#!/usr/bin/python
# -*- coding: utf-8 -*
from django.db import models
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _
from django.utils.http import urlquote
from brightway.apps.base.fields.uuidfield import UUIDField

class AutomaticURLModel(models.Model):
    """
Abstract base class that provides common URL methods.

URL Methods::

#. get_absolute_url: resolves to get_view_url
#. get_view_url
#. get_add_url
#. get_edit_url
#. get_delete_url

Needed Parameters::

#. _model_string: a string, e.g. "process". Convention is lower case model name.
#. _url_dict: a dictionary of parameters passed to url, e.g. {'name': self.name, 'id': self.id}

    """
    uuid = UUIDField()

    class Meta:
        abstract = True

    def get_absolute_url(self):
        return self.get_view_url()

    @models.permalink
    def get_view_url(self):
        return ('%s-view' % self._model_string, (), self._url_dict)

    @models.permalink
    def get_add_url(self):
        return ('%s-add' % self._model_string, (), {})

    @models.permalink
    def get_edit_url(self):
        return ('%s-edit' % self._model_string, (), self._url_dict)

    @models.permalink
    def get_delete_url(self):
        return ('%s-delete' % self._model_string, (), self._url_dict)

    @models.permalink
    def get_admin_edit_url(self):
        return ('admin:%s_%s_change' % (self._meta.app_label,
            self._model_string), (self.id,), {})

    @models.permalink
    def get_admin_add_url(self):
        return ('admin:%s_%s_add' % (self._meta.app_label, 
            self._model_string), (), {})

    # Module name is the way Django refers to the model, e.g. a "process" 
    # model. It is a transform of the name derived from the custom manager. 
    # Models with custom managers should redefine this.
    @property
    def _module_name(self):
        return self._meta.module_name

    # This is the "name" column, e.g. Process.process. In most models in
    # brightway, this is just the lowercase class name, but it doesn't have
    # to be.
    @property
    def _model_string(self):
        return self._module_name

    lookup_fields = None

    def _default_lookup_fields(self):
        return [self._module_name, "id"]

    @property
    def _url_dict(self):
        fields = self.lookup_fields if self.lookup_fields else \
            self._default_lookup_fields()
        return dict([(o, urlquote(getattr(self, o))) for o in fields])
