#!/usr/bin/python
# -*- coding: utf-8 -*
from django.db import models
from django.utils.translation import ugettext_lazy as _
from brightway.apps.base.fields.automaticurlmodel import AutomaticURLModel

class CompoundNameModel(AutomaticURLModel):
    '''
Create a compound name field, which is a combination of parent.compoundname 
and self.__unicode__. Creation and updating (including chld models) are 
handled automatically.

.. note::

    This model is most efficient if it is registered with MPTT.

.. note::

    Inherited model must define self.__unicode__

    '''
    compoundname = models.TextField(_('compound name'), editable=False)
    parent = models.ForeignKey('self', null=True, blank=True,
        related_name='%(class)s_children', verbose_name=_('parent'))

    def save(self, refresh_compound_name=False, *args, **kwargs):
        _new_name = self._generate_compound_name()
        if not self.compoundname == _new_name:
            # Should catch both new and changed values
            refresh_compound_name = True
            self.compoundname = _new_name
        super(CompoundNameModel, self).save(*args, **kwargs)
        if refresh_compound_name:
            try:
                _children = self.get_children()
            except: # No MPTT
                _children = self.objects.filter(parent__id=self.id)
            for child in _children:
                child.save(refresh_compound_name=True) # Yeah recursion.

    class Meta:
        abstract = True

    def _generate_compound_name(self):
        if self.parent:
            return "%s - %s" % (self.parent.compoundname, self.__unicode__())
        else:
            return self.__unicode__()

    @property
    def root(self):
        # Try with MPTT, and fall back to recursion
        try:
            return self.get_root()
        except AttributeError:
            if not self.parent:
                return self
            else:
                obj = self
                while obj.parent is not None:
                    obj = obj.parent
                return obj

    # @classmethod
    # def generate_missing_compound_names(model):
    #     for instance in model.objects.filter(parent__isnull=True, compoundname__isnull=True):
    #         instance.save()

