from django.db import models

# Want to use native DB fields if possible
try:
    from django.conf.settings import DATABASE_ENGINE
except ImportError:
    DATABASE_ENGINE = None

import uuid

class UUIDVersionError(Exception):
    pass

class UUIDField(models.CharField):
    """
UUIDField for Django, supports all uuid versions which are natively suported 
by the uuid python module.
    """
    __metaclass__ = models.SubfieldBase
    
    def __init__(self, verbose_name=None, version=4, 
            name=None, namespace=None, # Custom args for versions 3+5
            node=None, clock_seq=None, # Custom args for version 1
            *args, **kwargs):
        self.version = version
        kwargs['blank'] = True # Can set manually or let be set automatically
        kwargs['editable'] = kwargs.get('editable', False)
        kwargs['unique'] = True
        kwargs['db_index'] = True
        kwargs['max_length'] = 36
        if self.version==1:
            self.node, self.clock_seq = node, clock_seq
        elif self.version in (3,5):
            self.namespace, self.name = namespace, name
        super(UUIDField, self).__init__(self, *args, **kwargs)
    
    def create_uuid(self):
        if not self.version or self.version==4:
            return uuid.uuid4()
        elif self.version==1:
            return uuid.uuid1(self.node, self.clock_seq)
        elif self.version==2:
            raise UUIDVersionError("UUID version 2 is not supported.")
        elif self.version==3:
            return uuid.uuid3(self.namespace, self.name)
        elif self.version==5:
            return uuid.uuid5(self.namespace, self.name)
        else:
            raise UUIDVersionError("UUID version %s is not valid." % \
                self.version)

    def pre_save(self, model_instance, add):
        value = getattr(model_instance, self.attname)
        if value:
            return value # Don't overwrite UUID if provided.
        else:
            value = unicode(self.create_uuid())
            setattr(model_instance, self.attname, value)
            return value

    def __unicode__(self):
        return u"UUID"

try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([
        (
            [UUIDField], # Class(es) these apply to
            [],         # Positional arguments (not used)
            {           # Keyword argument
                "version": ["version", {"default": 4}],
                "db_index": ["db_index", {"default": True}],
                "unique": ["unique", {"default": True}],
                "blank": ["blank", {"default": True}],
                "max_length": ["max_length", {"default": 36}]
            },
        ),
    ], ["^brightway\.core\.fields\.uuidfield"])
except ImportError:
    pass
