# Adapted from 
# http://www.arnebrodowski.de/blog/distributing-mediafiles-with-django-apps.html
import os
import sys
from django.conf import settings
from django.core.management.color import color_style

def link_plugin_media(sender, verbosity, **kwargs):
    """
This function is called whenever django's `post_syncdb` signal is fired.
It looks if the sending app has its own media directory ('media'). 
If a directory with this name exists under the app directory
and has a subdirectory named as the app itself, this subdirectory is then
symlinked to a ``MEDIA_ROOT/apps`` directory, if it doesn't already exist.

Example:

    An app called ``foo`` (as listed in INSTALLED_APPS) needs to
    distribute some JavaScript files. The files are stored in
    `/plugins/foo/media/foo/js/*`.
    
    Now, whenever `manage.py syncdb` is run, the directory
    `/plugins/foo/media/foo` is linked to MEDIA_ROOT/plugins/foo and therefore the
    JavaScript files are accessible in the templates or as form media via:

        {{ MEDIA_ROOT }}plugins/foo/js/example.js

The symlink will not be created if a resource with the destination name
already exists.
    """
    app_name = sender.__name__.split('.')[-2]
    app_dir = os.path.dirname(sender.__file__)

    PLUGIN_MEDIA_DIRNAME = 'media'
    PLUGIN_MEDIA_SITES = 'plugins'

    app_media = os.path.join(app_dir, PLUGIN_MEDIA_DIRNAME, app_name)
    if os.path.exists(app_media):
        dest = os.path.join(settings.STATIC_ROOT, PLUGIN_MEDIA_SITES, 
            app_name)
        if not os.path.exists(dest):
            try:
                os.symlink(app_media, dest) # will not work on windows.
                print "symlinked app_media dir for app: %s" % app_name
            except:
                # windows users should get a note, that they should copy the
                # media files to the destination.
                error_msg = "Failed to link media for '%s'\n" % app_name
                instruction = ("Please copy the media files to the", 
                    " STATIC_ROOT manually\n")
                sys.stderr.write(color_style().ERROR(str(error_msg)))
                sys.stderr.write(" ".join(instruction))