# -*- coding: utf-8 -*
import random
import string
from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.utils.http import urlquote
import mptt
from brightway.apps.base.fields.automaticurlmodel import AutomaticURLModel

# TODO: Create links to processes, exchanges, etc.

class Person(AutomaticURLModel):
    person = models.TextField(_('person'))
    # Embed as hCard in web pages; AUMP as parser?
    vcard = models.TextField(_('vcard'), blank=True)

    class Meta:
        verbose_name = _('person')
        verbose_name_plural = _('people')

    def __unicode__(self):
        return self.person

    def save(self, *args, **kwargs):
        # TODO: Add vobject verification of vCard?
        # TODO: Should be able to supply file, not cut and paste
        # TODO: Need to parse XML files correctly
        if not self.vcard:
            import vobject
            vcard = vobject.vobject.vCard()
            vcard.add('n')
            vcard.add('fn') # Required elements
            names = self.person.split(" ")
            if not length(names) > 1:
                # TODO: Raise validation error?
                last_name = names[0]
                first_name = u'Unknown'
            else:
                last_name = u" ".join(names[1:])
                first_name = names[0]
            vcard.n.value = vobject.crad.Name(family=last_name, 
                given=first_name)
            vcard.fn.value = self.person
            self.vcard = vcard.serialize()
        super(Person, self).save(*args, **kwargs)


def get_path(instance, filename):
    """To preserve sanity, separate data files into separate directories based on the objects they are linked to. Also adds a random suffix to filenames."""
    split = filename.split(".")
    filename = ".".join(split[:-1]) + "." + string.join(random.sample( 
        string.letters, 4), '') + "." + split[-1]
    elems = ("uploads/data_source", str(instance.content_type.app_label), 
        str(instance.object_id), filename)
    return "/".join(elems)

class Source(AutomaticURLModel):
    UNDEFINED = 0
    ARTICLE = 1
    CHAPTER = 2
    SEPARATE_PUBLICATION = 3
    ON_SITE_MEASUREMENT = 4
    ORAL_COMMUNICATION = 5
    WRITTEN_COMMUNICATION = 6
    QUESTIONNAIRE = 7

    SOURCE_TYPE_CHOICES = ( # TODO: Defined in bibtex model?
        (UNDEFINED, _('undefined')),
        (ARTICLE, _('article')),
        (CHAPTER, _('chapter')),
        (SEPARATE_PUBLICATION, _('separate publication')),
        (ON_SITE_MEASUREMENT, _('on-site measurement')),
        (ORAL_COMMUNICATION, _('oral communication')),
        (WRITTEN_COMMUNICATION, _('written communication')),
        (QUESTIONNAIRE, _('questionnaire')),
    )

    title = models.TextField(_('title'))
    people = models.ManyToManyField(Person, verbose_name=_('people'), 
        null=True, blank=True)
    file = models.FileField(_('source document'), upload_to=get_path, 
        null=True, blank=True)
    link = models.URLField(_('source link'), blank=True, null=True)
    bibtex = models.TextField(_('bibtex'), blank=True, default='')
    content_type = models.ForeignKey(ContentType, verbose_name=_(
        'content type'))
    object_id = models.PositiveIntegerField(_('object_id'))
    content_object = generic.GenericForeignKey('content_type', 'object_id')

    class Meta:
        verbose_name = _('source')
        verbose_name_plural = _('sources')

    def __unicode__(self):
        return self.title

    @property
    def _module_name(self):
        return "title"

    @property
    def author_count(self):
        return self.people.count()

class DataEntry(AutomaticURLModel):
    """
A unique pair of data entry people and sources.
    """
    data_monkey = models.ForeignKey(Person, verbose_name=_('person'))
    data_source = models.ForeignKey(Source, verbose_name=_('source'))

    class Meta:
        verbose_name = _('data entry person')
        verbose_name_plural = _('data entry people')
        unique_together = ('data_monkey', 'data_source')

    def __unicode__(self):
        return self.data_monkey.__unicode__() + u": " + \
            self.data_source.__unicode__()


class Validation(AutomaticURLModel):
    details = models.TextField(_('validation'), blank=True, default='')
    validator = models.ForeignKey(Person, verbose_name=_('validator'))
    source = models.ForeignKey(Source, verbose_name=_('source'))

    class Meta:
        verbose_name = _('validation')
        verbose_name_plural = _('validations')

    def __unicode__(self):
        return self.validator.__unicode__() + u":" + \
            self.source.__unicode__()

