# -*- coding: utf-8 -*
from datetime import datetime
from django.utils.http import urlquote
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.contenttypes import generic
from brightway.apps.group_permissions.models import GroupPermission
from brightway.apps.base.fields.automaticurlmodel import AutomaticURLModel
from brightway.apps.group_permissions.managers import PermissionManager
from brightway.apps.license.models import License
import bw_attributes as attributes
from django_dag.mixins import GraphMixin
from django_dag.models import Graph
from brightway.apps.geo.errors import MultipleGlobalGeographies, \
    GlobalGeographyDatabase
from brightway.apps.geo import WORLD

class GeographyDatabase(AutomaticURLModel, GraphMixin):
    database = models.TextField(_('name'), unique=True)
    license = models.ForeignKey(License, null=True, blank=True, 
        related_name="geography_database_license", verbose_name=_("license"))
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True, 
        default='')
    modified = models.DateTimeField(_('modified'), editable=False, 
        auto_now=True)

    group_permission_set = generic.GenericRelation(GroupPermission)
    permitted = PermissionManager()

    objects = models.Manager()

    class Meta:
        verbose_name = _('geography database')
        verbose_name_plural = _('geography databases')

    def __unicode__(self):
        return self.database

    def save(self, *args, **kwargs):
        self.modified = datetime.now()
        super(GeographyDatabase, self).save(*args, **kwargs)

    @property
    def _module_name(self):
        return "database"

    def geography_count(self):
        return self.geography_set.count()
    geography_count.short_description = _('geography count')

    # Set default graph for graph mixin
    @property
    def default_graph(self):
        return Graph.objects.get_or_create(name="geography database")[0]


class Geography(AutomaticURLModel):
    geography = models.TextField(_('name'),)
    # No validation until a design decision is made on whether or not to
    # support invalid or obsolete codes.
    alpha2 = models.CharField(_('ISO alpha 2 letter code'), max_length=2,
        blank=True, default='',
        help_text='http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2')
    alpha3 = models.CharField(_('ISO alpha 3 letter code'), max_length=3,
        blank=True, default = '',
        help_text='http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3')
    numeric = models.IntegerField(_('numeric code'), blank=True, null=True,
        help_text='http://unstats.un.org/unsd/methods/m49/m49regin.htm')
    imported_database_code = models.CharField(_(
        'id number or hash in imported database'), max_length=255, blank=True, 
        default = '') # Not unique
    description = models.TextField(_('description'), blank=True, default='')
    longitude = models.FloatField(_('longitude'), blank=True, null=True)
    latitude = models.FloatField(_('latitude'), blank=True, null=True)
    other_info = models.TextField(_('other information'), blank=True,
        default = '')
    database = models.ForeignKey(GeographyDatabase, 
        verbose_name=_('database'))
    modified = models.DateTimeField(_('modified'), editable=False, 
        auto_now=True)

    # a set of common values for regions. Other possible attributes:
    # FIPS Code (http://www.census.gov/geo/www/fips/fips.html)
    # UN Code
    # Region and subregion codes
    area = models.FloatField(_('area'), blank=True, null=True)
    pop_density_km = models.FloatField(_('population density (per km^2)'),
        blank=True, null=True)
    pop_total = models.FloatField(_('population'), blank=True, null=True)
    # Commenting out for now - should be restored at some point
    # shape = models.TextField(_('shape'), blank=True, default='') # dummy

    lookup_names = ["database", "geography"]

    class Meta:
        verbose_name = _('geography')
        verbose_name_plural = _('geographies')
        unique_together = ('geography', 'database') 

    def __unicode__(self):
        return self.geography

    def __repr__(self):
        return "<Geography: %s>" % self.geography

    @property
    def _module_name(self):
        return "geography"

    @models.permalink
    def get_view_url(self):
        return ('geography-view', (), {
            'database': urlquote(self.database.database), 
            'geography': urlquote(self.geography)})

    def save(self, skip_datetime_update=False, **kwargs):
        '''
Update database modified field - used by LCA logic to determine whether caches 
are up to date.
        '''
        self._test_global()
        super(Geography, self).save(**kwargs)
        if not skip_datetime_update:
            self.database.save()

    def _test_global(self):
        """Only allow one geography named 'Global', and only allow 'Global' in 'Global' database."""
        # Only check for errors after initial defaults are created.
        if not Geography.objects.filter(geography="Global").exclude( 
                id=self.id).exists():
            return
        if self.geography == "Global":
            # Can only have one *Global* geography in database
            raise MultipleGlobalGeographies, \
                "Can only have one geography named 'Global'."
        if self.database.id == WORLD().database.id:
            raise GlobalGeographyDatabase, \
                "Can't save new geographies to 'Global' database."

    def fullname(self):
        # Includes database
        return u'%s: %s' % (self.database, self.geography)

    @property
    def code(self):
        if self.alpha2:
            return self.alpha2
        elif self.alpha3:
            return self.alpha3
        elif self.imported_database_code:
            return self.imported_database_code
        else:
            return None

    def process_count(self):
        return self.process_set.count()
    process_count.short_description = _('process count')

    def weighting_count(self):
        return self.weighting_geography_set.count()
    weighting_count.short_description = _('weighting count')

    def has_lat_long(self):
        return bool(self.longitude) and bool(self.latitude)
    has_lat_long.boolean=True
    has_lat_long.short_description = _('lat/lon coordinates')


attributes.register(Geography)
