from django.contrib.auth.models import Group, User
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.contenttypes.models import ContentType
from brightway.apps.group_permissions.models import GroupPermission

def get_permission_for_instance_and_user(instance, user):
    """
Utility function to get the most permissive permission level for an 
instance/user combination.
    """
    if user.is_superuser:
        return 0
    elif hasattr(instance, 'creator') and instance.creator == user:
        return 0
    else:
        contenttype = ContentType.objects.get_for_model(instance)
        g = GroupPermission.objects.filter( 
                group__id__in=user.groups.values_list('id', flat=True), 
                content_type=contenttype, object_id=instance.id 
                ).order_by('-permission')
        if g:
            return g[0].permission
        else:
            return 3

def filter_queryset_by_permission(queryset, permission=3, user_or_request=None):
    if not user_or_request and permission > 3:
        raise AttributeError
    if isinstance(user_or_request, User):
        user = user_or_request
    elif hasattr(user_or_request, 'user'):
        user = user_or_request.user
    else:
        user = None

    # Admins can see everything
    if hasattr(user, "is_superuser") and user.is_superuser:
        return queryset

    content_type = ContentType.objects.get_for_model(queryset.model)
    # How to do this with Django 1.1 features?

    if user:
        groups = user.groups.all()
    else:
        groups = Group.objects.filter(name="Everyone")

    # TODO: As this is called a lot, perhaps it could be made more efficient
    # to force joins?
    return queryset.filter(id__in=GroupPermission.objects.filter(
        permission__gte=permission, 
        content_type=content_type, 
        group__in=groups).values_list('object_id', flat=True))

def create_update_permission(instance, group, permission):
    """
    Utility function for easy setting and updating of GroupPermission objects.
    """
    contenttype_id = ContentType.objects.get_for_model(instance).id
    try:
        g = GroupPermission.objects.get(content_type__id=contenttype_id, 
            object_id=instance.id, group=group)
        g.permission = permission
        g.save()
    except ObjectDoesNotExist:
        g = GroupPermission.objects.create(content_object=instance, 
            group=group, permission=permission)
    return g