# -*- coding: utf-8 -*
from datetime import datetime
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.http import urlquote
import mptt
from brightway.apps.parameter.models import ParameterModel
from brightway.apps.base.fields.compoundnamemodel import CompoundNameModel
from brightway.apps.base.fields.automaticurlmodel import AutomaticURLModel
from brightway.apps.inventory.models import Process
from brightway.apps.unit.models import Unit
from brightway.apps.impact_assessment.managers import MethodManager
from brightway.apps.base.fields.jsonfield import JSONField


class Method(CompoundNameModel):
    method = models.TextField(_('name'),)
    description = models.TextField(_('description'), blank=True, default='')
    unit = models.ForeignKey(Unit, verbose_name=_("unit"))
    other_info = models.TextField(_('other information'), blank=True,
        default='')
    modified = models.DateTimeField(_('modified'), editable=False, 
        auto_now=True)

    objects = MethodManager()

    lookup_names = ["compoundname"]

    class Meta:
        verbose_name = _('method')
        verbose_name_plural = _('methods')
        unique_together = ('compoundname',)
        db_table = 'ia_method'

    def __unicode__(self):
        return self.method

    @models.permalink
    def get_view_url(self):
        return ('method-view', (), {
            'compoundname': urlquote(self.compoundname)})

    def save(self, skip_datetime_update=False, *args, **kwargs):
        super(Method, self).save(*args, **kwargs)
        if not skip_datetime_update:
            self.get_ancestors().update(modified=datetime.now())

    def fullname(self):
        return u'%s: %s' % self.compoundname
    fullname.short_description = _('full name')

    def weighting_count(self):
        return self.weighting_set.count()

    def get_unique_process_strings_for_method(self):
        rec_ids = Weighting.objects.filter(method__id=self.id).values_list( 
            'process', flat=True)
        return set(Process.objects.filter(id__in=rec_ids).values_list( 
            'process', flat=True))

mptt.register(Method, order_insertion_by=['method',])


class Weighting(ParameterModel):
    """A LCIA characterisation factor linked to a biosphere flow."""
    method = models.ForeignKey(Method, verbose_name=_('method'))
    process = models.ForeignKey(Process, limit_choices_to= \
        {'process_type': Process.EMISSION_TYPE},
        related_name='weighting_set', verbose_name=_('process'))
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True,
        default='')

    lookup_names = ["uuid"]

    class Meta:
        unique_together = ('method', 'process', 'geography')
        verbose_name = _('weighting')
        verbose_name_plural = _('weightings')
        db_table = 'ia_weighting'

    def __unicode__(self):
        return u'%.2g: %s' % (self.amount, self.process)

    def save(self, skip_datetime_update=False, *args, **kwargs):
        super(Weighting, self).save(*args, **kwargs)
        if not skip_datetime_update:
            self.method.get_ancestors().update(modified=datetime.now())

    @property
    def _url_dict(self):
        return {'uuid': self.uuid}


class Calculation(AutomaticURLModel):
    """
    A saved setup for LCI/LCIA calculations.
    
    Instead of many-to-many fields, store setup as a dictionary:
    {"processes": [{"process": process name,
            "process_view_url": process url,
            "id": process id,
            "amount": amount of process desired},
            ],
    "methods": [{"method": method name,
            "method_view_url": method_view_url, 
            "id": method id, 
            "submethods": include submethods or not},
            ],
    "datetime": datetime.datetime.now(),
        }
    """
    name = models.TextField(_("name"), blank=True, default="")
    description = models.TextField(_("description"), blank=True, default="")
    setup = JSONField()

    class Meta:
        verbose_name = _("LCI/LCIA calculation")
        verbose_name_plural = _("LCI/LCIA calculations")
        db_table = 'ia_calculation'

    def __unicode__(self):
        return self.name if self.name else _(
            u"Anonymous LCI/LCIA Calculation")
