# -*- coding: utf-8 -*-
from brightway.core.lca.lca import LCA
from brightway.apps.impact_assessment.models import Method, Weighting
from brightway.apps.inventory.models import Category
from brightway.tests.lca_test_case import LCATestCase
from brightway.core.inheritance.queries import \
    recursive_weightings_by_method, recursive_weightings_count_by_method
from brightway.core import Container


class NoInheritanceTestCase(LCATestCase):
    '''Make sure custom SQL works without inheritance.'''
    def test_weighting_qs(self):
        """Make sure all weightings are correct with no inheritance."""
        self.assertEqual(Weighting.objects.filter( 
            method=self.methods.human_health).count(), 2)
        self.assertEqual(len(list(recursive_weightings_by_method( 
            self.methods.human_health))), 2)
        self.assertEqual(recursive_weightings_count_by_method( 
            self.methods.human_health), 2)


class InheritanceTestCase(LCATestCase):
    """Test that custom queries and LCA correctly implements weighting inheritance"""
    def setUp(self):
        """
Construct a tree of biosphere flows and weightings.

Biosphere flows:

(self.categories.airemissions) Air--+- NOx (self.processes.nox)
                               |    +- PM (self.processes.pm)
                               |    +- Toxin (self.process.toxin)
                               |
       (self.categories.hpd) High Pop. Density--+- NOx (self.toxin.nox)
                                                +- Toxin (self.children.toxin)
                        (self.processes.dioxin) +- Dioxin 

(self.categories.earth) Earth--+- Coal Resource (self.processes.coalresource)
                          |
                          +-----+
                                |
(self.categories.mountains) Mountains--+- Coal Resource (self.children.coal)

Method: Human Health (self.methods.human_health)

Weightings:

* NOx, Air, 1 (LCA test case)
* PM, Air, 2 (LCA test case)
* NOx, HPD, 0
* Coal Resource, HPD, 15
* Toxin, Air, 10
* Toxin, HPD, 10 (implied)
* Dioxin, HPD, 1000

Test cases:
* Overwritten weighting should have new value: NOx Air, HPD
* Child flow should inherit weighting: Toxin Air, HPD
* Independent trees should work correctly: Air, Earth
* Child flow weighting should not propogate to parent: Coal Resource
* Flow defined on subcategory should exist independently: Dioxin
        """
        super(InheritanceTestCase, self).setUp()
        self.categories.hpd = self.create_category(
            category="High Population Density", 
            parent=self.categories.airemissions)
        self.categories.mountains = self.create_category(
            category="Mountains", parent=self.categories.earth)

        self.children = Container()
        self.children.nox = self.create_process("NOx", 
            self.categories.hpd, process_type=2, parent=self.processes.nox)
        self.children.coal = self.create_process("Coal Resource", 
            self.categories.mountains, process_type=1, 
            parent=self.processes.coalresource)
        self.processes.toxin = self.create_process("Toxin", 
            category=self.categories.airemissions, process_type=2)
        self.children.toxin = self.create_process("Toxin", 
            category=self.categories.hpd, process_type=2, 
            parent=self.processes.toxin)
        self.processes.dioxin = self.create_process("Dioxin", 
            category=self.categories.hpd, process_type=2)

        self.create_weighting(self.methods.human_health, self.children.nox, 
            0)
        self.create_weighting(self.methods.human_health, self.children.coal, 
            15)
        self.create_weighting(self.methods.human_health, self.processes.toxin, 
            10)
        self.create_weighting(self.methods.human_health, 
            self.processes.dioxin, 1000)

    def test_number_of_weightings(self):
        """Count explicit and implied weightings."""
        self.assertEquals(len(list(recursive_weightings_by_method( 
            self.methods.human_health))), 7)

    def test_count_query(self):
        """Count explicit and implied weightings with count query."""
        self.assertEqual(recursive_weightings_count_by_method( 
            self.methods.human_health), 7)

    def test_weighting_values(self):
        results = {}
        for obj in recursive_weightings_by_method(self.methods.human_health):
            results[obj['process_id']] = obj['amount']
        self.assertEqual(results[self.processes.nox.id], 1)
        self.assertEqual(results[self.processes.pm.id], 2)
        self.assertEqual(results[self.children.nox.id], 0)
        self.assertEqual(results[self.children.coal.id], 15)
        self.assertEqual(results[self.processes.toxin.id], 10)
        self.assertEqual(results[self.children.toxin.id], 10)
        self.assertEqual(results[self.processes.dioxin.id], 1000)
