from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from django.http import HttpResponseRedirect 
from django.core.urlresolvers import reverse 
from brightway.apps.inventory.models import *
from brightway.apps.group_permissions.admin import GroupPermissionAdminInline


class DatabaseAdmin(admin.ModelAdmin):
    save_on_top = True
    search_fields = ('database',)
    list_display = ('database', 'description', 'category_count', 
        'process_count', 'creator')
    # list_display_links = ('database',)
    # date_hierarchy = 'created'
    ordering = ('database', 'creator')
    list_filter = ("creator", "license")
    fieldsets = (
        (None, {
            'fields': (('database', 'description', 'license'), ('creator', 
                'modifier')),
            }
        ),
    )
    # inlines = [GroupPermissionAdminInline,]

class DatabasePermissionAdmin(admin.ModelAdmin):
    save_on_top = True
    search_fields = ('group',)
    list_display = ('database', 'group', 'permission')
    list_filter = ('database', 'permission')
    ordering = ('database', 'group')
    save_as = True
    radio_fields = {
        'permission': admin.VERTICAL,
        }
    fieldsets = (
        (None, {
            'fields': (('group', 'database',), 'permission'),
            }
        ),
    )


class CategoryAdmin(TreeEditor):
    save_on_top = True
    search_fields = ('category', 'parent__category',)
    list_display = ('category', 'parent', 'database', 'process_count', 
        'description')
    list_filter = ('database',)
    ordering = ('database', 'parent__category', 'category',)
    fieldsets = (
        (None, {
            'fields': (('category', 'description',), 'parent', 'database'),
            'description': "Category name/database combinations must be unique."
            }
        ),
    )

admin.site.register(Category, CategoryAdmin)


class ProcessAdmin(admin.ModelAdmin): #VersionAdmin): 
    #TODO: Use MPTT properties to do something nice here.
    save_on_top = True
    search_fields = ('process',)
    list_display = ('process', 'parent', 'unit', 'database', 
        'process_type', 'category', 'selection_type', 'description', 
        'creator')
    # list_display_links = ('process',)
    # list_editable = ('unit', 'database', 'process_type', 'selection_type')
    ordering = ('process',)
    radio_fields = {
        'process_type': admin.VERTICAL,
        'selection_type': admin.VERTICAL,
        }
    list_filter = ('process_type', 'selection_type', 'database')
    raw_id_fields = ('parent', 'geography', 'category')
    # date_hierarchy = 'created'
    save_as = True
    fieldsets = (
        (None, {
            'fields': ('process', 'description', 'parent', 'unit', 'database', 
                'category', 'geography', 'process_type', 'selection_type'),
            }
        ),
        ('User & Creation', {
            'classes': ('collapse',),
            'fields': ('creator', 'modifier'),
            }
        ),
        ('Details', {
            'classes': ('collapse',),
            'fields': ('imported_database_code', 'cas_number', 
                'chemical_formula', 'original_format',),
            }
        )
    )
    # inlines = [AttributeAdminInline,]
    actions = ['export_process', 'export_process_excel']
    
    def export_process(self, request, queryset):
        process_count = queryset.count()
        # self.message_user(request, _(u"%s process(es) being exported." % \
        # process_count))
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        return HttpResponseRedirect("%s?ids=%s&zipped=1" % \
            (reverse('export_process_csv'), ",".join(selected)))
    export_process.short_description = _(u"Export process to zipped CSV")

    def export_process_excel(self, request, queryset):
        process_count = queryset.count()
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        return HttpResponseRedirect("%s?ids=%s" % (reverse('export_process', 
            {'export_type': 'excel'}), ",".join(selected)))
    export_process.short_description = _(u"Export process to spreadsheet")


class ExchangeAdmin(admin.ModelAdmin): #VersionAdmin): 
    #TODO: Use MPTT properties to do something nice here.
    save_on_top = True
    search_fields = ('process',)
    list_display = ('amount', 'has_formula', 'uncertainty_type', 
        'description', )
    radio_fields = {
        'uncertainty_type': admin.VERTICAL,
        'input_group': admin.VERTICAL,
        'output_group': admin.VERTICAL,
        }
    raw_id_fields = ('input', 'output')
    date_hierarchy = 'created'
    save_as = True
    list_per_page = 25
    fieldsets = (
        (None, {
            'fields': ('amount', 'formula', 'input', 'input_group', 'output', 
                'output_group', 'description'),
            }
        ),
        ('User & Creation', {
            'classes': ('collapse',),
            'fields': ('creator', 'modifier'),
            }
        ),
        ('Uncertainty', {
            'classes': ('collapse',),
            'fields': ('uncertainty_type', 'sigma', 'minimum', 'maximum'),
            }
        ),
    )
    # inlines = [AttributeAdminInline,]

admin.site.register(Process, ProcessAdmin)
admin.site.register(Exchange, ExchangeAdmin)
admin.site.register(Database, DatabaseAdmin)
admin.site.register(DatabasePermission, DatabasePermissionAdmin)
