from django.db import connection, models
from django.db.models.query import ValuesQuerySet, QuerySet
# Using aggregates across table joins is slow, because the generated SQL
# stinks; see http://code.djangoproject.com/ticket/10060
# These are utility functions that can accept Querysets, ValuesQuerysets, or 
# lists of dictionaries. In general, use ids for foreign keys because they 
# have indices in the database.

DATABASE_CATEGORY_COUNT_SQL = """SELECT database_id, COUNT(*) 
FROM inventory_category 
    WHERE database_id IN %s 
    GROUP BY database_id"""

DATABASE_PROCESS_COUNT_SQL = """SELECT database_id, COUNT(*) 
FROM inventory_process
    WHERE database_id IN %s 
    GROUP BY database_id"""

CATEGORY_PROCESS_COUNT_SQL = """SELECT category_id, COUNT(*) 
FROM inventory_process
    WHERE category_id IN %s 
    GROUP BY category_id"""

def ids_for_object(qs):
    if isinstance(qs, (list, tuple)):
        ids = [dic['id'] for dic in qs]
    elif isinstance(qs, ValuesQuerySet):
        ids = [values['id'] for values in qs]
    elif isinstance(qs, QuerySet):
        ids = qs.values_list('id', flat=True)
    else:
        raise ValueError
    return ids

def database_category_count(qs):
    cursor = connection.cursor()
    cursor.execute(DATABASE_CATEGORY_COUNT_SQL, [tuple(ids_for_object(qs))])
    dct = {}
    for database, count in cursor.fetchall():
        dct[database] = {'count': count}
    return dct

def database_process_count(qs):
    cursor = connection.cursor()
    cursor.execute(DATABASE_PROCESS_COUNT_SQL, [tuple(ids_for_object(qs))])
    dct = {}
    for database, count in cursor.fetchall():
        dct[database] = {'count': count}
    return dct

def category_process_count(qs):
    cursor = connection.cursor()
    cursor.execute(CATEGORY_PROCESS_COUNT_SQL, [tuple(ids_for_object(qs))])
    dct = {}
    for category, count in cursor.fetchall():
        dct[category] = {'count': count}
    return dct


class SQLCreatorMinCase(models.sql.aggregates.Aggregate):
    """Allow case statement with inside aggregate.
    
    Used for testing database permissions, which are a function of creator and group database permissions."""
    # Always need to check to make sure SQL generated is correct
    # Based on http://djangosnippets.org/snippets/2099/
    # See also http://www.voteruniverse.com/Members/jlantz/blog/conditional-aggregates-in-django
    is_ordinal = True
    sql_function = 'MIN'
    sql_template = "( CASE WHEN creator_id = %(user_id)s THEN 0 ELSE %(function)s(%(field)s) END )"

class CreatorMinCase(models.Aggregate):
    """Class used in queryset. Usage: CreatorMinCase(column, user_id=user.id)."""
    name = 'MIN'

    def add_to_query(self, query, alias, col, source, is_summary):
        aggregate = SQLCreatorMinCase(col, source=source, 
            is_summary=is_summary, **self.extra)
        query.aggregates[alias] = aggregate