# -*- coding: utf-8 -*
from datetime import datetime
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.http import urlquote
from django.db.models.signals import post_save
from django.contrib.auth.models import Group
from django.core.exceptions import ObjectDoesNotExist
import bw_attributes as attributes
from django_dag.mixins import GraphMixin
from django_dag.models import Graph
from django_dag.errors import CircularReferenceError
import mptt
from brightway.apps.parameter.models import ParameterModel
from brightway.apps.base.fields.compoundnamemodel import CompoundNameModel
from brightway.apps.base.fields.autopopulatemodel import AutopopulateModel
from brightway.apps.base.fields.jsonfield import JSONField
from brightway.apps.geo.models import Geography
from brightway.apps.unit.models import Unit
from brightway.apps.license.models import License
from brightway.core.inheritance.queries import recursive_exchanges_by_process
from brightway.apps.inventory.managers import ExchangeManager, DatabaseManager
from brightway.apps.inventory.errors import GeographyBiosphereError, \
    BiosphereFlowAlreadyExists, UnfreezableProcess, FrozenOutput, FrozenParent
from brightway.apps.geo import WORLD
from brightway.apps.inventory.signals import \
    add_default_database_everyone_permission

class Database(AutopopulateModel, GraphMixin):
    DATABASE_TYPES = (
        (0, _('Library')), 
        (1, _('Template')),
        (2, _('Project')),
        (3, _('Private Demo')),
        (4, _('Public Demo')),
    )

    database = models.TextField(_('name'))
    license = models.ForeignKey(License, null=True, blank=True, 
        related_name="database_license", verbose_name=_("license"))
    goal = models.TextField(_('goal'), blank=True, default='')
    scope = models.TextField(_('scope'), blank=True, default='')
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True, 
        default='')
    database_type = models.IntegerField(_('type'), choices=DATABASE_TYPES, 
        default=2)

    objects = DatabaseManager()
    lookup_fields = ["database"]

    class Meta:
        verbose_name = _('database')
        verbose_name_plural = _('databases')
        unique_together = ('database', 'database_type')

    def __unicode__(self):
        return self.database

    @property
    def default_graph(self):
        return Graph.objects.get(name="database")

    def get_permission(self, user=None):
        try:
            permission = Database.objects.filter(id=self.id).permissions( 
                user=user).get().permission
        except ObjectDoesNotExist:
            permission = 4
        return permission

    def process_count(self):
        return self.process_set.count()
    process_count.short_description = _('process count')

    def category_count(self):
        return self.category_set.filter(parent__isnull=True).count()
    category_count.short_description = _('category count')

post_save.connect(add_default_database_everyone_permission, sender=Database)


class DatabasePermission(models.Model):
    PERMISSION_CHOICES = (
        (0, _('open')), # Freely editable by anyone
        (1, _('usable')), # Freely readable, but not changable
        (2, _('restricted')), # Can't create variants or export
        (3, _('private')), # Private
        )

    group = models.ForeignKey(Group, verbose_name=_("group"))
    database = models.ForeignKey(Database, verbose_name=_("database"))
    permission = models.IntegerField(_("permission"), 
        choices=PERMISSION_CHOICES)

    class Meta:
        verbose_name = _('database permission')
        verbose_name_plural = _('database permissions')
        unique_together = ('group', 'database')

    def __unicode__(self):
        return "%s permission for group %s - %s" % (self.database, self.group, 
            self.permission)


class Category(CompoundNameModel):
    category = models.TextField(_('name'),)
    database = models.ForeignKey(Database, null=True, blank=True,
        related_name='category_set', verbose_name=_('database'))
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True, 
        default='')

    lookup_fields = ["_database_name", "compoundname"]

    @property
    def _database_name(self):
        return self.database.database

    class Meta:
        verbose_name = _('category')
        verbose_name_plural = _('categories')
        unique_together = ('category', 'database', 'parent')

    def __unicode__(self):
        return self.category

    def process_count(self):
        return self.process_set.count()
    process_count.short_description = _('process count')

mptt.register(Category, order_insertion_by=['category',])


class Process(AutopopulateModel):
    '''
A technological process is something with material inputs, outputs, and 
emissions. For example, a car manufacture process requires steel, rubber, 
and plastic, and outputs one car.
    '''
    TECHNOLOGICAL_PROCESS_TYPE = 0
    RESOURCE_CONSUMPTION_TYPE = 1
    EMISSION_TYPE = 2
    FROZEN_TYPE = 3

    PROCESS_TYPE_CHOICES = (
        (TECHNOLOGICAL_PROCESS_TYPE, _('technological process')),
        (RESOURCE_CONSUMPTION_TYPE, _('resource consumption')),
        (EMISSION_TYPE, _('emission'))
        )

    NO_SELECTION = 0
    ALLOCATED_SELECTION = 1
    MOST_LIKELY_SELECTION = 2
    PROBABILITY_SELECTION = 3

    SELECTION_TYPE_CHOICES = (
        (NO_SELECTION, _('not a selection process (default)')),
        (ALLOCATED_SELECTION, _('allocated among choices')),
        (MOST_LIKELY_SELECTION, _('most likely choice taken')),
        (PROBABILITY_SELECTION, _('choice based on given probabilities')),
        )

    database = models.ForeignKey(Database, verbose_name=_('database'))
    process = models.TextField(_('name'))

    parent = models.ForeignKey('self', null=True, blank=True,
        related_name='childprocess_set', verbose_name=_('parent'))

    description = models.TextField(_('description'), blank=True, default='')
    category = models.ForeignKey(Category, verbose_name=('category'))
    process_type = models.IntegerField(_('type'),
        choices=PROCESS_TYPE_CHOICES, default=0, db_index=True)
    unit = models.ForeignKey(Unit, verbose_name=_('unit'))
    selection_type = models.IntegerField(_('selection field type'),
        choices=SELECTION_TYPE_CHOICES, default=0, db_index=True)
    # Chemical Abstract Service
    cas_number = models.CharField(_('Chemical Abstract Service number'), 
        max_length=11, blank=True, default='')
    chemical_formula = models.TextField(_('chemical formula'), blank=True, 
        default='')
    imported_database_code = models.CharField(_(
        'id number or hash in imported database'), max_length=255, blank=True, 
        default='') # Not unique
    other_info = models.TextField(_('other information'), blank=True, 
        default='')
    original_format = models.FileField(null=True, blank=True,
        upload_to='xml/originals')
    frozen_data = JSONField(blank=True, default="")

    # Attributes
    geography = models.ForeignKey(Geography, verbose_name=_('geography'))

    objects = models.Manager()

    class Meta:
        verbose_name = _('process')
        verbose_name_plural = _('processes')

    def __unicode__(self):
        return self.process

    def save(self, skip_datetime_update=False, 
            skip_database_inheritance_check=False, *args, **kwargs):
        """
Update database modified field - used by LCA logic to determine whether caches
are up to date.
        """
        if self.parent and self.parent.id == self.id:
            raise CircularReferenceError, \
                "Can't be parent and child at same time - it is gross."
        if self.parent and self.parent.process_type == self.FROZEN_TYPE:
            raise FrozenParent("Can't create children of a frozen process")
        if self.process_type in (self.RESOURCE_CONSUMPTION_TYPE, 
                self.EMISSION_TYPE) and self.geography != WORLD():
            raise GeographyBiosphereError, \
                "Biosphere flows can only have 'Global' geography."
        if not skip_database_inheritance_check:
            # Link databases in "inheritance" graph
            if self.parent and self.database_id != self.parent.database_id:
                self.parent.database.add_child(self.database, 
                    graph=Graph.objects.get(name="inheritance"))
        if self.process_type in (self.RESOURCE_CONSUMPTION_TYPE, 
                self.EMISSION_TYPE) and Process.objects.filter( 
                database=self.database, process=self.process, 
                category=self.category).exclude(id=self.id, 
                process_type=self.process_type).exists():
            raise BiosphereFlowAlreadyExists, \
                "This biosphere flow already exists."
        super(Process, self).save(*args, **kwargs)
        if not skip_datetime_update:
            self.database.save()

    @property
    def children(self):
        return Process.objects.filter(parent__id=self.id)

    def get_inherited_exchanges(self, as_dict=True, add_links=True):
        # TODO: Will this route correctly is there are multiple dbs?
        if self.process_type == self.FROZEN_TYPE:
            raise ValueError("This process is frozen and has no edges")
        return recursive_exchanges_by_process(self, as_dict=as_dict, 
            add_links=add_links)

    @property
    def input_set(self):
        return Exchange.objects.filter(output__id=self.id, 
            input__process_type=0).extra(where=['input_id <> output_id']
            )

    @property
    def emission_set(self):
        return Exchange.objects.filter(output__id=self.id, 
            input__process_type=2).extra(where=['input_id <> output_id']
            )

    @property
    def resource_set(self):
        return Exchange.objects.filter(output__id=self.id, 
            input__process_type=1).extra(where=['input_id <> output_id']
            )

    @property
    def output_set(self):
        return Exchange.objects.filter(input__id=self.id).extra(where=[
            'input_id <> output_id'])

    def freeze(self):
        if self.process_type not in (self.TECHNOLOGICAL_PROCESS_TYPE,):
            raise UnfreezableProcess("This process type can't be frozen")
        if self.database.database_type not in (2,3,4):
            raise UnfreezableProcess("This project doesn't allow freezing")
        if Process.objects.filter(parent__id=self.id).exists():
            raise UnfreezableProcess("Can't freeze a process with children")
        from brightway.core.lca.lca import LCA
        lca = LCA(self.id)
        lca.lci()
        biosphere_vector = lca.calculated_biosphere.data.sum(axis=1)
        results = dict([(int(key), float(biosphere_vector[lca.dicts.biosphere[ 
            key], 0])) for key in lca.dicts.biosphere if \
            biosphere_vector[lca.dicts.biosphere[key], 0] != 0])
        self.process_type = self.FROZEN_TYPE
        self.frozen_data = results
        self.save()

    def thaw(self):
        self.frozen_data = None
        self.process_type = self.TECHNOLOGICAL_PROCESS_TYPE
        self.save()


attributes.register(Process)


class Exchange(ParameterModel):
    # Input group choices. Definitions from EcoSpold.
    MATERIALS_AND_FUELS = 1
    ELECTRICITY_AND_HEAT = 2
    SERVICES = 3
    FROM_NATURE = 4
    FROM_TECHNOSPHERE = 5

    INPUTGROUP_CHOICES = (
        (MATERIALS_AND_FUELS, _('materials or fuels')),
        (ELECTRICITY_AND_HEAT, _('electricity or heat')),
        (SERVICES, _('services')),
        (FROM_NATURE, _('from nature')),
        (FROM_TECHNOSPHERE, _('from technosphere')),
        )

    # Output group choices. Definitions from EcoSpold.
    REFERENCE_PRODUCT = 0
    INCLUDE_AVOIDED_PRODUCT_SYSTEM = 1
    ALLOCATED_BY_PRODUCT = 2
    WASTE_TO_TREATMENT = 3
    TO_NATURE = 4
    SELECTION = 5

    # TODO: Integrate into allocation model
    OUTPUTGROUP_CHOICES = (
        (REFERENCE_PRODUCT, _('reference product')),
        (INCLUDE_AVOIDED_PRODUCT_SYSTEM, _('include avoided product system')),
        (ALLOCATED_BY_PRODUCT, _('allocated by product')),
        (WASTE_TO_TREATMENT, _('waste to treatment')),
        (TO_NATURE, _('to nature')),
        (SELECTION, _('selection')), # This is new - not in Ecoinvent
        )

    ### Nodes ###
    input = models.ForeignKey(Process, related_name='exchange_input_set',
        verbose_name=_('input'))
    output = models.ForeignKey(Process, related_name='exchange_output_set',
        verbose_name=_('output'))

    ### Other meta-data ###
    input_group = models.IntegerField(_('input group'),
        choices=INPUTGROUP_CHOICES, null=True, blank=True)
    output_group = models.IntegerField(_('output group'),
        choices=OUTPUTGROUP_CHOICES, null=True, blank=True)
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True, 
        default='')

    objects = ExchangeManager()

    class Meta:
        verbose_name = _('exchange')
        verbose_name_plural = _('exchanges')
        # This is a good idea, but how to deal with what already exists, e.g.
        # 12 exchanges of "pesticide, unspecified" to "apples"?
        # Ideally these would be addressed by better inventory construction.
        # unique_together = ('input', 'output')

    def __unicode__(self):
        return u'%.3g %s to %s' % (self.amount, self.input, self.output)

    def save(self, skip_datetime_update=False, check_frozen_output=True,
            skip_database_dependency_check=False, *args, **kwargs):
        # if self.input_id == self.output_id:
        #     raise ValueError
        if not skip_database_dependency_check:
            # Create links between databases is needed
            if self.input.database_id != self.output.database_id:
                self.input.database.add_child(self.output.database)
        if check_frozen_output and self.output.process_type == \
                Process.FROZEN_TYPE:
            raise FrozenOutput
        super(Exchange, self).save(*args, **kwargs)
        if not skip_datetime_update:
            db = self.input.database
            db.modified = datetime.now()
            db.save()
            if not self.input.database_id == self.output.database_id:
                db = self.output.database
                db.modified = datetime.now()
                db.save()

    def unit(self):
        return self.input.unit
    unit.short_description = _('unit')

    def get_database(self):
        _input_db = self.input.database.database
        _output_db = self.output.database.database
        if _input_db == _output_db:
            return _input_db
        else:
            return _input_db + '/' + _output_db
    get_database.short_description = _('database')

attributes.register(Exchange)