# -*- coding: utf-8 -*-
from brightway.tests.lca_test_case import LCATestCase
from brightway.apps.inventory.models import Process
from brightway.apps.inventory.errors import UnfreezableProcess, FrozenOutput,\
    FrozenParent
from brightway.core.lca.lca import LCA

class FrozenProcessTestCase(LCATestCase):
    """Test freezing and thawing"""
    def setUp(self):
        super(FrozenProcessTestCase, self).setUp()
        self.fill_out_widget_process()

    def fill_out_widget_process(self):
        """Add exchanges to the widget process. Only it can be frozen."""
        self.create_exchange(self.processes.coal, self.processes.widget, 1)
        self.create_exchange(self.processes.steel, self.processes.widget, 1)
        self.create_exchange(self.processes.electricity.mix, 
            self.processes.widget, 1)

    def test_emission_type(self):
        """Test that emissions can't be frozen"""
        self.assertRaises(UnfreezableProcess, self.processes.nox.freeze)

    def test_wrong_database_type(self):
        """Test that on project processes can be frozen"""
        self.assertRaises(UnfreezableProcess, self.processes.steel.freeze)

    def test_has_children(self):
        """Test that freezing a process with children raises an error"""
        p = self.create_process("Test child", self.categories.products, 
            parent=self.processes.widget, database=self.databases.user)
        self.assertRaises(UnfreezableProcess, self.processes.widget.freeze)

    def test_add_exchange(self):
        """Test that adding an exchange to a frozen process raises and error"""
        self.processes.widget.freeze()
        p = self.create_process("Another process", self.categories.products, 
            database=self.databases.user)
        kwargs = {'output': self.processes.widget, 
            'input': p, 
            'amount': 1}
        self.assertRaises(FrozenOutput, self.create_exchange, **kwargs)

    def test_add_child(self):
        """Test that adding a child to a frozen process raises an error"""
        self.processes.widget.freeze()
        kwargs = {'name': "Another child", 
            'category': self.categories.products,
            'parent': Process.objects.get(id=self.processes.widget.id),
            'database': self.databases.user}
        self.assertRaises(FrozenParent, self.create_process, **kwargs)

    def test_freezing(self):
        """Test that freezing functions correctly"""
        self.processes.widget.freeze()
        widget = Process.objects.get(id=self.processes.widget.id)
        self.assertEqual(Process.FROZEN_TYPE, widget.process_type)
        self.assertTrue(isinstance(widget.frozen_data, dict))

    def test_thawing(self):
        """Test that thawing functions correctly"""
        num_exchanges = self.processes.widget.exchange_output_set.count()
        self.processes.widget.freeze()
        self.processes.widget.thaw()
        widget = Process.objects.get(id=self.processes.widget.id)
        self.assertEqual(Process.TECHNOLOGICAL_PROCESS_TYPE, 
            widget.process_type)
        self.assertFalse(widget.frozen_data)
        self.assertEqual(num_exchanges, widget.exchange_output_set.count())

    def test_lca_results(self):
        """Test that freezing a process doesn't change LCA results"""
        # Inventory
        # IA
        L = LCA(self.processes.widget.id, method=self.methods.human_health, 
            read_from_cache=False, write_to_cache=False, 
            **self.redirect_kwargs)
        L.lci()
        L.lcia()
        ia_score = L.score
        inventory_score = L.calculated_biosphere.sum()
        self.processes.widget.freeze()
        L = LCA(self.processes.widget.id, method=self.methods.human_health, 
            read_from_cache=False, write_to_cache=False, 
            **self.redirect_kwargs)
        L.lci()
        L.lcia()
        self.assertAlmostEqual(L.score, ia_score, places=5)
        self.assertAlmostEqual(L.calculated_biosphere.sum(), inventory_score, 
            places=5)

    def test_lca_results_intermediate_frozen_process(self):
        """Test that linking to a frozen process doesn't change LCA results"""
        p = self.create_process("Another process", self.categories.products, 
            database=self.databases.user)
        self.create_exchange(self.processes.widget, p, 1)
        L = LCA(p.id, method=self.methods.human_health, 
            read_from_cache=False, write_to_cache=False, 
            **self.redirect_kwargs)
        L.lci()
        L.lcia()
        ia_score = L.score
        self.processes.widget.freeze()
        L = LCA(p.id, method=self.methods.human_health, 
            read_from_cache=False, write_to_cache=False, 
            **self.redirect_kwargs)
        L.lci()
        L.lcia()
        self.assertAlmostEqual(L.score, ia_score, places=5)
