# -*- coding: utf-8 -*-
from brightway.apps.inventory.models import Database, Exchange
from brightway.core.inheritance.queries import *
from brightway.core.inheritance.proxies import InheritedExchangeProxy
from brightway.tests.lca_test_case import LCATestCase

class NoInheritanceTestCase(LCATestCase):
    '''Make sure custom SQL works without inheritance.'''
    def test_database_exchange_qs(self):
        """Test the method that gets all exchanges for a database."""
        self.assertEquals(Exchange.objects.all().count(), 20)
        self.assertEquals(len(recursive_exchanges_by_database( 
            Database.objects.all().values_list('id', flat=True), 
            as_dict=False)), 20)

class InheritanceTestCase(LCATestCase):
    '''Test process-exchange inheritance functionality.'''
    def setUp(self):
        super(InheritanceTestCase, self).setUp()
        """Child processes"""
        self.fusion = self.create_process("Fusion", 
            self.categories.electricity, self.geographies.world, 
            self.units.kwh, parent=self.processes.electricity.wind)
        self.fusion.attributes['dirty'] = False
        # Delete PM emission
        self.create_exchange(self.processes.pm, self.fusion, 0, 
            uncertainty_type=0)
        # Add input of electricity from wind
        self.create_exchange(self.processes.electricity.wind, self.fusion, 
            0.1)
        # Change input of electricity from mix
        self.create_exchange(self.processes.electricity.mix, self.fusion, 0.7)

    def test_exchange_numbers(self):
        """Test that the correct number of exchanges are found."""
        self.assertEquals(Exchange.objects.all().count(), 23)

    def test_unmodified_exchange(self):
        """Test that an untouched exchange is retrieved correctly."""
        e = recursive_exchange(self.processes.steel, self.fusion)
        self.assertEqual(e.amount, 0.02)

    def test_modified_exchange(self):
        """Test that an exchange modified in a child process is correct."""
        e = recursive_exchange(self.processes.electricity.mix, self.fusion)
        self.assertEqual(e.amount, 0.7)

    def test_added_exchange(self):
        """Test that an exchange added to a child process is found."""
        e = recursive_exchange(self.processes.electricity.wind, self.fusion)
        self.assertEqual(e.amount, 0.1)

    def test_zeroed_exchange(self):
        """Test that zeroed exchanges are returned, and have an amount of 0."""
        e = recursive_exchange(self.processes.pm, self.fusion)
        self.assertEqual(e.amount, 0)

    def test_process_exchange_qs(self):
        """Test the method that gets all exchanges from a process, including inherited exchanges."""
        self.assertEqual(4, len(list(recursive_exchanges_by_process( 
            self.fusion))))

    def test_database_exchange_qs(self):
        """Test the method that gets all exchanges for a database."""
        self.assertEquals(len(recursive_exchanges_by_database( 
            Database.objects.all().values_list('id', flat=True), 
            as_dict=False)), 24)

    def test_parent_process(self):
        """Test that a parent process is not affected by child modifications."""
        p = self.processes.electricity.wind
        e = recursive_exchange(self.processes.pm, p)
        self.assertEqual(e.amount, -0.01)
        e = recursive_exchange(self.processes.steel, p)
        self.assertEqual(e.amount, 0.02)
        e = recursive_exchange(self.processes.electricity.mix, p)
        self.assertEqual(e.amount, 0.01)
        self.assertEqual(3, len(list(recursive_exchanges_by_process(p))))

    def test_exchange_hierarchy(self):
        """Test than an entire hierarchy of exchanges functions correctly."""
        self.child = self.create_process("Child",
            self.categories.electricity, self.geographies.world, 
            self.units.kwh, parent=self.fusion)
        self.assertEquals(len(list(recursive_exchanges_by_database( 
            Database.objects.all().values_list('id', flat=True), 
            as_dict=False))), 28)
        self.assertEquals(len(list(recursive_exchanges_by_database( 
            Database.objects.all()))), 28)
        self.assertEqual(4, len(list(recursive_exchanges_by_process( 
            self.child))))

    def test_multiple_child_processes(self):
        """Test that multiple child processes of a single parent function correctly."""
        self.child = self.create_process("Child", 
            self.categories.electricity, self.geographies.world, 
            self.units.kwh, parent=self.processes.electricity.wind)
        self.assertEquals(len(recursive_exchanges_by_database( 
            Database.objects.all().values_list('id', flat=True), 
            as_dict=False)), 27)
        self.assertEqual(3, len(list(recursive_exchanges_by_process( 
            self.child))))

    def test_independent_child_processes(self):
        """Test that multiple, independent (i.e. different parent) processes function correctly."""
        self.child = self.create_process("Child", 
            self.categories.metals, self.geographies.world, 
            self.units.kg, parent=self.processes.steel)
        self.assertEquals(len(recursive_exchanges_by_database( 
            Database.objects.all().values_list('id', flat=True), 
            as_dict=False)), 28)
        self.assertEqual(4, len(list(recursive_exchanges_by_process( 
            self.child))))

    def test_inherited_description(self):
        # Inheritance of attributes is not yet supported
        pass

