from numpy import zeros, log, NaN
from django.utils.translation import ugettext_lazy as _
from django.db import models
from stats_toolkit.uncertainty_choices import uncertainty_choices
from stats_toolkit.distributions import LognormalUncertainty

class UncertaintyMixin(models.Model):
    """Mixin that provides uncertainty fields"""
    uncertainty_type = models.IntegerField(_('uncertainty type'),
        choices=uncertainty_choices.choices_tuple, default=0)

    amount = models.FloatField(_("mean or geometric mean"))
    sigma = models.FloatField(_('standard deviation'), default=NaN)
    minimum = models.FloatField(_('miniumum'), default=NaN)
    maximum = models.FloatField(_('maximum'), default=NaN)

    # Include extra parameters for future distributions
    uncertainty_param_one = models.FloatField(default=NaN)
    uncertainty_param_two = models.FloatField(default=NaN)
    uncertainty_param_three = models.FloatField(default=NaN)
    uncertainty_param_four = models.FloatField(default=NaN)
    
    # TODO: Add default values for Bernoulli
    # def save(self, *args, **kwargs):
    #     if not self.amount:
    #     if self.uncertainty_type in (UNCERTAINTY_TRIANGULAR, 
    #             UNCERTAINTY_UNIFORM):
    #         self.amount = (self.minimum + self.maximum)/2
    #     else:
    #         raise ValueError, "Must specify an amount"
    #     super(UncertaintyMixin, self).save(*args, **kwargs)

    class Meta:
        abstract = True

    @property
    def uncertainty(self):
        return uncertainty_choices.id_dict[self.uncertainty_type]

    def params(self):
        if hasattr(self, "input") and hasattr(self, "output"):
            model_type = "exchange"
        else:
            model_type = "generic"
        params = zeros((1,), dtype=[('input', 'u4'), ('output', 'u4'), 
            ('amount', 'f4'), ('negative', 'b1'), ('sigma', 'f4'), ('minimum', 
            'f4'), ('maximum', 'f4')])
        negative = self.amount < 0
        if self.uncertainty == LognormalUncertainty:
            amount = log(abs(self.amount))
        else:
            amount = self.amount

        if model_type == "exchange":
            params[:] = (self.input.id, self.output.id, amount, 
                negative, self.sigma, self.minimum, 
                self.maximum)
        else:
            params[:] = (self.id, 0, amount, 
                negative, self.sigma, self.minimum, 
                self.maximum)
        return params
