#!/usr/bin/python
# -*- coding: utf-8 -*
import math
from django.db import models
from django.db.models.signals import post_save
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.contenttypes.models import ContentType
from django.db import connection
from brightway.apps.base.fields.autopopulatemodel import AutopopulateModel
from brightway.apps.parameter.errors import ScopeError
from brightway.apps.parameter.parsing import Parser
from brightway.apps.parameter.context import ReferenceContext
from django_dag.mixins import GraphMixin
from django_dag.models import Graph, Node, Edge
from brightway.apps.parameter.models.mixins import UncertaintyMixin
from brightway.apps.parameter import functions

safe_list = ['math', 'functions']
SAFE_DICT = dict([ (k, locals().get(k, None)) for k in safe_list ])

REFERENCE_SQL = """
select uuid, id, amount, 'inventory_exchange' from inventory_exchange 
    where uuid in %s
union all
select uuid, id, amount, 'parameter_local_variable' from parameter_local_variable 
    where uuid in %s
union all 
select uuid, id, amount, 'parameter_global_variable'
    from parameter_global_variable 
    where uuid in %s
"""

LOCAL_VARIABLE_PROCESS_SQL = """
with recursive parents(p_id, parent_id, variable, v_id, amount, level) as (
    select p.id, p.parent_id, v.variable, v.id, v.amount, 1 
        from parameter_local_variable v
        left join inventory_process p
            on p.id = v.process_id
        where p.id = %s
    union all
    select p.id, p.parent_id, v.variable, v.id, v.amount, parents.level + 1
        from parents, parameter_local_variable v
        left join inventory_process p
            on p.id = v.process_id
        where p.id = parents.parent_id)
    select distinct on (variable) p_id, variable, v_id, amount, level
        from parents
        order by variable, level
"""

GLOBAL_VARIABLE_SQL = """
select variable, amount, id from parameter_global_variable 
    where database_id = %s
    and variable in %s
"""

class CachedContentTypeDict(object):
    def __init__(self):
        self._ct_dict = None

    @property
    def ct_dict(self):
        if not self._ct_dict:
            self._ct_dict = dict([("%s_%s" % (obj.app_label, 
                obj.name.replace(" ", "_")), obj.id) for obj in \
                ContentType.objects.all()])
        return self._ct_dict

    def __getitem__(self, key):
        return self.ct_dict[key]

CT_DICT = CachedContentTypeDict()

class CacheParameterGraph(object):
    def __init__(self):
        self._pg = None

    @property
    def graph(self):
        if not self._pg:
            try:
                self._pg = Graph.objects.get(name="parameter")
            except ObjectDoesNotExist:
                self._pg = Graph.objects.create(name="parameter")
        return self._pg

PARAMETER_GRAPH = CacheParameterGraph()

class ParameterModel(AutopopulateModel, UncertaintyMixin, GraphMixin):
    """Abstract base class that includes parameterization with formulas."""
    formula = models.TextField(_('formula'), null=True, blank=True, 
        help_text=_("Formulas override amount, if provided."))
    _parsed_formula = models.TextField(null=True, editable=False)

    class Meta:
        abstract = True

    def __float__(self):
        return self.amount

    @property
    def default_graph(self):
        return PARAMETER_GRAPH.graph

    def node(self, graph=None):
        # import pdb; pdb.set_trace()
        if not hasattr(self, "_node_cache"):
            # Call GraphMixin node finding procedure
            self._node_cache = super(ParameterModel, self).node( 
                graph=self.default_graph)
        return self._node_cache

    def save(self, reparse_formula=True, update_children=True, atomic=True, 
            *args, **kwargs):
        if not self.formula and self.amount is None:
            # Need to allow 0 as an amount, for overridinng exchanges with 
            # inheritance.
            raise AttributeError, "Must set either amount or formula"

        if atomic:
            self.start_transaction()

        try:
            if self.formula and reparse_formula:
                self.context = ReferenceContext(self)
                self.parse_formula()
                self.check_local_variable_scope()
                self.check_global_variable_scope()
                self.check_generic_reference_scope()
                self.amount = self.evaluate()

            super(ParameterModel, self).save(*args, **kwargs)

            if self.formula and reparse_formula:
                self.update_references()

            if self.formula and update_children:
                self.update_children()

        except:
            if atomic:
                self.rollback_transaction()
            raise

        if atomic:
            self.commit_transaction()

    def has_formula(self):
        return bool(self.formula)
    has_formula.boolean = True
    has_formula.short_description = _("formula")

    def update_references(self):
        existing_edges = Edge.objects.filter(graph=self.default_graph, 
            node_to=self.node()).values("node_from__content_type", 
            "node_from__object_id")
        formatted_edges = set(["ref:%i:%i" % obj["node_from__content_type", 
            "node_from__object_id"] for obj in existing_edges])
        this_node = self.node()
        this_graph = self.default_graph
        new_edges = set(self.context.keys())
        for obj in formatted_edges.difference(new_edges):
            dummy, ct_id, obj_id = obj.split(":")
            Edge.objects.filter(graph=this_graph, node_to=this_node,
                node_from__content_type=int(ct_id), 
                node_from__object_id=int(obj_id)).delete()

        for obj in new_edges.difference(formatted_edges):
            dummy, ct_id, obj_id = obj.split(":")
            try:
                node_from = Node.objects.get(graph=this_graph, 
                    content_type=ContentType.objects.get(id=int(ct_id)),
                    object_id=int(obj_id))
            except ObjectDoesNotExist:
                node_from = Node.objects.create(graph=this_graph, 
                    content_type=ContentType.objects.get(id=int(ct_id)),
                    object_id=int(obj_id))
            Edge.objects.create(graph=this_graph, node_to=this_node, 
                node_from=node_from)

    def update_children(self, atomic=True):
        if atomic:
            self.start_transaction()

        try:
            # Ordered by depth, don't need to call recursively
            for obj in self.get_descendants():
                obj.context = self.context
                try:
                    obj.amount = obj.evaluate()
                except KeyError:
                    for ref in obj.get_parents():
                        ct_id = ContentType.objects.get_for_model(ref).id
                        self.context.references["ref:%i:%i" % (ct_id, ref.id)] =\
                            {'ct_id': ct_id,
                            'id': ref.id,
                            'amount': ref.amount}
                    obj.amount = obj.evaluate()
                obj.save(reparse_formula=False, update_children=False, 
                    atomic=False)

        except:
            if atomic:
                self.rollback_transaction()
            raise

        if atomic:
            self.commit_transaction()

    def parse_formula(self):
        """Use pyparsing to parse EcoSpold2 formula syntax to python syntax."""
        self.parser = Parser(self.formula)
        self._parsed_formula = self.parser.translate()

    def get_process_id(self):
        raise NotImplementedError, "Must define for each ParameterModel subclass"

    def get_local_variables(self):
        cursor = connection.cursor()
        cursor.execute(LOCAL_VARIABLE_PROCESS_SQL, (self.get_process_id(),))
        return [{'process_id': row[0], 'variable': row[1], 'id': row[2], 
            'amount': row[3], 'level': row[4]} for row in cursor.fetchall()]

    def check_local_variable_scope(self):
        if not self.parser.local_variables:
            return

        existing = self.get_local_variables()

        missing = set(self.parser.local_variables).difference(
            set([x['variable'] for x in existing]))
        if missing:
            error_msg = "The following %i local variables not found:" % \
                len(missing) + "\n\t".join(missing)
            raise ScopeError, error_msg

        content_type_id = CT_DICT["parameter_local_variable"]

        matching = [x for x in existing if x["variable"] in \
            self.parser.local_variables]

        for obj in matching:
            obj["ct_id"] = content_type_id
            key = "ref:%i:%i" % (content_type_id, obj["id"])
            self._parsed_formula = self._parsed_formula.replace("local:%s" % \
                obj["variable"], "%%(%s)f" % key)
            self.context.references[key] = obj

    def get_database(self):
        try:
            return self.database
        except AttributeError:
            raise AttributeError, "Must define self.database for " + \
                "ParameterModel subclasses."

    def get_global_variables(self):
        if not hasattr(self, "database"):
            raise NotImplementedError
        cursor = connection.cursor()
        # Use raw SQL to avoid circular imports
        cursor.execute(GLOBAL_VARIABLE_SQL, (self.database.id, 
            tuple(self.parser.global_variables)))
        return [{'variable': row[0], 'amount': row[1], 'id': row[2]} for row \
            in cursor.fetchall()]

    def check_global_variable_scope(self):
        if not self.parser.global_variables:
            return

        existing = self.get_global_variables()

        missing = set(self.parser.global_variables).difference(
            set([x['variable'] for x in existing]))
        if missing:
            error_msg = "The following %i global variables not found:" % \
                len(missing) + "\n\t".join(missing)
            raise ScopeError, error_msg

        content_type_id = CT_DICT["parameter_global_variable"]

        matching = [x for x in existing if x["variable"] in \
            self.parser.global_variables]

        for obj in matching:
            obj["ct_id"] = content_type_id
            key = "ref:%i:%i" % (content_type_id, obj["id"])
            self._parsed_formula = self._parsed_formula.replace("global:%s" %\
                obj["variable"], "%%(%s)f" % key)
            self.context.references[key] = obj

    def get_generic_references(self):
        cursor = connection.cursor()
        cursor.execute(REFERENCE_SQL, (tuple(self.parser.references),))
        return [{'uuid': row[0], 'id': row[1], 'amount': row[2], 
            'ct_id': CT_DICT[row[3]]} for row in cursor.fetchall()]

    def check_generic_reference_scope(self):
        if not self.parser.references:
            return

        existing = self.get_generic_references()

        if len(existing) != len(self.parser.references):
            ex_vars = set([x['uuid'] for x in existing])
            missing_uuids = set(self.parser.references).difference(ex_vars)
            error_msg = "The following %i UUID references not found:" \
                % len(missing_uuids) + "\n\t".join(missing_uuids)
            raise ScopeError, error_msg

        for obj in existing:
            key = "ref:%i:%i" % (obj["ct_id"], obj["id"])
            self._parsed_formula = self._parsed_formula.replace("ref:%s" % \
                obj["uuid"], "%%(%s)f" % key)
            self.context.references[key] = obj

    def evaluate(self):
        if not self.formula:
            raise ValueError, "No formula to evaluate"

        if not self._parsed_formula and not self.context:
            self.parse_formula()

        return eval(self._parsed_formula % self.context, 
            {'__builtins__':None}, SAFE_DICT)

    ### Transactions ###
    def start_transaction(self):
        transaction.enter_transaction_management()
        transaction.managed(True)

    def rollback_transaction(self):
        transaction.rollback()

    def commit_transaction(self):
        transaction.commit()
