from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.db import IntegrityError
from brightway.apps.parameter.errors import IllegalLocalVariable, \
    DuplicateGlobalVariable
from brightway.apps.parameter.models import ParameterModel


class Variable(ParameterModel):
    """Abstract class for a node in a directed acyclic graph of parameters."""
    variable = models.TextField(_('name'))
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True, 
        default='')

    class Meta:
        abstract = True

    def __unicode__(self):
        return self.variable


class GlobalVariable(Variable):
    """Global variables can only exist once in a database. Their use is not 
allowed in EcoSpold2, and they are generally discouraged. However, they are 
sometimes useful for work with projects."""
    database = models.ForeignKey('inventory.Database', 
        verbose_name=_('database'))

    class Meta:
        unique_together = ('database', 'variable')
        verbose_name = _('global variable')
        verbose_name_plural = _('global variables')
        app_label = "parameter"
        db_table = 'parameter_global_variable'

    def check_local_variable_scope(self):
        if self.parser.local_variables:
            raise IllegalLocalVariable, "Global variables can't reference " +\
                "local variables; parsed the following local variables:\n\t"+\
                "\n\t".join(self.parser.local_variables)

    def save(self, *args, **kwargs):
        try:
            super(GlobalVariable, self).save(*args, **kwargs)
        except IntegrityError:
            raise DuplicateGlobalVariable, "A global variable named " + \
                self.variable + " already exists in this database."


class LocalVariable(Variable):
    """Local variables are specific to a technological process tree. There can 
be multiple local variables with the same name in a process tree."""
    process = models.ForeignKey("inventory.Process", 
        verbose_name=_("process reference"), null=False)

    def get_process_id(self):
        return self.process.id

    class Meta:
        verbose_name = _('local variable')
        verbose_name_plural = _('local variables')
        app_label = "parameter"
        db_table = 'parameter_local_variable'

