#~/usr/bin/python

"""
A parsing grammar that interprets formulas from the EcoSpold2 data format.
Based partially on example from the pyparsing wiki.

Because we want to be able to dynamically substitute in variable values, we 
only need to parse to an evaluable python string, and provide the references and 
variables in the current context. Because we are only doing substitution and 
some basic manipulation, the code is much simpler than it would be if we were 
doing evaluation using pyparsing as well.

The parser then only checks 3 things:

 *. Anything not defined as a function is interpreted as a variable. Variables 
 checked to make sure they exist in the current scope, and are reformatted for 
 string substitution.
 *. References are reformatted for string substitution, and are checked to 
 ensure they are in the current scope.
 *. Parentheses are checked to make sure they are balanced.

Functions are substituted for Python function names as needed.

Supported functions
===================

=================== =================== ======================================================
OpenFormula syntax  Python equivalent   Comments
=================== =================== ======================================================
ABS                 abs
ACOS                math.acos
AND                 and                 Python evaluates more values as True than OpenFormula
ASIN                math.asin
ATAN                math.atan
ATAN2               math.atan2
AVERAGE             numpy.average       Is this the best approach?
CHOOSE == ? 
COLUMNS (!?) 
COS 
COUNT (!?) 
COUNTA (!?) 
COUNTBLANK (!?) 
COUNTIF (!?) 
DATE 
DAVERAGE 
DAY 
DCOUNT 
DCOUNTA 
DDB 
DEGREES 
DGET 
DMAX 
DMIN 
DPRODUCT 
DSTDEV 
DSTDEVP 
DSUM 
DVAR 
DVARP 
EVEN 
EXACT 
EXP == math.exp 
FACT 
FALSE == False 
FIND (!?) 
FV 
HLOOKUP 
HOUR 
IF 
INDEX 
INT 
IRR 
ISBLANK 
ISERR 
ISERROR 
ISLOGICAL 
ISNA 
ISNONTEXT 
ISNUMBER 
ISTEXT 
LEFT 
LEN 
LN 
LOG 
LOG10 
LOWER 
MATCH 
MAX 
MID 
MIN 
MINUTE 
MOD 
MONTH 
N 
NA 
NOT 
NOW 
NPER 
NPV 
ODD 
OR 
PI 
PMT 
POWER 
PRODUCT 
PROPER 
PV 
RADIANS 
RATE 
REPLACE 
REPT 
RIGHT 
ROUND 
ROWS 
SECOND 
SIN 
SLN 
SQRT 
STDEV 
STDEVP 
SUBSTITUTE 
SUM 
SUMIF 
SYD 
T 
TAN 
TIME 
TODAY 
TRIM 
TRUE 
TRUNC 
UPPER 
VALUE 
VAR 
VARP 
VLOOKUP 
WEEKDAY 
YEAR

Functions that should be supported, but aren't yet:

None yet...

Numbers can be 1, 1.5, 1E+1, 1E-1, 1e1

Inequalities: =, <>, <=, >=, <, >

Weirdness: % == /100
"""
from primitives import parser
from pyparsing_mod import ParseResults
from errors import MisformedUUIDError

class Parser(object):
    def __init__(self, formula):
        self.formula = formula

    def unfold(self, element):
        if isinstance(element, ParseResults):
            return "(" + " ".join([self.unfold(obj) for obj in element]) + ")"
        else:
            return element

    def recursively_build_scope(self, element):
        if isinstance(element, ParseResults):
            self.md.update(element.asMultipleDict())
            for obj in element:
                self.recursively_build_scope(obj)
        else:
            return

    def translate(self):
        parsed = parser.parseString(self.formula)
        self.md = parsed.asMultipleDict()
        # Add nested ParsedResults dicts to self.md
        for obj in parsed:
            self.recursively_build_scope(obj)

        self.local_variables = [x[6:] for x in self.md["locals"]] if "locals" \
            in self.md else ()
        if "REF" in self.local_variables:
            raise MisformedUUIDError, "An incomplete or malformed UUID was probably present"
        self.global_variables = [x[7:] for x in self.md["globals"]] if \
            "globals" in self.md else ()
        self.references = [x[4:] for x in self.md["references"]] if \
            "references" in self.md else ()

        # Simple passing wraps EVERYTHING in parentheses, which is icky.
        return " ".join([self.unfold(obj) for obj in parsed])
