# from django.test import TestCase as DjangoTestCase
from unittest import TestCase
from brightway.apps.parameter.primitives import parser
from brightway.apps.parameter.parsing import Parser
from brightway.apps.parameter.errors import ScopeError, MisformedUUIDError
from brightway.apps.parameter.pyparsing_mod import ParseException

class PrimitivesTestCase(TestCase):
    """Test primitives.py parsing fundamentals."""
    def return_parser_string(self, expression):
        return str(parser.parseString(expression))

    def test_sexp(self):
        simple = "( foo )"
        complicated = "( foo * (bar - 2 ) )"
        without_spaces = "(foo)"
        unbalanced = "(foo"
        self.assertEqual("[['local:foo']]", self.return_parser_string(simple))
        self.assertEqual("[['local:foo', '*', ['local:bar', '-', '2.0']]]", 
            self.return_parser_string(complicated))
        self.assertRaises(ParseException, parser.parseString, unbalanced)

    def test_numbers(self):
        self.assertEqual("['2.0']", self.return_parser_string("2"))
        self.assertEqual("['2.0']", self.return_parser_string("2.0000"))
        self.assertEqual("['200.0']", self.return_parser_string("2e2"))
        self.assertEqual("['200.0']", self.return_parser_string("2e+2"))
        self.assertEqual("['0.2']", self.return_parser_string("2e-1"))

    def test_variables(self):
        self.assertEqual("['local:foo']", self.return_parser_string("foo"))
        self.assertEqual("['local:foo', '*', 'local:bar']", 
            self.return_parser_string("foo * bar"))
        self.assertEqual("['global:foo']", self.return_parser_string("global:foo"))
        self.assertEqual("['local:foo', '*', 'global:bar']", 
            self.return_parser_string("foo * global:bar"))

    def test_asmultipledict(self):
        pass

    def test_uuid(self):
        self.assertEqual("['ref:39435fa2-8b43-11de-8a52-0019e336be3a']", 
            self.return_parser_string("REF:<39435fa2-8b43-11de-8a52-" + \
            "0019e336be3a>"))
        # Not recognized; will raise an error in Parser
        self.assertEqual("['local:REF']", self.return_parser_string("REF:<" + \
            "39435fa2-8b43-11de-8a52-0019e336be3>"))

class ParserTestCase(TestCase):
    """Test Parser class"""
    def test_unfolding(self):
        p = Parser("a - ( + complicated * ( / expression ) * (with) - (some " + \
            "+ (parentheses)))")
        self.assertEqual("local:a - (+ local:complicated * (/ local:" + \
            "expression) * (local:with) - (local:some + (local:parentheses)))", 
            p.translate())
        p = Parser("ACOS((a + 1))")
        self.assertEqual("math.acos ((local:a + 1.0))", p.translate())
        p = Parser("ACOS(1) + 2")
        self.assertEqual("math.acos (1.0) + 2.0", p.translate())

    def test_local_variable_handling(self):
        p = Parser("a + 2")
        self.assertEqual("local:a + 2.0", p.translate())
        self.assertEqual(["a",], p.local_variables)
        self.assertFalse(p.global_variables)
        self.assertFalse(p.references)

    def test_functions_and_variables(self):
        # Checks against poor pyparsing usage where parseAction wouldn't 
        # return locals or globals back into the main ParseResults scope
        p = Parser("ACOS(a)")
        self.assertEqual("math.acos (local:a)", p.translate())
        self.assertEqual(["a",], p.local_variables)
        self.assertFalse(p.global_variables)
        self.assertFalse(p.references)

    def test_global_variable_handling(self):
        p = Parser("global:a")
        # print p.translate()
        self.assertEqual("global:a", p.translate())
        self.assertEqual(["a",], p.global_variables)
        self.assertFalse(p.local_variables)
        self.assertFalse(p.references)

    def test_reference_handling(self):
        p = Parser("REF:<39435fa2-8b43-11de-8a52-0019e336be3a> + 2")
        self.assertEqual("ref:39435fa2-8b43-11de-8a52-0019e336be3a + 2.0", 
            p.translate())
        self.assertEqual(["39435fa2-8b43-11de-8a52-0019e336be3a",], 
            p.references)
        self.assertFalse(p.global_variables)
        self.assertFalse(p.local_variables)

    def test_malformed_uuid(self):
        p = Parser("REF:<39435fa2-8b43-11de-8a52-0019e336be3>")
        self.assertRaises(MisformedUUIDError, p.translate)
        p = Parser("REF:<39435fa28b4311de8a520019e336be3a>")
        self.assertRaises(MisformedUUIDError, p.translate)
        