# -*- coding: utf-8 -*
from django.db import models
from django.utils.translation import ugettext_lazy as _
from brightway.apps.base.fields.automaticurlmodel import AutomaticURLModel


class UnitType(AutomaticURLModel):
    '''
Fundamental unit types (e.g. mass). Used primarily to determine which
conversions are allowed.
    '''
    unittype = models.TextField(_('unit type'),)
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True, 
        default='')

    class Meta:
        verbose_name_plural = _('unit types')
        verbose_name = _('unit type')

    def __unicode__(self):
        return self.unittype

    def unit_count(self):
        return self.unit_set.count()
    unit_count.short_description = _('unit count')

    def process_count(self):
        # Importing at the top causes circular import
        from brightway.apps.inventory.models import Process
        return Process.objects.filter(unit__in= \
            Unit.objects.filter(unittype__id=self.id)).count()
    process_count.short_description = _('process count')


class Unit(AutomaticURLModel):
    abbreviation = models.TextField(_('abbreviation'), blank=True, 
        default='')
    unit = models.TextField(_('name'), unique=True)
    unittype = models.ForeignKey(UnitType, verbose_name=_('unit type'), 
        related_name='unit_unittype_set')
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True, 
        default='')

    lookup_fields = ["unit"]

    class Meta:
        verbose_name = _('unit')
        verbose_name_plural = _('units')

    def __unicode__(self):
        return self.unit

    def process_count(self):
        return self.process_set.count()
    process_count.short_description = _('process count')

class UnitConversion(AutomaticURLModel):
    '''
Should perform automatic unit conversion when possible. 
Syntax: 1.0 input = amount of child.
    '''
    amount = models.FloatField(_('amount'))
    input = models.ForeignKey(Unit, related_name='unitconversion_input_set',
        verbose_name=_('input unit'))
    output = models.ForeignKey(Unit, related_name='unitconversion_output_set',
        verbose_name=_('output unit'))
    description = models.TextField(_('description'), blank=True, default='')
    other_info = models.TextField(_('other information'), blank=True,
        default='')

    class Meta:
        verbose_name = _('unit conversion')
        verbose_name_plural = _('unit conversions')
        unique_together = ('input', 'output')

    def __unicode__(self):
        return u'1 %s = %1.2e %s' % (self.input, self.amount, self.output)

    def reverse(self):
        return u'1 %s = %1.2e %s' % (self.output, 1.0/self.amount, 
            self.output)
    reverse.short_description = _('inverse')

    def consistent_unittypes(self):
        return self.input.unittype == self.output.unittype
    consistent_unittypes.boolean = True
    consistent_unittypes.short_description = _('consistent units')

