# -*- coding: utf-8 -*
import numpy
import numpy
import scipy.sparse
import unittest

class DifferentIndexTypesError(StandardError):
    pass


class InconsistentSlicingError(StandardError):
    pass


class ArrayProxy(object):
    """
Provides a dictionary-based interface from database row ids to array indices.

ArrayProxy provides a matrix or array whose indices are translated through a
lookup dictionary. Slices are not generally supported.
    """
    def __init__(self, data, row_dict, col_dict=None):
        super(ArrayProxy, self).__init__()
        self.data = data
        self.row_dict = row_dict
        if col_dict:
            self.col_dict = col_dict

    def translate(self, obj, type="row"):
        if hasattr(obj, "id"):
            obj = obj.id
        if isinstance(obj, slice):
            if obj == slice(None):
                return obj
            else:
                # Could support for slicing using Django's SortedDict
                # (http://code.djangoproject.com/browser/django/trunk/django/utils/datastructures.py)
                # or other sorted dictionary implementations, e.g. 
                # http://code.activestate.com/recipes/496761/ or
                # http://www.voidspace.org.uk/downloads/odict.py
                # However, this is not desperately needed...
                raise NotImplementedError, "Slices are not supported, "+\
                    "because their meaning is unclear when translated into "+\
                    "array indices. If you are confident, you can call "+\
                    "slices directly on self.data."
        elif isinstance(obj, (list, tuple)):
            return tuple([self.translate(_obj, type) for _obj in obj])
        else: # No longer test for integer keys; subclasses can use strings
            try:
                if type == "row":
                    return self.row_dict[obj]
                elif type == "col":
                    return self.col_dict[obj]
            except KeyError:
                raise KeyError, "Provided object %s is not a valid %s key"\
                    % (obj, type)

    def __getitem__(self, *args):
        assert len(args) == 1
        x = self.translate(args[0][0], "row")
        y = self.translate(args[0][1], "col")
        return self.data.__getitem__((x,y),)

    def __setitem__(self, *args):
        assert len(args) == 2
        x = self.translate(args[0][0], "row")
        y = self.translate(args[0][1], "col")
        return self.data.__setitem__((x,y), args[1])

    def __repr__(self):
        return '<%s for %s>' % (self.__class__.__name__, self.data.__repr__(
            )[1:-1])

    def __str__(self):
        return self.data.__str__()

    # Numpy functions
    def any(self, *args, **kwargs):
        return self.data.any(*args, **kwargs)

    def all(self, *args, **kwargs):
        return self.data.all(*args, **kwargs)

    def sum(self, *args, **kwargs):
        return self.data.sum(*args, **kwargs)

    def min(self, *args, **kwargs):
        # This should work for both ndarrays and sparse matrices
        try:
            m = self.data.min(*args, **kwargs)
        except AttributeError:
            m = self.data.data.min(*args, **kwargs)
        while isinstance(m, numpy.ndarray):
            m = m[0]
        return m

    def max(self, *args, **kwargs):
        try:
            m = self.data.max(*args, **kwargs)
        except AttributeError:
            m = self.data.data.max(*args, **kwargs)
        while isinstance(m, numpy.ndarray):
            m = m[0]
        return m

    def cumsum(self, *args, **kwargs):
        return self.data.cumsum(*args, **kwargs)

    def mean(self, *args, **kwargs):
        return self.data.mean(*args, **kwargs)

    def row(self, id):
        return self.row_dict[id]

    def col(self, id):
        return self.col_dict[id]

    @property
    def shape(self):
        return self.data.shape

    @property
    def row_dict_rev(self):
        """Build only upon demand"""
        if not hasattr(self, "_row_dict_rev"):
            self._row_dict_rev = dict(zip(self.row_dict.values(),
                self.row_dict.keys()))
        return self._row_dict_rev

    @property
    def col_dict_rev(self):
        if not hasattr(self, "_col_dict_rev"):
            self._col_dict_rev = dict(zip(self.col_dict.values(),
                self.col_dict.keys()))
        return self._col_dict_rev


class OneDimensionalArrayProxy(ArrayProxy):
    """
A special case of ArrayProxy limited to one-dimensional arrays.

Used for supply and demand arrays in LCA calculations.
    """
    def __init__(self, data, row_dict):
        if not len(data.shape) == 1:
            raise AttributeError, "Must only be used for one-dimensional array"
        super(OneDimensionalArrayProxy, self).__init__(data, row_dict)

    def __getitem__(self, *args):
        assert len(args) == 1
        x = self.translate(args[0], "row")
        return self.data.__getitem__((x,),)

    def __setitem__(self, *args):
        assert len(args) == 2
        x = self.translate(args[0], "row")
        self.data.__setitem__((x,), args[1])


class SparseMatrixProxy(ArrayProxy):
    """
Provides a dictionary-based interface from database row ids to array indices 
for sparse matrices. Does not assume a certain sparsity structure.
    """
    def __init__(self, data, row_dict, col_dict, *args, **kwargs):
        self.row_dict = row_dict
        self.col_dict = col_dict
        self.data = data
        try:
            self.format = self.data.getformat()
        except AttributeError:
            raise TypeError, "Must pass a Scipy sparse matrix"

    def __getitem__(self, *args):
        assert len(args) == 1
        x = self.translate(args[0][0], "row")
        y = self.translate(args[0][1], "col")

        if x == slice(None) or y == slice(None):
            raise NotImplementedError, "SparseMatrixProxy doesn't support " +\
                "slices; use CompressedSparseMatrixProxy for efficient slices"
        return self.data.__getitem__((x,y),)

    def __setitem__(self, *args):
        assert len(args) == 2
        x = self.translate(args[0][0], "row")
        y = self.translate(args[0][1], "col")

        if x == slice(None) or y == slice(None):
            raise NotImplementedError, "SparseMatrixProxy doesn't support " +\
                "assignment by slice"
        self.data.__setitem__((x,y), args[1])

    def get_row_as_dict(self, row):
        raise NotImplementedError, "Use CompressedSparseMatrixProxy for " + \
            "dictionary matrix slices"

    def get_col_as_dict(self, col):
        raise NotImplementedError, "Use CompressedSparseMatrixProxy for " + \
            "dictionary matrix slices"

    def toarray(self):
        return self.data.toarray()

    def todense(self):
        return self.data.todense()

    @property
    def nnz(self):
        # Used to use getnzmax, but that seems to be missing from newest scipy
        return self.data.getnnz()


class CompressedSparseMatrixProxy(SparseMatrixProxy):
    """
Subclass of SparseMatrixProxy for CSR (comma separated row) or CSC (comma 
separated column) matrices.

This class will create, on demand, the complementary matrix type (e.g. CSC for 
CSR) for efficient slicing and multiple assignment. If a matrix is changed, a 
*dirty* flag is set that tells the class to re-create the complementary matrix 
before accessing its information. All updates to complementary matrices are 
lazy.
    """

    __slots__ = ['row_dict', 'col_dict', 'format', 'dirty', '_csr', '_csc']

    def __init__(self, data, row_dict, col_dict, *args, **kwargs):
        self.row_dict = row_dict
        self.col_dict = col_dict
        self.dirty = True
        # try:
        self.format = data.getformat()
        if self.format == "csc":
            self._csc = data
        else:
            self._csr = data.tocsr()
            self.format = "csr"
        # except AttributeError:
        #     raise TypeError, "Must pass a Scipy sparse matrix"

    def __getitem__(self, *args):
        assert len(args) == 1
        x = self.translate(args[0][0], "row")
        y = self.translate(args[0][1], "col")

        if isinstance(x, slice) and isinstance(y, slice):
            raise NotImplementedError, \
                "Convert to dense matrix to do slices on multiple dimensions"
        elif isinstance(x, slice):
            if self.dirty and self.format == "csr":
                self._csc = self._csr.tocsc()
            return self._csc.__getitem__((x,y),)
        elif isinstance(y, slice):
            if self.dirty and self.format == "csc":
                self._csr = self._csc.tocsr()
            return self._csr.__getitem__((x,y),)
        else:
            return self.data.__getitem__((x,y),)

    def _get_data(self):
        if self.format == "csr":
            return self._csr
        else:
            return self._csc

    def _set_data(self, data):
        self.dirty = True
        if self.format == "csr":
            self._csr = data
        else:
            self._csc = data

    data = property(_get_data, _set_data)

    def __setitem__(self, *args):
        # Set dirty flag to indicate that matrix has changed, and need to
        # recompute the partner matrix if is accessed
        self.dirty = True
        # TODO: Warn if setting slicing on compressed sparse matrix?
        super(CompressedSparseMatrixProxy, self).__setitem__(*args)

    def get_row_as_dict(self, row):
        obj = self.__getitem__((row,slice(None)),)
        return dict([(self.row_dict_rev[obj.indices[index]], obj.data[index] 
            ) for index in xrange(obj.indices.shape[0])])

    def get_col_as_dict(self, col):
        obj = self.__getitem__((slice(None),col),)
        return dict([(self.col_dict_rev[obj.indices[index]], obj.data[index] 
            ) for index in xrange(obj.indices.shape[0])])


class ListArrayProxy(object):
    """
An interface to a list of objects that translates lookeups from foo[bar,baz] to 
foo.indices.index(bar)[baz]. If baz is a slice, returns a generator.

Takes list_, the list of objects, and optionally index_objs, which is an 
iterable of objects used as indices to list_. index_objs must all be of the 
same type.
    """

    __slots__ = ["indices", "list", "index_type"]

    def __init__(self, list_, index_objs=None):
        self.list = list_
        if index_objs:
            self.check_indice_types(index_objs)
            if len(index_objs) != len(list_):
                raise ValueError, "index_objs must have same length as list_"
            self.indices = list(index_objs)
        else:
            self.indices = None

    def check_indice_types(self, objs):
        l = [type(x) for x in objs]
        if not len(set(l)) == 1:
            raise DifferentIndexTypesError

    def sum(self):
        return sum([sum(obj) for obj in self.list])

    def __unicode__(self):
        if self.indices:
            return "<ListArrayProxy for %s>" % self.indices[0]
        else:
            return "<ListArrayProxy for unknown objects (id %s)>" % id(self)

    def translate_slice(self, sl):
        """Possible translate slice arguments into self.indices terms"""
        if sl.start in self.indices:
            start = self.indices.index(sl.start)
            start_translated = True
        else:
            start = sl.start
            start_translated = False
        if sl.stop in self.indices:
            stop = self.indices.index(sl.stop)
            stop_translated = True
        else:
            stop = sl.stop
            stop_translated = False
        if start_translated != stop_translated and start != None and stop != \
                None:
            raise InconsistentSlicingError, \
                "Only one slice element could be found in the indices"
        return slice(start, stop, sl.step)

    def __getitem__(self, args):
        if args == None:
            raise SyntaxError

        if isinstance(args, tuple):
            list_pos = args[0]
        else:
            list_pos = args

        if self.indices and isinstance(list_pos, slice):
            list_pos = self.translate_slice(list_pos)
        elif self.indices and list_pos in self.indices:
            list_pos = self.indices.index(list_pos)

        if not isinstance(args, tuple):
            return self.list[list_pos]
        elif isinstance(list_pos, slice):
            return (obj.__getitem__(*args[1:]) for obj in self.list[list_pos])
        else:
            return self.list[list_pos].__getitem__(*args[1:])

    def __setitem__(self):
        raise NotImplementedError

    def __iter__(self):
        return iter(self.list)

    def __len__(self):
        return len(self.list)

    def __repr__(self):
        return self.__unicode__()


class ArrayProxyTest(unittest.TestCase):
    def test_array_proxy(self):
        a = numpy.zeros((3,3))
        for x in range(3):
            for y in range(3):
                a[x,y] = x+3*y
        a = ArrayProxy(a, {10: 0, 11: 1, 12: 2}, {20: 0, 21: 1, 22: 2})
        self.assertEqual(a[10, 20], 0)
        self.assertTrue(numpy.allclose(a[11, (20, 21)], numpy.array((1,4))))

    def test_array_proxy_object(self):
        class Dummy(object):
            def __init__(self, id):
                self.id = id

        a = numpy.zeros((3,3))
        for x in range(3):
            for y in range(3):
                a[x,y] = x+3*y
        a = ArrayProxy(a, {10: 0, 11: 1, 12: 2}, {20: 0, 21: 1, 22: 2})
        obj = Dummy(10)
        self.assertEqual(a[obj, 20], 0)

    def test_array_proxy_key_error(self):
        a = numpy.zeros((3,3))
        for x in range(3):
            for y in range(3):
                a[x,y] = x+3*y
        a = ArrayProxy(a, {10: 0, 11: 1, 12: 2}, {20: 0, 21: 1, 22: 2})
        self.assertRaises(KeyError, a.__getitem__, (20, 10))
        class Vanilla(object):
            pass
        v = Vanilla()
        self.assertRaises(KeyError, a.__getitem__, (v, 20))

    def test_array_proxy_slices(self):
        a = numpy.zeros((3,3))
        for x in range(3):
            for y in range(3):
                a[x,y] = x+3*y
        a = ArrayProxy(a, {10: 0, 11: 1, 12: 2}, {20: 0, 21: 1, 22: 2})
        self.assertRaises(NotImplementedError, a.__getitem__, (slice(None, 11,
            None),20))
        # Note: numpy slices don't preserve shape!
        self.assertTrue(numpy.allclose(a[:,21], numpy.array((3, 4, 5,))))
        self.assertTrue(numpy.allclose(a[11, :], numpy.array((1,4,7))))

    def test_reverse_dict(self):
        a = numpy.zeros((3,3))
        for x in range(3):
            for y in range(3):
                a[x,y] = x+3*y
        a = ArrayProxy(a, {10: 0, 11: 1, 12: 2}, {20: 0, 21: 1, 22: 2})
        self.assertFalse(hasattr(a, "_row_dict_rev"))
        self.assertFalse(hasattr(a, "_col_dict_rev"))
        self.assertTrue(a.row_dict_rev)
        self.assertTrue(a.col_dict_rev)
        self.assertEquals(a.row_dict_rev[0], 10)
        self.assertEquals(a.col_dict_rev[0], 20)

    def test_one_dimensional_proxy(self):
        b = numpy.zeros((3,))
        for x in range(3):
            b[x] = x+3
        b = OneDimensionalArrayProxy(b, {10: 0, 11: 1, 12: 2})
        self.assertEqual(b[11], 4)
        self.assertTrue(numpy.allclose(b[:], numpy.array((3, 4, 5))))
        b[11] = 13
        self.assertEqual(b[11], 13)
        self.assertTrue(numpy.allclose(b[:], numpy.array((3, 13, 5))))
        b = numpy.zeros((3,3))
        self.assertRaises(AttributeError, OneDimensionalArrayProxy, b, {})
        self.assertRaises(TypeError, OneDimensionalArrayProxy, b, {}, {})

    def test_sparse_matrix_proxy(self):
        mat = scipy.sparse.lil_matrix((3,3))
        for x in range(3):
            for y in range(3):
                if x == y: continue
                mat[x,y] = x+3*y
        mat = SparseMatrixProxy(mat, {10: 0, 11: 1, 12: 2}, {20: 0, 21: 1, 
            22: 2})
        self.assertEqual(mat[10, 20], 0)
        self.assertEqual(mat[11, 22], 7)
        self.assertEqual(mat.nnz, 6)
        mat[11,21] = 3
        self.assertEqual(mat[11,21], 3)
        self.assertRaises(NotImplementedError, mat.__getitem__, (slice(None), 
            21))
        self.assertRaises(NotImplementedError, mat.__setitem__, (slice(None), 
            21), 1)

    def test_compressed_sparse_matrix_proxy(self):
        c = scipy.sparse.lil_matrix((3,3))
        for x in range(3):
            for y in range(3):
                if x == y: continue
                c[x,y] = x+3*y
        c = CompressedSparseMatrixProxy(c, {10: 0, 11: 1, 12: 2}, 
            {20: 0, 21: 1, 22: 2})
        self.assertEqual(c[10, 20], 0)
        self.assertEqual(c[11, 22], 7)
        self.assertEqual(c.nnz, 6)
        self.assertTrue(isinstance(c.data, scipy.sparse.csr.csr_matrix))
        self.assertTrue(numpy.allclose(c[:,21].todense(), numpy.array(((3,),(
            0,),(5,)))))
        self.assertTrue(numpy.allclose(c[11,:].todense(), numpy.array((1,0,7))))
        c[11,21] = 3
        assert c[11,21] == 3


class ListArrayProxyTest(unittest.TestCase):
    def test_list_array_proxy_without_indices(self):
        l = ListArrayProxy((range(10), range(10)))
        self.assertEqual(l[0], range(10))
        self.assertEqual(list(l[:,1]), [1, 1])
        self.assertEqual(list(l[:,:]), [range(10), range(10)])
        self.assertEqual(list(l[:,4:6]), [range(10)[4:6], range(10)[4:6]])

    def test_objects_as_indices(self):
        class A(object):
            pass
        m = A()
        n = A()
        l = ListArrayProxy((range(5), range(5,10)), [m, n])
        self.assertEqual(l[n], range(5, 10))
        self.assertEqual(l.sum(), sum(range(10)))
        self.assertEqual(l[n,:2], [5, 6])

    def test_mismatched_index_length(self):
        self.assertRaises(ValueError, ListArrayProxy, range(5), range(4))

    def test_list_array_proxy_with_indices(self):
        class A(object):
            pass
        class B(object):
            pass
        m = A()
        o = B()
        self.assertRaises(DifferentIndexTypesError, ListArrayProxy, ((), ()), 
            [m, o])

    def test_slices_with_indices(self):
        l = ListArrayProxy((range(3), range(5), range(7)), 
            (3, 5, 7))
        self.assertEqual(l[:], (range(3), range(5), range(7)))
        self.assertRaises(InconsistentSlicingError, l.__getitem__, slice(3,4))
        self.assertFalse(l[:3])
        self.assertEqual(l[:5], (range(3),))
        self.assertEqual([x for x in l[::2]], [range(3), range(7)])
        self.assertEqual([x for x in l[5:]], [range(5), range(7)])

