import os
import cPickle
import numpy
import scipy.sparse
import tarfile
import tempfile
import random
import logging
import datetime
import time
import unittest
from django.conf import settings
from brightway.core.logging_decorators import timelogger, catch_ioerror
from brightway.core.utils.file_utils import check_if_expired
from brightway.core.errors import CacheDirectoryError

class CacheInterface(object):
    """
.. centered:: An easy interface for caching SciPy, Numpy, and Python objects.

Takes an object or a list of objects as an input if saving, and returns a list
of objects if retrieving. Most python objects are pickled, but Numpy and Scipy
matrices and arrays get special treatment. For Numpy arrays, the shape and 
datatype are written to a special meta file; for Scipy sparse matrices (only 
CSR and CSC matrices are supported) the 'data', 'indices', and 'indptr' 
attributes are stored (they are NumPy arrays themselves).

**Internals**

To store some objects where meta information is necessary to reconstruct them 
efficiently (e.g. SciPy objects), a tar file is created for instance 
of CacheInterface used. Decompression is done into a temporary directory, 
which is then deleted.

**Inputs**

    #. cache_dir: The directory to store the cache.
    #. filename: The filename of the cache to save or load.
    #. objs: The objects to cache. None if loading the cache.
    #. action: The cache is saved or loaded if action is True.
    #. logging: A boolean or logging config dictionary.
    #. overwrite: Overwrite filename is it exists already.
    #. expiration: Don't load a cache older than expiration. Optional.

**Outputs**

When saving objects, CacheInterface does not return a value. When retrieving 
objects, CacheInterface returns False is loading was unsuccesful, or a list of 
objects if loading was successful.
    """
    def __init__(self, cache_dir, filename, objs=None, action=True,
            logging=False, overwrite=True, expiration=False):
        super(CacheInterface, self).__init__()
        self.cache_dir = cache_dir
        self.check_cache_dir()

        self.filename = filename
        self.tar_filename = self.filename + ".tar"

        # Attempting to iterate sometimes produces incorrect results, e.g.
        # for dictionaries, strings, sparse matrices (you get each row or col. 
        # separately). CacheInterface always returns a list anyway.
        if objs is not None and not isinstance(objs, (list, tuple)):
            self.objs = (objs,)
        else:
            self.objs = objs

        self.overwrite = overwrite
        if expiration and not isinstance(expiration, (datetime.datetime,
                datetime.timedelta)):
            raise TypeError, "expiration must be a python datetime." + \
                "datetime. You passed %s (type %s)" % (expiration, 
                type(expiration))
        self.expiration = expiration
        self.meta = []
        self.logging, self.logger = self.setup_logging(logging)
        if self.logging:
            self.logger.debug(str(
                "Calling CacheInterface with parameters:\n"+\
                "\t\tcache_dir: %s\n\t\tfilename: %s\n\t\tobjs: %s\n\t\t"+\
                "action: %s\n\t\toverwrite: %s\n\t\tlogging: True") % (
                self.cache_dir, self.filename, repr(self.objs), action, 
                self.overwrite))

        if action:
            self.make_directory()

            if objs is not None:
                self.success = self.save()
            else:
                self.success = self.load()

            if not self.success:
                self.force_cleanup()
            else:
                self.cleanup()

    def flatten(self, l, ltypes=(list, tuple)):
        """From http://code.activestate.com/recipes/363051/
Via http://rightfootin.blogspot.com/2006/09/more-on-python-flatten.html"""
        if self.logging:
            self.logger.debug("Calling flatten with list: %s" % l)
        i = 0
        while i < len(l):
            while isinstance(l[i], ltypes):
                if not l[i]:
                    l.pop(i)
                    if not len(l):
                        break
                else:
                    l[i:i+1] = list(l[i])
            i += 1
        return l

    def check_cache_dir(self):
        # This can be tricky with Vista, e.g. 
        # http://stackoverflow.com/questions/450210/how-to-check-if-a-file-can-be-created-inside-given-directory-on-ms-xp-vista
        if not os.path.exists(self.cache_dir):
            raise CacheDirectoryError, "Cache directory doesn't exist"
        if not os.access(self.cache_dir, os.W_OK):
            raise CacheDirectoryError, "Can't write to given cache dir.: %s"%\
                self.cache_dir

    def cleanup(self):
        """Delete all files listed in self.meta, then remove self.temp_dir."""
        try:
            files = self.flatten([x[2] for x in self.meta])
            for file_ in files:
                os.remove(os.path.join(self.temp_dir, file_))
            os.remove(os.path.join(self.temp_dir, 'tar.meta'))
            os.rmdir(self.temp_dir)

        except OSError, inst:
            if self.logging:
                self.logger.critical("Unable to cleanup temporary directory"+\
                    " %s" % self.temp_dir)
                self.logger.critical(inst)
            raise OSError, inst

    def force_cleanup(self):
        """Delete all files in self.temp_dir"""
        try:
            for file_ in os.listdir(self.temp_dir):
                os.remove(os.path.join(self.temp_dir, file_))
            os.rmdir(self.temp_dir)
        except OSError, inst:
            if self.logging:
                self.logger.critical("Unable to cleanup temporary "+\
                    "directory %s" % self.temp_dir)
                self.logger.critical(inst)
            raise OSError, inst

    def setup_logging(self, l):
        if not l:
            return False, None
        elif l == True:
            return True, logging.getLogger('bw.cache')
        else: # l is a logging level
            try:
                level = int(l)
                logger = getLogger('bw.cache')
                logger.setLevel(l)
                return True, logger
            except:
                raise ValueError, \
                    "'logging' must be a logging level, e.g. logging.DEBUG"

    def make_directory(self):
        try:
            self.temp_dir = tempfile.mkdtemp()
        except OSError:
            if self.logging:
                self.logger.critical("Unable to create temporary directory")

    def add_files_to_tarfile(self):
        files = self.flatten([x[2] for x in self.meta])
        for file_ in files:
            self.tar_file.add(os.path.join(self.temp_dir, file_), file_)
            self.tar_file.add(os.path.join(self.temp_dir, 'tar.meta'),
                'tar.meta')

    @timelogger(logging.DEBUG)
    @catch_ioerror
    def write_array(self, obj, filename):
        if not isinstance(obj, numpy.ndarray):
            raise AttributeError, "CacheInterface.load_array obj must be a"+\
                " numpy array."
        with open(os.path.join(self.temp_dir, filename + '.array_data'), "wb"
                ) as f:
            numpy.save(f, obj)
        self.meta.append((filename,
            'array',
            (filename + '.array_data',)))

    @timelogger(logging.DEBUG)
    @catch_ioerror
    def write_sparse_matrix(self, obj, filename):
        if not isinstance(obj, (scipy.sparse.csr_matrix,
                scipy.sparse.csc_matrix)):
            raise AttributeError, "CacheInterface.write_sparse_matrix obj" + \
                " must be a CSC or CSR matrix."
        with open(os.path.join(self.temp_dir, filename + '.matrix_data'), 
                "wb") as f:
            numpy.save(f, obj.data)
        with open(os.path.join(self.temp_dir, filename + '.matrix_indices'), 
                "wb") as f:
            numpy.save(f, obj.indices)
        with open(os.path.join(self.temp_dir, filename + '.matrix_indptr'), 
                "wb") as f:
            numpy.save(f, obj.indptr)
        with open(os.path.join(self.temp_dir, filename + '.matrix_meta'), 
                "wb") as f:
            meta = {'shape': obj.shape, 'dtype': obj.dtype, 
                'indices_dtype': obj.indices.dtype, 
                'indptr_dtype': obj.indptr.dtype}
            if isinstance(obj, scipy.sparse.csr_matrix):
                meta['type'] = 'csr_matrix'
            else:
                meta['type'] = 'csc_matrix'
            cPickle.dump(meta, f)
        self.meta.append((filename,
            'sparse_matrix',
            (filename + '.matrix_data',
            filename + '.matrix_indices',
            filename + '.matrix_indptr',
            filename + '.matrix_meta',)))

    @timelogger(logging.DEBUG)
    @catch_ioerror
    def write_pickle(self, obj, filename):
        with open(os.path.join(self.temp_dir, filename + '.pickle'), "wb"
                ) as f:
            cPickle.dump(obj, f, cPickle.HIGHEST_PROTOCOL)
        self.meta.append((filename, 'pickle', (filename + '.pickle',)))

    def save_object(self, index, obj):
        if isinstance(obj, numpy.ndarray):
            filename = index + '.array'
            self.write_array(obj, filename)
        elif isinstance(obj, (scipy.sparse.csr_matrix,
                scipy.sparse.csc_matrix)):
            filename = index + '.sparse_matrix'
            self.write_sparse_matrix(obj, filename)
        # Explicit list of types to prevent non-standard objects from writing
        # huge files and killing server.
        elif isinstance(obj, (dict, list, tuple, str, unicode, int, float,
            set, datetime.datetime, datetime.timedelta, datetime.date,
            datetime.time)):
            filename = index + '.pickled'
            self.write_pickle(obj, filename)
        #TODO: Add other data sources, e.g. dense matrices
        else:
            raise TypeError, "Object of type %s cannot be cached." % type(obj)

    @timelogger(logging.INFO)
    def save(self):
        for index, obj in enumerate(self.objs):
            # Will write relevant info to self.meta
            self.save_object(str(index), obj)

        with open(os.path.join(self.temp_dir, "tar.meta"), "wb") as \
                self.meta_file:
            cPickle.dump(self.meta, self.meta_file)
        self.tar_file = tarfile.open(os.path.join(self.cache_dir,
            self.tar_filename), "w")
        self.add_files_to_tarfile()
        self.tar_file.close()

        # Add one second so that the cache tests pass - Unix rounds file 
        # creation to the nearest second, and tar files were seen as before
        # the database modification times.
        created_time = os.stat(os.path.join(self.cache_dir,
            self.tar_filename))[7]
        os.utime(os.path.join(self.cache_dir, self.tar_filename), 
            (created_time, created_time + 1))
        return True

    @timelogger(logging.DEBUG)
    @catch_ioerror
    def load_array(self, filename):
        with open(os.path.join(self.temp_dir, filename + '.array_data'), 'rb'
                ) as f:
            obj = numpy.load(f)
        return obj

    @timelogger(logging.DEBUG)
    @catch_ioerror
    def load_sparse_matrix(self, filename):
        with open(os.path.join(self.temp_dir, filename + '.matrix_meta'),
                "rb") as f:
            meta = cPickle.load(f)
        with open(os.path.join(self.temp_dir, filename + '.matrix_data'), 
                'rb') as f:
            data = numpy.load(f)
        with open(os.path.join(self.temp_dir, filename + '.matrix_indices'), 
                'rb') as f:
            indices = numpy.load(f)
        with open(os.path.join(self.temp_dir, filename + '.matrix_indptr'), 
                'rb') as f:
            indptr = numpy.load(f)

        if meta['type'] == 'csr_matrix':
            try:
                obj = scipy.sparse.csr_matrix((data,indices,indptr), 
                    shape=meta['shape'], dtype=meta['dtype'])
            except TypeError:
                # Scipy < 0.7
                obj = scipy.sparse.csr_matrix((data,indices,indptr), 
                    dims=meta['shape'], dtype=meta['dtype'])
        elif meta['type'] == 'csc_matrix':
            try:
                obj = scipy.sparse.csc_matrix((data,indices,indptr),
                    shape=meta['shape'], dtype=meta['dtype'])
            except TypeError:
                # Scipy < 0.7
                obj = scipy.sparse.csc_matrix((data,indices,indptr), 
                    dims=meta['shape'], dtype=meta['dtype'])
        return obj

    @timelogger(logging.DEBUG)
    @catch_ioerror
    def load_pickle(self, filename):
        f = open(os.path.join(self.temp_dir, filename + '.pickle'), 'rb')
        obj = cPickle.load(f)
        f.close()
        return obj

    def verify_tar_file_date(self):
        """
Return true is expiration and file data are valid, and expiration is later 
than last file modification.
        """
        if not self.expiration:
            # Always load if no expiration given
            return True
        if isinstance(self.expiration, datetime.timedelta):
            self.expiration = datetime.datetime.now() - self.expiration
        try:
            expired = check_if_expired(os.path.join(self.cache_dir, 
                self.tar_filename), self.expiration)
        except OSError:
            # If file doesn't exist, then return false. This keeps the temp
            # dir structure so that it can be written to later.
            return False
        if expired:
            return False
        else:
            return True

    @timelogger(logging.DEBUG)
    def load_tar_file(self):
        if not self.verify_tar_file_date():
            return False

        try:
            self.tar_file = tarfile.open(os.path.join(self.cache_dir,
                self.tar_filename), "r")
            self.tar_file.extractall(self.temp_dir)

            if self.logging:
                self.logger.info(str("Extracted tar file %s to %s. Tar "+\
                    "file contents: %s") % (self.tar_filename, self.temp_dir, 
                    os.listdir(self.temp_dir)))

            self.meta_file = open(os.path.join(self.temp_dir, "tar.meta"),
                "rb")
            self.meta = cPickle.load(self.meta_file)
            self.meta_file.close()

        except (IOError, tarfile.ReadError), inst:
            if self.logging:
                self.logger.error("Unable to load tarfile %s" % \
                    self.tar_filename)
                self.logger.error(inst)
            return False

        return True

    @timelogger(logging.INFO)
    def load(self):
            if not self.load_tar_file():
                return False

            self.objs = []

            for obj_name, obj_type, obj_files in self.meta:

                if self.logging:
                    self.logger.info("Loading file %s; type %s; files: %s" % \
                        (obj_name, obj_type, obj_files))

                try:
                    if obj_type == 'array':
                        self.objs.append(self.load_array(obj_name))
                    elif obj_type == 'sparse_matrix':
                        self.objs.append(self.load_sparse_matrix(obj_name))
                    elif obj_type == 'pickle':
                        self.objs.append(self.load_pickle(obj_name))

                except IOError, inst:
                    if self.logging:
                        self.logger.error("Unable to load file %s" % obj_name)
                        self.logger.error(inst)
                    return False

            return self.objs

    def expire(self):
        if self.temp_dir:
            self.force_cleanup()
        os.remove(os.path.join(self.cache_dir, self.tar_filename))


class CacheInterfaceTest(unittest.TestCase):
    def setUp(self):
        super(CacheInterfaceTest, self).setUp()
        self.cache_dir = os.path.join(settings.BRIGHTWAY_ROOT_PATH, 
            'core/tests/cache/')

    def tearDown(self):
        self.clean_cache_directory()
        super(CacheInterfaceTest, self).tearDown()

    def clean_cache_directory(self):
        def tar_filter(x):
            if x.lower()[-4:] == ".tar":
                return True
            return False

        files = os.listdir(self.cache_dir)
        tar_files = filter(tar_filter, files)
        for tf in tar_files:
            os.remove(os.path.join(self.cache_dir, tf))

    def test_return_false_from_missing_cache(self):
        result = CacheInterface(self.cache_dir, "missing")
        self.assertFalse(result.objs)

    def test_overwrite_existing_cache(self):
        l = ["a", 1.3, u"unicode"]
        t = (1, 2, 3)
        Input = CacheInterface(self.cache_dir, "test_overwrite", objs=l)
        Input = CacheInterface(self.cache_dir, "test_overwrite", objs=t)
        Output = CacheInterface(self.cache_dir, "test_overwrite")
        self.assertEqual(Output.objs, list(t))

    def test_numpy_array(self):
        a = numpy.ones((3,3))
        b = a * 10
        Input = CacheInterface(self.cache_dir, "test1", objs=a)
        Output = CacheInterface(self.cache_dir, "test1", None)
        self.assertTrue(numpy.allclose(a, Output.objs))
        Input = CacheInterface(self.cache_dir, "test2", objs=b)
        Output = CacheInterface(self.cache_dir, "test2", None)
        self.assertTrue(numpy.allclose(b, Output.objs))
        Input = CacheInterface(self.cache_dir, "test3", objs=(a,b))
        Output = CacheInterface(self.cache_dir, "test3", None)
        aa, bb = Output.objs
        self.assertTrue(numpy.allclose(a, aa))
        self.assertTrue(numpy.allclose(b, bb))

    def test_struct_array(self):
        g = numpy.zeros((5,), dtype=[
            ('foo', 'u4'), ('baloney', 'b1'), ('bar', 'f4')])
        g[0] = (12, True, 3.14)
        Input = CacheInterface(self.cache_dir, "test_struct", objs=g)
        Output = CacheInterface(self.cache_dir, "test_struct")
        h = Output.objs[0]
        self.assertEqual(g.shape, h.shape)
        self.assertTrue(numpy.allclose(g['foo'], h['foo']))
        self.assertTrue(numpy.allclose(g['baloney'], h['baloney']))
        self.assertTrue(numpy.allclose(g['bar'], h['bar']))

    def test_array_dtype(self):
        """Test array saving with non-default dtype"""
        a = numpy.ones((3,3), dtype='int16')
        Input = CacheInterface(self.cache_dir, "test1", objs=a)
        Output = CacheInterface(self.cache_dir, "test1", None)
        self.assertTrue(numpy.allclose(a, Output.objs[0]))

    def test_csc_matrix(self):
        b = scipy.sparse.lil_matrix((2,2))
        b[0,0] = 1
        b[0,1] = 2
        b[1,0] = 3
        b = b.tocsc()
        Input = CacheInterface(self.cache_dir, "test_csc", objs=b)
        Output = CacheInterface(self.cache_dir, "test_csc")
        c = Output.objs[0]
        self.assertTrue(numpy.allclose(b.todense(), c.todense()))
        self.assertEqual(b.shape, c.shape)
        self.assertTrue(isinstance(b, scipy.sparse.csc_matrix))

    def test_csr_matrix(self):
        b = scipy.sparse.lil_matrix((2,2))
        b[0,0] = 1
        b[0,1] = 2
        b[1,0] = 3
        b = b.tocsr()
        Input = CacheInterface(self.cache_dir, "test_csr", objs=b)
        Output = CacheInterface(self.cache_dir, "test_csr")
        c = Output.objs[0]
        self.assertTrue(numpy.allclose(b.todense(), c.todense()))
        self.assertEqual(b.shape, c.shape)
        self.assertTrue(isinstance(b, scipy.sparse.csr_matrix))

    def test_csr_dtype(self):
        """Test sparse matrix saving with non-default dtype"""
        b = scipy.sparse.lil_matrix((2,2), dtype='int8')
        b[0,0] = 1
        b[0,1] = 2
        b[1,0] = 3
        b = b.tocsr()
        Input = CacheInterface(self.cache_dir, "test_csr", objs=b)
        Output = CacheInterface(self.cache_dir, "test_csr")
        c = Output.objs[0]
        self.assertTrue(numpy.allclose(b.todense(), c.todense()))
        self.assertEqual(b.shape, c.shape)
        self.assertTrue(isinstance(b, scipy.sparse.csr_matrix))

        b = scipy.sparse.lil_matrix((2,2), dtype='f4')
        b[0,0] = 1
        b[0,1] = 2
        b[1,0] = 3
        b = b.tocsr()
        Input = CacheInterface(self.cache_dir, "test_csr", objs=b)
        Output = CacheInterface(self.cache_dir, "test_csr")
        c = Output.objs[0]
        self.assertTrue(numpy.allclose(b.todense(), c.todense()))
        self.assertEqual(b.shape, c.shape)
        self.assertTrue(isinstance(b, scipy.sparse.csr_matrix))

    def test_lil_matrix_fails(self):
        b = scipy.sparse.lil_matrix((2,2))
        b[0,0] = 1
        b[0,1] = 2
        b[1,0] = 3
        self.assertRaises(TypeError, CacheInterface, self.cache_dir, "this", 
            b)

    def test_arbitrary_object_fails(self):
        class Foo(object):
            """A dummy class used for tests"""
            def __init__(self, arg):
                self.arg = arg

        self.assertRaises(TypeError, CacheInterface, self.cache_dir, 
            "test_object", Foo("this"))

    def test_list_tuple(self):
        l = ["a", 1.3, u"unicode"]
        t = (1, 2, 3)
        Input = CacheInterface(self.cache_dir, "test_lt", objs=(l, t))
        Output = CacheInterface(self.cache_dir, "test_lt")
        self.assertEqual(Output.objs, [l, t])

    def test_nested_objects(self):
        a = ({1:2}, datetime.date(2009, 1, 1))
        b = ("monkey", 3.4)
        Input = CacheInterface(self.cache_dir, "test_nested", objs=(a, b))
        Output = CacheInterface(self.cache_dir, "test_nested")
        self.assertEqual(Output.objs, [a, b])

    def test_dict(self):
        dic = {1:2, "a":"b"}
        Input = CacheInterface(self.cache_dir, "test_dict", objs=dic)
        Output = CacheInterface(self.cache_dir, "test_dict")
        self.assertEqual(Output.objs[0], dic)

    def test_datetime(self):
        date = datetime.date(2009, 1, 1)
        time = datetime.time(1, 1, 1)
        dt = datetime.datetime(2009, 1, 1, 1, 1)
        Input = CacheInterface(self.cache_dir, "test_dt", objs=(date, time, 
            dt))
        Output = CacheInterface(self.cache_dir, "test_dt")
        self.assertEqual(Output.objs, [date, time, dt])

    def test_slice_fails(self):
        self.assertRaises(TypeError, CacheInterface, self.cache_dir, 
            "test_slice", slice(1, 5, 2))

    def test_set(self):
        s = set((1, 2, 3))
        Input = CacheInterface(self.cache_dir, "test_set", objs=s)
        Output = CacheInterface(self.cache_dir, "test_set")
        self.assertEqual(s, Output.objs[0])

    def test_invalid_path(self):
        self.assertRaises(CacheDirectoryError, CacheInterface, 
            "/foo/not/really", "false", ())

    def test_invalid_expiration(self):
        self.assertRaises(TypeError, CacheInterface.__init__, self.cache_dir,
            "nothing", {'objs': (), 'expiration': '2008-10-27'})
        self.assertRaises(TypeError, CacheInterface.__init__, self.cache_dir,
            "nothing", {'objs': (), 'expiration': (2010, 2, 16, 10, 6, 20,  
            107421)})

    def test_expiration_timedelta(self):
        saved = CacheInterface(self.cache_dir, "test_timedelta", objs={1:2})
        time.sleep(1) # The cache adds one second to file creation times; wait
        stale = CacheInterface(self.cache_dir, "test_timedelta", 
            expiration=datetime.timedelta(0, 0, 1)) # 1 microsecond)
        fresh = CacheInterface(self.cache_dir, "test_timedelta", 
            expiration=datetime.timedelta(0, 10)) # 1 second
        self.assertFalse(stale.objs)
        self.assertEqual(fresh.objs, [{1:2},])

    def test_expiration(self):
        saved = CacheInterface(self.cache_dir, "test_expiration", objs={1:2})
        # Expiration is date that database changed, not cache file
        can_load = CacheInterface(self.cache_dir, "test_expiration", 
            expiration=datetime.datetime.now() - datetime.timedelta(0, 2))
        no_expiration = CacheInterface(self.cache_dir, "test_expiration")
        stale = CacheInterface(self.cache_dir, "test_expiration", 
            expiration=datetime.datetime.now() + datetime.timedelta(0, 2))
        self.assertFalse(stale.objs)
        self.assertTrue(no_expiration.objs)
        self.assertTrue(can_load.objs)

    def test_corrupted_tarfile(self):
        outfile = open(os.path.join(self.cache_dir, "corrupted.tar"), "w")
        outfile.write("This is not a tar file\n")
        outfile.close()
        cache = CacheInterface(self.cache_dir, "corrupted")
        self.assertFalse(cache.objs)

    def test_manual(self):
        # action=False
        pass

    def test_overwrite_false(self):
        pass

