-- Create an in-memory table with the following columns that is recursively
-- populated as the tree is followed upwards.
with recursive exchange_recursive (
    -- The ID of the current level exchange (fixed to exchange)
    exchange_id, 
    -- The ID of the current process's parent (changes with recursion)
    intermediate_parent_id, 
    -- The ID of the exchange output (fixed to exchange)
    exchange_output_id,
    -- The ID of the exchange input (fixed to exchange)
    input_id, 
    -- The ID of the child process (doesn't change)
    output_id, 
    -- The exchange amount (fixed to exchange)
    amount, 
    -- The level of recursion (changes with recursion)
    level,
    -- The output process type (doesn't change)
    output_type,
    -- The input process type (fixed to exchange)
    input_type,
    -- Exchange uncertainty: Sigma (fixed to exchange)
    sigma,
    -- Exchange uncertainty: Type (fixed to exchange)
    uncertainty_type,
    -- Exchange uncertainty: Minimum (fixed to exchange)
    minimum,
    -- Exchange uncertainty: Maximum (fixed to exchange)
    maximum
    ) as (
    select e.id, starting_table.parent_id, e.output_id, e.input_id, starting_table.id, e.amount, 0, starting_table.process_type, input.process_type, e.sigma, e.uncertainty_type, e.minimum, e.maximum
        from inventory_process starting_table
        -- The key is a left join, as some rows may not have exchanges. We 
        -- need the empty row with the parent id for the recursive part, so
        -- we can pull in the exchanges of the parent.
        left join inventory_exchange e
            on starting_table.id = e.output_id
        -- Need to filter by the exchange input as well
        left join inventory_process input
            on input.id = e.input_id
        where starting_table.id = %s
        -- and (e.input_id =  or e.input_id is null)
    union all
    -- This is the recursive part; the first section is the initial rows.
    -- These are the recursive rows. Level is the level up the tree.
    -- Use exchange_recursive.output_id instead of e.output_id because we
    -- need to keep track of the original output_id to do the distinct clause
    -- below. 
    select e.id, recursed_table.parent_id, e.output_id, e.input_id, exchange_recursive.output_id, e.amount, exchange_recursive.level + 1, exchange_recursive.output_type, input.process_type, e.sigma, e.uncertainty_type, e.minimum, e.maximum
        from exchange_recursive, inventory_process recursed_table
        left join inventory_exchange e
            on recursed_table.id = e.output_id
        left join inventory_process input
            on input.id = e.input_id
        -- Recursive link by parent id
        where recursed_table.id = exchange_recursive.intermediate_parent_id
        -- This was causing a problem for parent - child - grandchild 
        -- relationships, because I think there were exchanges found for the
        -- child upon recursion, but as they did not have the correct ID they
        -- where not selected. Because there were exchanges found, there was
        -- not a input_id is null row upon which to recurse up to the top.
        -- and (e.input_id =  or e.input_id is null)
    )
select distinct on (input_id, output_id) * 
    from exchange_recursive
    where input_id = %s
    -- and input_id is not null
    and output_id is not null
    -- Ascending order so it picks lowest level first
    order by input_id, output_id, level asc
    limit 1