from django.db.models.query import QuerySet
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from brightway.core.sql import sql, dict_cursor
from brightway.apps.geo import WORLD

def recursive_exchanges_by_database(dbs, using="default", as_dict=True, 
        count=True):
    """
Use a recursive query to get all inherited exchanges for a set of database ids.

This can't return a queryset, or Exchange objects, as many of the objects returned here do no correspond to a database row.

.. rubric:: Inputs

* dbs: An list or set of database ids.
* using: String. Database connection label.
* as_dict: Boolean. Return results as a list of dictionaries instead of tuples.

.. rubric:: Outputs

A list of dictionaries, or tuples, with the following elements. See the inheritance documentation for a graphical illustration of each term.

#. exchange_id : The id of the inherited exchange
#. intermediate_parent_id : The id of the parent process providing the exchange
#. exchange_output_id : The id of the inherited exchange's output
#. input_id : The id of the input process
#. output_id : The id of the output process
#. amount : The floating-point amount
#. level : The numeric tree depth, going upwards from 0 as the child process
#. output_type : The numeric type of the output process
#. input_type : The numeric type of the input process
#. sigma : The sigma of the uncertainty distribution
#. uncertainty_type : The uncertainty type of the uncertainty distribution
#. minimum : The minimum of the uncertainty distribution
#. maximum : The maximum of the uncertainty distribution
    """
    if isinstance(dbs, int):
        d = (dbs,)
    elif isinstance(dbs, QuerySet):
        d = tuple(dbs.values_list('id', flat=True))
    elif hasattr(dbs, "id"):
        d = (dbs.id,)
    else:
        d = tuple(dbs)
    cursor = sql("core/inheritance/exchanges_for_databases.sql", (d,d), 
        using=using, as_dict=as_dict)
    if as_dict:
        # Return generator
        return cursor
    else:
        return cursor.fetchall()

def recursive_exchanges_count_by_database(dbs, using="default", as_dict=True, 
        count=True):
    """
Use a recursive query to count all inherited exchanges for a set of database ids.

.. rubric:: Inputs

* dbs: An list or set of database ids.
* using: String. Database connection label.

.. rubric:: Outputs

Row count. Integer.
    """
    if isinstance(dbs, int):
        d = (dbs,)
    elif isinstance(dbs, QuerySet):
        d = tuple(dbs.values_list('id', flat=True))
    elif hasattr(dbs, "id"):
        d = (dbs.id,)
    else:
        d = tuple(dbs)
    cursor = sql("core/inheritance/count_for_databases.sql", (d,d), 
        using=using)
    return cursor.fetchone()[0]

def recursive_exchanges_by_process(process, using="default", as_dict=True, 
        add_links=True):
    """
Use a recursive query to get all inherited exchanges for a process.

This can't return a queryset, or Exchange objects, as many of the objects returned here do no correspond to a database row.

.. rubric:: Inputs

* process: A process id or Process object.
* using: String. Database connection label.
* as_dict: Boolean. Return results as a list of dictionaries instead of tuples.

.. rubric:: Outputs

A list of dictionaries, or tuples, with the following elements. See the inheritance documentation for a graphical illustration of each term.

#. exchange_id : The id of the inherited exchange
#. intermediate_parent_id : The id of the parent process providing the exchange
#. exchange_output_id : The id of the inherited exchange's output
#. exchange_output_name : The name of the inherited exchange's output
#. input_id : The id of the input process
#. input_name : The name of the input process
#. output_id : The id of the output process
#. output_name : The name of the output process
#. amount : The floating-point amount
#. level : The numeric tree depth, going upwards from 0 as the child process
#. output_type : The numeric type of the output process
#. input_type : The numeric type of the input process
#. sigma : The sigma of the uncertainty distribution
#. uncertainty_type : The uncertainty type of the uncertainty distribution
#. minimum : The minimum of the uncertainty distribution
#. maximum : The maximum of the uncertainty distribution
#. geography : The name of the input process geography
#. category : The name of the input process category
#. database : The name of the input process database
#. unit : The name of the input process unit
    """
    if hasattr(process, "id"):
        process = process.id
    else:
        process = int(process)
    cursor = sql("core/inheritance/exchanges_for_process.sql", (process,), 
        using=using)
    if as_dict:
        if add_links:
            return _add_links(dict_cursor(cursor))
        else:
            return dict_cursor(cursor)
    else:
        return cursor.fetchall()

def _add_links(cursor):
    while 1:
        values = cursor.next()
        values.update({"input_url": reverse("process-view", kwargs={ 
            "process": values["input_name"], "id": values["input_id"]}), 
            "parent_url": reverse("process-view", kwargs={
            "process": values["exchange_output_name"], 
            "id": values["exchange_output_id"]})})
        yield values

def recursive_exchange(input_, output, using="default"):
    """
Use a recursive query to get all properties for an (possibly inherited) exchange.

.. rubric:: Inputs

* input_ : A process id or Process object.
* output : A process id or Process object.
* using : String. Database connection label.

.. rubric:: Outputs

A proxies.InheritedExchangeProxy object.
    """
    from proxies import InheritedExchangeProxy
    if hasattr(input_, "id"):
        input_ = input_.id
    else:
        input_ = int(input_)
    if hasattr(output, "id"):
        output = output.id
    else:
        output = int(output)
    cursor = sql("core/inheritance/exchange.sql", (output, input_), 
        using=using)
    description = [x[0] for x in cursor.description]
    result = cursor.fetchone()
    if not result:
        raise ObjectDoesNotExist
    return InheritedExchangeProxy(dict(zip(description, result)))

def recursive_weightings_by_method(method, using="default", as_dict=True):
    """
Use a recursive query to get LCIA weightings for biosphere flows which as assigned either to the flow or the flow's parent.

.. rubric:: Inputs

* method: A method id or Method object
* using : String. Database connection label.
* as_dict: Boolean. Return results as a list of dictionaries instead of tuples.

.. rubric:: Outputs

#. weighting_id : The id of the weighting row
#. process_id : The id of the biosphere flow
#. intermediate_parent_id : The id of the parent process providing the weighting
#. amount : The floating-point amount
#. level : The numeric tree depth, going upwards from 0 as the child process
#. sigma : The sigma of the uncertainty distribution
#. uncertainty_type : The uncertainty type of the uncertainty distribution
#. minimum : The minimum of the uncertainty distribution
#. maximum : The maximum of the uncertainty distribution
    """
    BIOSPHERE_TYPES = (1,2)
    if hasattr(method, "id"):
        method = method.id
    else:
        method = int(method)

    cursor = sql("core/inheritance/weightings_for_method.sql", 
        (BIOSPHERE_TYPES, method, WORLD().id, method, WORLD().id), 
        using=using, as_dict=as_dict)
    if as_dict:
        return cursor
    else:
        return cursor.fetchall()

def recursive_weightings_count_by_method(method, using="default"):
    """
Use a recursive query to get LCIA weightings for biosphere flows which as assigned either to the flow or the flow's parent.

.. important:: Only counts weightings with the ``Global`` geography. If the LCIA method is regionalized, use ``recursive_weightings_count_regionalized``.

.. rubric:: Inputs

* method: A method id or Method object
* using : String. Database connection label.
* as_dict: Boolean. Return results as a list of dictionaries instead of tuples.

.. rubric:: Outputs

Row count. Integer.
    """
    BIOSPHERE_TYPES = (1,2)
    if hasattr(method, "id"):
        method = method.id
    else:
        method = int(method)
    cursor = sql("core/inheritance/count_for_method.sql", (BIOSPHERE_TYPES, 
        method, WORLD().id, method, WORLD().id), using=using)
    return cursor.fetchone()[0]
