# -*- coding: utf-8 -*
from __future__ import division
import time
import os
import operator
import logging
from numpy import ones, zeros, argsort, array, arange, abs, hstack, where, \
    NaN, isnan
from scipy import sparse, stats
from scipy.sparse.linalg import dsolve
from django.conf import settings
from django.db.models.query import QuerySet
from brightway.apps.inventory.models import Database, Process, Exchange
from brightway.apps.impact_assessment.models import Method, Weighting
from brightway.core.simple_cache import SimpleCache
from brightway.core.cache_interface import CacheInterface
from brightway.core.errors import CacheError
from brightway.core.array_proxies import OneDimensionalArrayProxy, \
    SparseMatrixProxy, CompressedSparseMatrixProxy, ListArrayProxy
from brightway.core.utils.safe_hash import generate_safe_hash
from brightway.core.logging_decorators import timelogger
from brightway.core.lca.errors import *
from brightway.core import Container
from brightway.core.inheritance.queries import \
    recursive_exchanges_by_database, recursive_exchanges_count_by_database, \
    recursive_weightings_by_method, recursive_weightings_count_by_method

# Maximum value for unsigned integer stored in 4 bytes
MAX_INT_32 = 4294967295

class LCA(object):
    """.. centered:: The master LCA class; all variants on LCA should inherit from this class.

.. warning:: Before subclassing, be sure you understand :ref:`caching-in-brightway`, proxy matrices, and :ref:`sparse-formats`.

This class performs all the logic necessary for constructing the technosphere 
and biosphere matrices, including loading them from cache if possible; solving 
the supply vector, and performing life cycle impact assessment. It is designed 
to be modular enough that key components can be replaced or extended easily.

.. rubric:: Caching

This class takes a modular array to caching data from the database. It is 
rather expensive to build matrices from the database, so caching is used to 
avoid database hits. Caching is on the database and method level, and the 
``modified`` field is used to determine if the cache is fresh or not.

The basic unit of caching is a NumPy structured array for a method or database 
that gives matrix index information, object id values, and amount and 
uncertainty values. The structure is the same for database and method arrays, 
although method arrays don't use the ``output_id``, ``matrix_col``, 
``input_type``, and ``output_type`` columns. This is the structured array 
constructor for databases::

    my_array = array((values,), dtype=[
        ('uncertainty_type', 'i2'), ('input_id', 'u4'), 
        ('output_id', 'u4'), ('matrix_row', 'u4'), ('matrix_col', 'u4'), 
        ('input_type', 'i2'), ('output_type', 'i2'), 
        ('geography', 'u4'), ('amount', 'f4'), ('sigma', 'f4'), 
        ('minimum', 'f4'), ('maximum', 'f4'), ('negative', 'b')])

Here is the structured array constructor for methods::

    my_array = array((values,), dtype=[
        ('uncertainty_type', 'i2'), ('input_id', 'u4'), 
        ('matrix_row', 'u4'), ('matrix_col', 'u4'), ('geography', 'u4'), 
        ('amount', 'f4'), ('sigma', 'f4'), ('minimum', 'f4'), 
        ('maximum', 'f4'), ('negative', 'b')])

These arrays are understood natively by the uncertainty classes.

Databases can thus be cached individually, as can methods. Databases that use 
inheritance need to check the latest ``modified`` field for all ancestor 
databases as well as the original object.

.. rubric:: Inputs

* process : An id number, Process, or dictionary with id number keys and amount values.
* method : A queryset, Method, an id, or a string. Note that method names are not unique.
* follow_method_tree : Boolean. Optional, default is False. If True, will additionally calculate LCIA scores for all the child submethods of the method(s) you specified.
* write_to_cache: Boolean. Optional, default is True. Write matrices generated by this class to cache if they would update stale cached objects.
* read_from_cache : Boolean. Optional, default is True. Use cached matrices when possible.
* force_cache : Boolean. Optional, default is False. Force matrices generated by this class to be written to cache, even if their cached data is not stale.
* logging: Boolean. Optional, default is False. Otherwise, this is a logging level.
* cache_dir : String. Optional, default is *settings.BRIGHTWAY_CACHE_PATH*.
* extra_string : String. Optional, default is nothing. Used for subclasses to provide extra salt to the algorithms that determine unique cache names, so that subclasses can be sure that their cache names won't collide.
* using : String. Optional, default is "default." Database label, used with multiple databases.
* use_inheritance : Boolean. Optional, default is True. Use inheritance when determining what exchanges each child process has.

Retrieved from database:

#. self.technosphere_matrix
#. self.biosphere_matrix
#. self.characterization_matrix

Constructed:

#. self.demand_array

Calculated:

#. self.supply_array (same dimensions as self.demand_array)
#. self.calculated_biosphere
#. self.weighted_biosphere (a list)
    """
    def __init__(self, process, method=None, follow_method_tree=False, 
            write_to_cache=True, read_from_cache=True, force_cache=False, 
            log=False, cache_dir=None, extra_string="", using="default", 
            use_inheritance=True):
        # Initial assignment
        self.force_cache = force_cache
        self.follow_method_tree = follow_method_tree
        self.write_to_cache = write_to_cache
        self.read_from_cache = read_from_cache
        self.use_inheritance = use_inheritance
        # Used for subclasses to add arbitrary string to prevent cache 
        # collisions
        self.extra_string = unicode(extra_string)

        self.lci_finished = self.lcia_finished = False

        # Lookup dictionaries for objects to matrix indices
        self.dicts = Container() 

        # Setup logging
        if not hasattr(self, "default_logger"):
            self.default_logger = logging.getLogger('bw.assessment')
        self.logging, self.logger = self.setup_logging(log)

        # Parse input
        self.using = self.parse_using(using)
        self.parse_process(process)
        self.database_ids, self.database = self.parse_database()
        self.parse_method(method)
        self.inject_post_parsing()

        if not self.use_inheritance:
            self.check_if_database_uses_inheritance()

        # Construct useful parameters
        self.construct_canonical_names()
        self.modify_timestamps_with_inheritance

        # Cache directory
        self.cache_dir = cache_dir if cache_dir else \
            settings.BRIGHTWAY_CACHE_PATH

    ### Methods to parse user input ###
    @timelogger(logging.DEBUG)
    def parse_using(self, using):
        if using not in settings.DATABASES:
            raise ValueError("using must be a database")
        return using

    @timelogger(logging.DEBUG)
    def parse_process(self, process):
        if isinstance(process, (int, long)):
            process_dict = {process: 1.0}
        elif isinstance(process, dict):
            process_dict = process
        elif isinstance(process, Process):
            process_dict = {process.id: 1.0}
        elif isinstance(process, QuerySet):
            process_dict = dict([(p.id, 1.0) for p in process])
            self.process_queryset = process
        else:
            raise LCASetupError("Parameter process is an unrecognized type")

        try:
            keys = [int(key) for key in process_dict.keys()]
            values = [float(value) for value in process_dict.values()]
        except ValueError:
            raise LCASetupError(
                "Can't convert process IDs to integers or amounts to floats")

        for key in keys:
            # TODO: Check process is technosphere type
            if not Process.objects.filter(id=key).using(self.using).exists():
                raise LCASetupError("Process with id %s doesn't exist" % key)

        self.process = dict(zip(keys, values))
        if not hasattr(self, 'process_queryset'):
            self.process_queryset = Process.objects.filter(id__in=keys).using(
                self.using)

    @timelogger(logging.DEBUG)
    def parse_database(self):
        """Build database queryset from self.process. Databases are linked together in a directed graph (see DAG app)."""
        database_ids = set()
        for process_id in self.process:
            database_ids.update(set([obj.id for obj in Process.objects.using( 
                self.using).get(id=process_id).database.get_ancestors(
                include_self=True)]))

        return list(database_ids), Database.objects.filter(id__in=database_ids 
            ).using(self.using)

    @timelogger(logging.DEBUG)
    def parse_method(self, method):
        """Allow passing many method descriptions, for CLI convenience"""
        if not method:
            self.method = ()
            return

        if isinstance(method, Method):
            method_id = method.id
        elif isinstance(method, (str, unicode)):
            method_id = Method.objects.get(name__iexact=method).using(
                self.using).id
        elif isinstance(method, (int, long)):
            method_id = Method.objects.using(self.using).get(id=method).id
        elif isinstance(method, QuerySet):
            raise NotImplementedError, "Method querysets are not implemented."
        else:
            raise LCASetupError, "Method not recognized (%s:%s)." % (method,
                type(method))

        self.method = Method.objects.filter(id__in=(method_id,)).using(
            self.using)

        if self.follow_method_tree:
            # Implicitly includes ordering
            self.method = self.method[0].get_descendants(include_self=True)

        # It is safe not to cache this, because any change will invalidate
        # the cache.
        self.dicts.method = dict(enumerate([method.id for method in \
            self.method]))
        self.method_count = len(self.dicts.method)

    @timelogger(logging.DEBUG)
    def check_if_database_uses_inheritance(self):
        """Raise NeedInheritanceError if database(s) use inheritance, but use_inheritance is False."""
        if Process.objects.filter(database__in=self.database, 
                parent__isnull=False).using(self.using).exists():
            raise NeedInheritanceError

    @timelogger(logging.DEBUG)
    def modify_timestamps_with_inheritance(self):
        """Set ``modified`` for databases and to latest date in inheritance tree. Methods are not allowed to inherit through other methods."""
        for database in self.database:
            database.modified = database.get_ancestors(graph="inheritance", 
                include_self=True).order_by('-modified')[0].modified

    ### Methods to create useful class parameters ###
    @timelogger(logging.DEBUG)
    def construct_canonical_names(self):
        """
Keep all name generation in the same function.

Called by __init__, and by get_solution_matrices for each process key in 
self.process.
        """
        base_string = unicode(self.using) + unicode(repr(self.__class__)
            ) + self.extra_string
        for method in self.method:
            method.hash = generate_safe_hash(base_string + method.uuid)
        for database in self.database:
            database.hash = generate_safe_hash(base_string + database.uuid)

    @timelogger(logging.DEBUG)
    def construct_reverse_dicts(self):
        """Provides lookup from matrix indices to database ids"""
        self.dicts.reverse = Container()
        self.dicts.reverse.technosphere = dict(zip(
            self.dicts.technosphere.values(), self.dicts.technosphere.keys()))
        self.dicts.reverse.biosphere = dict(zip(
            self.dicts.biosphere.values(), self.dicts.biosphere.keys()))

    def setup_logging(self, l):
        if not l:
            return False, None
        elif l == True:
            return True, self.default_logger
        else: # l is a logging level
            try:
                level = int(l)
                self.default_logger.setLevel(l)
                return True, self.default_logger
            except:
                raise ValueError, \
                    "'logging' must be a logging level, e.g. logging.DEBUG"

    ### Methods to load objects from the cache ###
    def simple_cache(self, filename, obj=None, expiration=None):
        """Use the SimpleCache"""
        return self.cache(filename, obj, expiration, cls=SimpleCache)

    def cache(self, filename, obj=None, expiration=None, cls=CacheInterface):
        """Uses CacheInterface for more complicated objects"""
        if obj is None:
            value = cls(self.cache_dir, filename, 
                expiration=expiration).load()
            if value not in (None, False):
                return value
            else:
                raise CacheError
        else:
            cls(self.cache_dir, filename, obj)

    @timelogger(logging.INFO)
    def get_lci_matrices(self):
        '''Load lci exchange arrays from cache, or build them'''
        exchange_arrays = []

        if self.logging:
            then = lci_start_time = time.time()

        for database in self.database:
            try:
                obj = self.simple_cache(filename=database.hash,
                    expiration=database.modified)
                database.from_cache = True
            except CacheError:
                obj = self.build_exchange_array_for_database(database)
                database.from_cache = False
                if self.write_to_cache:
                    self.simple_cache(database.hash, obj)

            exchange_arrays.append(obj)

        for process in Process.objects.filter(
                database__id__in=self.database_ids, 
                process_type=Process.FROZEN_TYPE).using(self.using):
            # TODO: Cache per database?
            exchange_arrays.append(self.build_frozen_exchange_array(process))

        if self.logging:
            now = time.time()
            self.logger.info("Build/load exchange arrays: %.4g (%.4g total)" \
                % (now - then, now - lci_start_time))
            then = now

        self.inventory_exchange_arrays = hstack(exchange_arrays)

        if self.logging:
            now = time.time()
            self.logger.info("Stack exchange arrays: %.4g (%.4g total)" % \
                (now - then, now - lci_start_time))
            then = now

        # Only need edges in the technosphere matrix that are from a 
        # technology or frozen process
        input_indices = hstack((
            where(self.inventory_exchange_arrays['input_type'] == \
            Process.TECHNOLOGICAL_PROCESS_TYPE)[0], # Tech processes as inputs
            where(self.inventory_exchange_arrays['input_type'] == \
            Process.FROZEN_TYPE)[0])) # Frozen processes as inputs
        # But frozen processes without inputs still need to be on the diagonal
        tech_process_ids = set(hstack((
            self.inventory_exchange_arrays['input_id'][input_indices],  
            self.inventory_exchange_arrays['output_id'][ 
            where(self.inventory_exchange_arrays['output_type'] == \
            Process.TECHNOLOGICAL_PROCESS_TYPE)]))) # Frozen processes
        self.technosphere_matrix = self.technosphere_matrix_from_array(
            self.inventory_exchange_arrays[input_indices], 
            obj_ids=tech_process_ids)
        self.technosphere_count = self.technosphere_matrix.shape[0]
        self.dicts.technosphere = self.technosphere_matrix.col_dict

        if self.logging:
            now = time.time()
            self.logger.info("Construct T matrix: %.4g (%.4g total)" % \
                (now - then, now - lci_start_time))
            then = now

        arry = self.inventory_exchange_arrays[where(
            self.inventory_exchange_arrays['input_type'] != \
                Process.TECHNOLOGICAL_PROCESS_TYPE)]

        if self.logging:
            now = time.time()
            self.logger.info("Filter exchange array for biosphere flows:" + \
                " %.4g (%.4g total)" % (now - then, now - lci_start_time))
            then = now

        self.biosphere_matrix = self.biosphere_matrix_from_array(
            arry, 
            col_dict=self.dicts.technosphere)
        self.biosphere_count = self.biosphere_matrix.shape[0]
        self.dicts.biosphere = self.biosphere_matrix.row_dict

        if self.logging:
            now = time.time()
            self.logger.info("Construct B matrix: %.4g (%.4g total)" % \
                (now - then, now - lci_start_time))
            then = now

    @timelogger(logging.INFO)
    def get_lcia_matrices(self):
        """The characterization matrix is a diagonal matrix with dimensions (biosphere flows by biosphere flows), which gives the characterization factors applied to each biosphere flow by the impact assessment method."""
        arrays = []

        for method in self.method:
            try:
                obj = self.simple_cache(method.hash, 
                    expiration=method.modified)
                method.from_cache = True
            except CacheError:
                obj = self.build_exchange_array_for_method(method)
                method.from_cache = False
                if self.write_to_cache:
                    self.simple_cache(method.hash, obj)

            arrays.append(obj)

        self.ia_exchange_arrays = [self.clean_characterization_array(obj 
            ) for obj in arrays]
        self.characterization_matrices = ListArrayProxy([ 
            self.build_sparse_matrix(obj['amount'], obj['matrix_row'], 
            obj['matrix_row'], self.biosphere_count, self.biosphere_count, 
            self.dicts.biosphere, self.dicts.biosphere) for obj in \
            self.ia_exchange_arrays], self.method)

    @timelogger(logging.DEBUG)
    def exchange_array_for_iterable(self, qs, dtype, row_keys, count=None):
        """
Build exchange array from itreable (either ValuesQuerySet or raw SQL result).

Used for construction of inventory and impact assessment parameter arrays, and by subclasses for other matrix construction.

.. rubric:: Inputs

* qs: Iterable with key lookups (either ValuesQuerySet or SQL converted to dict)
* dtype: Tuple of tuple of strings. NumPy format strings for each name, e.g. ('foo', 'u4').
* row_keys: Tuple of strings. Column names, which can be nicer than default Django names.
* count: Integer. Total number of records in iterable. ``qs`` must support ``.count()`` method if not specified. 
        """
        """Build a numpy array, with rows of: uncertainty_type, input_id, output_id, amount, sigma, minimum, maximum, negative.
        
        Contains all exchanges, for both technosphere and biosphere matrices."""
        sarray = zeros((count,), dtype=dtype + [('matrix_row', 'u4'), 
            ('matrix_col', 'u4'), ('negative', 'b')])

        if count is None:
            count = qs.count()

        # Should be optimzed, use fromiter or fromrecords?
        # See http://docs.scipy.org/doc/numpy/reference/generated/numpy.core.records.fromrecords.html
        # http://docs.scipy.org/doc/numpy/user/basics.rec.html
        # http://stackoverflow.com/questions/1741107/how-do-i-convert-a-django-queryset-to-numpy-record-array
        # Use MAX_INT_32 as a mask for matrix indices which aren't known yet.
        for index, row in enumerate(qs):
            sarray[index] = tuple([row[key] for key in row_keys] + [ 
                MAX_INT_32, MAX_INT_32, False])

        return sarray

    @timelogger(logging.DEBUG)
    def build_frozen_exchange_array(self, process):
        data = process.frozen_data
        count = len(data)
        qs = ({'uncertainty_type': 0, # Uncertainty type
            'input_id': key, 
            'output_id': process.id,
            'input_type': Process.EMISSION_TYPE, # Generic biosphere type
            # Output is still a technosphere process
            'output_type': Process.TECHNOLOGICAL_PROCESS_TYPE, 
            'geography_id': process.geography_id, 
            'amount': value, 
            'sigma': NaN, 
            'minimum': NaN, 
            'maximum': NaN} for key, value in data.iteritems())
        dtype = [('uncertainty_type', 'i2'), ('input_id', 'u4'), 
            ('output_id', 'u4'), ('input_type', 'i2'), ('output_type', 'i2'), 
            ('geography_id', 'u4'), ('amount', 'f4'), ('sigma', 'f4'), 
            ('minimum', 'f4'), ('maximum', 'f4')]
        row_keys = [x[0] for x in dtype]
        return self.exchange_array_for_iterable(qs, dtype, row_keys, count)

    @timelogger(logging.DEBUG)
    def build_exchange_array_for_database(self, database):
        dtype = [('uncertainty_type', 'i2'), ('input_id', 'u4'), 
            ('output_id', 'u4'), ('input_type', 'i2'), ('output_type', 'i2'), 
            ('geography_id', 'u4'), ('amount', 'f4'), ('sigma', 'f4'), 
            ('minimum', 'f4'), ('maximum', 'f4')]

        if self.use_inheritance:
            count = recursive_exchanges_count_by_database(database, 
                self.using)
            qs = recursive_exchanges_by_database(database, self.using)
            row_keys = [x[0] for x in dtype]
        else:
            qs = Exchange.objects.using(self.using).filter( 
                output__database__id__in=self.database_ids).exclude(
                output__process_type=Process.FROZEN_TYPE)
            count = qs.count()
            row_keys = ('uncertainty_type', 'input_id', 'output_id', 
                'input__process_type', 'output__process_type', 'geography_id', 
                'amount', 'sigma', 'minimum', 'maximum')
            qs = qs.values(*row_keys)

        return self.exchange_array_for_iterable(qs, dtype, row_keys, count)

    @timelogger(logging.DEBUG)
    def build_exchange_array_for_method(self, method):
        dtype = [('uncertainty_type', 'i2'), ('process_id', 'u4'), 
            ('amount', 'f4'), ('sigma', 'f4'), ('minimum', 'f4'), 
            ('maximum', 'f4')]
        row_keys = [x[0] for x in dtype]

        if self.use_inheritance:
            count = recursive_weightings_count_by_method(method, self.using)
            qs = recursive_weightings_by_method(method, self.using)
        else:
            qs = Weighting.objects.filter(method=method)
            count = qs.count()
            qs = qs.values(*row_keys)

        return self.exchange_array_for_iterable(qs, dtype, row_keys, count)

    @timelogger(logging.INFO)
    def add_matrix_indices(self, array_from, array_to, mapping):
        """
Modifies an array in place that gives the rank of each element in an array, with consecutive ranks and ties assigned the same value. These ranks are used as  matrix indices.

Chosen algorithm is faster than using rankdata, or simple iteration. map(dict.get, data) seems faster, but doesn't work with missing values. Setup code to try new approaches::

    from numpy import *
    from time import time
    a = random.random_integers(5000, size=100000)
    indices = sorted(list(set([int(x) for x in a])))
    mapping = dict(zip(indices, range(len(indices))))
    b = zeros(a.shape[0])

        """
        array_to[:] = array([mapping.get(x, MAX_INT_32) for x in array_from])

    @timelogger(logging.INFO)
    def technosphere_matrix_from_array(self, arry, obj_ids=None, 
            common_dict=None):
        """Turn exchanges array into technosphere matrix proxy. Construct dictionaries if necessary."""
        if common_dict:
            count = len(common_dict)
            indices = range(count)
        else:
            if not obj_ids:
                obj_ids = set(arry['input_id']).union(set(arry['output_id']))
            count = len(obj_ids)
            indices = range(count)
            common_dict = dict(zip(obj_ids, indices))

        self.add_matrix_indices(arry['input_id'], arry['matrix_row'], 
            common_dict)
        self.add_matrix_indices(arry['output_id'], arry['matrix_col'], 
            common_dict)

        self.technosphere_array = arry

        return self.build_sparse_matrix(arry['amount'], arry['matrix_row'], 
            arry['matrix_col'], count, count, common_dict, common_dict, 
            is_technosphere=True, indices=indices)

    @timelogger(logging.INFO)
    def biosphere_matrix_from_array(self, arry, row_dict=None, col_dict=None):
        """Turn exchanges array into technosphere matrix proxy. Construct dictionaries if necessary."""
        if not row_dict:
            obj_ids = set(arry['input_id'])
            row_count = len(obj_ids)
            indices = range(row_count)
            row_dict = dict(zip(obj_ids, indices))
        else:
            row_count = len(row_dict)
            row_indices = range(row_count)

        if not col_dict:
            obj_ids = set(arry['output_id'])
            col_count = len(obj_ids)
            indices = range(col_count)
            col_dict = dict(zip(obj_ids, indices))
        else:
            col_count = len(col_dict)
            col_indices = range(col_count)

        self.add_matrix_indices(arry['input_id'], arry['matrix_row'], 
            row_dict)
        self.add_matrix_indices(arry['output_id'], arry['matrix_col'], 
            col_dict)

        # Eliminate references to biosphere flows that don't appear in this
        # assessment; they are masked with MAX_INT_32 values
        arry = arry[arry['matrix_col'] != MAX_INT_32]

        self.biosphere_array = arry

        return self.build_sparse_matrix(arry['amount'], arry['matrix_row'], 
            arry['matrix_col'], row_count, col_count, row_dict, col_dict)

    @timelogger(logging.DEBUG)
    def build_sparse_matrix(self, data, rows, cols, row_count, col_count, 
            row_dict=None, col_dict=None, return_proxy=True, 
            is_technosphere=False, indices=None):
        """Common functionality to build coo_matrix and compress."""
        if is_technosphere:
            assert row_count == col_count
            indices = indices or range(len(row_dict))
            # Add ones along the diagonal
            data = hstack((-1 * data, ones((row_count,))))
            rows = hstack((rows, indices))
            cols = hstack((cols, indices))
        # coo_matrix construction is coo_matrix((values, (rows, cols)), 
        # (row_count, col_count))
        coo_matrix = sparse.coo_matrix((data, (rows, cols)), 
            (row_count, col_count))
        if return_proxy:
            if (not row_dict) or (not col_dict):
                raise ValueError, \
                    "Must provide dictionaries to construct a sparse proxy"
            return CompressedSparseMatrixProxy(coo_matrix.tocsr(), 
                row_dict, col_dict)
        else:
            return coo_matrix.tocsr()

    @timelogger(logging.INFO)
    def clean_characterization_array(self, arry):
        self.add_matrix_indices(arry['process_id'], arry['matrix_row'], 
            self.dicts.biosphere)

        # Eliminate references to biosphere flows that don't appear in this
        # assessment; they are masked with MAX_INT_32 values
        return arry[arry['matrix_row'] != MAX_INT_32]

    ### Methods to build arrays and matrices ###
    @timelogger(logging.DEBUG)
    def build_demand_array(self):
        self.demand_array = OneDimensionalArrayProxy(zeros(
            self.technosphere_count), self.dicts.technosphere)
        for key in self.process:
            self.demand_array[key] = self.process[key]

    ### Injection functions for arbitrary modification of LCA procedures ###
    def inject_post_parsing(self, *args, **kwargs):
        pass

    def inject_post_demand_array(self, *args, **kwargs):
        pass

    def inject_post_lci_exchanges_array(self, *args, **kwargs):
        pass

    def inject_post_lcia_exchanges_array(self, *args, **kwargs):
        pass

    def inject_post_solve_linear_system(self, *args, **kwargs):
        pass

    def inject_post_lci(self, *args, **kwargs):
        pass

    def inject_post_lcia(self, *args, **kwargs):
        pass

    ### Methods for LCA mathematics ###
    @timelogger(logging.INFO)
    def solve_linear_system(self, *args): # Can pass a, b
        """
Master solution function for linear system a*x=b. 

This is separate from 'solve_system' so that it can be easily subclassed or 
replaced. Can be passed the technosphere matrix and demand vector, but is 
normally can be called directly; normal parameters are assumed.

Default uses UMFpack, and stores the context so that solving subsequent times 
are quick. "To most numerical analysts, matrix inversion is a sin." - Nicolas Higham, Accuracy and Stability of Numerical Algorithms, 2002, p. 260.
        """
        if args:
            if not len(args) == 2:
                raise LCASetupError, "Invalid inputs to solve_linear_system function (%s)" % args
            a, b = args
        else:
            a, b = self.technosphere_matrix.data, self.demand_array.data

        if dsolve:
            self.solver = dsolve.factorized(a.tocsc())
            return self.solver(b)
        else:
            # Scipy < 0.7
            self.solver = None
            return spsolve(a, b)

    @timelogger(logging.DEBUG)
    def lci(self):
        """
Calcuate the life-cycle inventory (supply vector, emissions & resource consumption) of the system.

This is the heart of the LCA calculation.

#. Load technosphere matrix; usually this is cached because it is expensive to load from the DB
#. Coerce self._process into an array with correct dimensions
#. Use one line of scipy magic! (or as otherwise provided in solve)
#. Generate biosphere matrix
#. Calculate emissions and resource use for system.
        """
        if self.logging:
            lci_start_time = time.time()

        self.get_lci_matrices()

        self.build_demand_array()
        self.inject_post_demand_array()

        self.supply_array = OneDimensionalArrayProxy(
            self.solve_linear_system(), self.dicts.technosphere)
        self.inject_post_solve_linear_system()

        self.calculated_biosphere = CompressedSparseMatrixProxy(
            self.biosphere_matrix.data * sparse.spdiags(
            [self.supply_array.data], [0], self.technosphere_count,
            self.technosphere_count), self.dicts.biosphere,
            self.dicts.technosphere)
        self.inject_post_lci()
        self.lci_finished = True

    @timelogger(logging.DEBUG)
    def lcia(self):
        """Life Cycle Impact Assessment"""
        #TODO: Add error messages if adequate components not provided
        if not self.lci_finished:
            raise LCAOperationalError, "Must do lci before doing lcia"

        if not self.method:
            raise LCASetupError, "Must specify a method to perform LCIA"

        if self.logging:
            now = time.time()

        self.get_lcia_matrices()

        self.weighted_biosphere = ListArrayProxy([CompressedSparseMatrixProxy(
            matrix.data*self.calculated_biosphere.data,
            self.dicts.biosphere, self.dicts.technosphere
            ) for matrix in self.characterization_matrices], 
            self.method)
        self.inject_post_lcia()
        self.lcia_finished = True

        if self.logging:
            self.logger.info("LCIA mathematics took %.5f seconds" % \
                (time.time() - now))

    @timelogger(logging.INFO)
    def redo_lci(self, process):
        """Redo lci with new process. 
        
        Use UMF factorized technosphere matrix if possible. Does not test if a cached solution is available."""
        self.parse_process(process)
        if not self.lci_finished:
            self.lci()
        else:
            self.build_demand_array()
            self.inject_post_demand_array()

            # Reuse factorization if possible.
            if self.solver is not None:
                self.supply_array = OneDimensionalArrayProxy(
                    self.solver(self.demand_array.data), 
                    self.dicts.technosphere)
            else:
                self.supply_array = OneDimensionalArrayProxy(
                    self.solve_linear_system(), self.dicts.technosphere)

            self.inject_post_solve_linear_system()
            self.calculated_biosphere = CompressedSparseMatrixProxy(
                self.biosphere_matrix.data * sparse.spdiags(
                [self.supply_array.data], [0], self.technosphere_count,
                self.technosphere_count), self.dicts.biosphere,
                self.dicts.technosphere)

    @timelogger(logging.INFO)
    def redo_lcia(self, method=None):
        """Redo lcia. New method is optional."""
        if not self.lcia_finished:
            if method:
                self.parse_method(method)
                self.inject_post_parsing()
            self.lcia()
        elif method:
            self.parse_method(method)
            self.inject_post_parsing()
            self.lcia()
        elif (self.dicts.technosphere != self.weighted_biosphere[0].col_dict
                ) or (self.dicts.biosphere != self.weighted_biosphere[ 
                0].row_dict):
            # Method is the same, but inventory matrices changed.
            self.lcia()
        else:
            self.weighted_biosphere = ListArrayProxy( 
                [CompressedSparseMatrixProxy(matrix.data * \
                self.calculated_biosphere.data,
                self.dicts.biosphere, self.dicts.technosphere
                ) for matrix in self.characterization_matrices], 
                self.method)
            self.inject_post_lcia()

    def _top(self, data, dict_, limit=10, limit_type="index", annotate=False, 
            absolute=False):
        """Common function for all top methods. Data is a one dimensional array of lci/lcia scores, and dict_ is a dictionary of indices to process ids."""
        if limit_type not in ("index", "percent"):
            raise ValueError("limit_type must be either 'percent' or 'index'.")
        if limit_type == "percent":
            if limit > 1:
                raise ValueError("Percentage limits must be between 0 and 1.")
            data_sum = abs(data.sum()) if absolute else data.sum()
            limit = (data > data_sum * limit).sum()

        results = zeros((data.shape[0],2))
        results[:,0] = data
        results[:,1] = arange(data.shape[0])
        if absolute:
            results = results[argsort(abs(data))[::-1]][:limit,:]
        else:
            results = results[argsort(data)[::-1]][:limit,:]

        for row in results:
            row[1] = dict_[row[1]]

        if annotate:
            return [(row[0], Process.objects.using(self.using).get(id=row[1]) 
                ) for row in results]
        else:
            return results

    def top_processes(self, index=0, limit=10, limit_type="index", axis=0,
            annotate=False):
        '''
Return an array of [value, database id] rows, or a sorted list with elements [value, Process, database id] if annotate is true.
        
        Axis=0 are columns, or processes in the weighted_biosphere matrix. Axis=1 are rows, or emissions in the weighted_biosphere matrix.
        '''
        if not hasattr(self, "weighted_biosphere"):
            return LCAOperationalError("Must perform LCIA to analyze LCIA" + \
                " results")
        if axis not in (0,1):
            raise ValueError("axis must be 0 or 1 (rows or columns)")
        if not hasattr(self.dicts, 'reverse'):
            self.construct_reverse_dicts()

        summed = array(self.weighted_biosphere[index].sum(axis=axis)).ravel()

        if axis == 0:
            dict_ = self.dicts.reverse.technosphere
        else:
            dict_ = self.dicts.reverse.biosphere

        return self._top(summed, dict_, limit=limit, limit_type=limit_type, 
            annotate=annotate)

    def top_emissions(self, **kwargs):
        '''Return sorted list of emissions by total score'''
        if 'axis' in kwargs:
            del kwargs['axis']
        return self.top_processes(axis=1, **kwargs)

    def top_submethods(self):
        '''Return sorted list of methods from self.method by total score'''
        if self.method_count < 2: 
            return None # No submethods
        return sorted([(self._W[x].sum(), self.method[x]) for x in range(1,
            self.method_count)], reverse=True)

    def top_emissions_for_process(self, process, **kwargs):
        if hasattr(process, "id"):
            process = process.id
        if not hasattr(self.dicts, 'reverse'):
            self.construct_reverse_dicts()
        return self._top(array(self.weighted_biosphere[:,process].todense( 
            )).ravel(), self.dicts.reverse.biosphere, **kwargs)

    def top_processes_for_emission(self, biosphere_flow, **kwargs):
        if hasattr(biosphere_flow, "id"):
            biosphere_flow = biosphere_flow.id
        if not hasattr(self.dicts, 'reverse'):
            self.construct_reverse_dicts()
        return self._top(array(self.weighted_biosphere[biosphere_flow,: 
            ].todense()).ravel(), self.dicts.reverse.technosphere, **kwargs)

    def top_processes_for_emission_inventory(self, emission, **kwargs):
        """Get the most important inventory processes for an emission"""
        if hasattr(emission, "id"):
            emission = emission.id
        if not hasattr(self.dicts, 'reverse'):
            self.construct_reverse_dicts()
        return self._top(array(self.calculated_biosphere[emission,:].todense( 
            )).ravel(), self.dicts.reverse.technosphere, **kwargs)

    @property
    def score(self):
        if hasattr(self, "weighted_biosphere"):
            return reduce(sum, [x.sum() for x in self.weighted_biosphere])
        else:
            raise LCAOperationalError, \
                "Must perform LCI/LCIA to calculate score"

