# -*- coding: utf-8 -*
from __future__ import division
import time
import logging
from scipy import sparse
from numpy import array, zeros, argsort, ones, hstack, vstack, arange, NaN, \
    where
from brightway.apps.inventory.models import Process, Exchange
from brightway.apps.impact_assessment.models import Weighting
from brightway.core.lca.lca import LCA
from brightway.core.lca.errors import *
from brightway.core.logging_decorators import timelogger
from stats_toolkit.uncertainty_choices import UncertaintyChoices
from stats_toolkit.random import MCRandomNumberGenerator
from brightway.core.cache_interface import CacheInterface
from brightway.core import Container

TECHNOSPHERE_TYPE = Process.TECHNOLOGICAL_PROCESS_TYPE
BIOSPHERE_TYPES = (Process.RESOURCE_CONSUMPTION_TYPE, 
    Process.RESOURCE_CONSUMPTION_TYPE)


class MethodArrayList(object):
    """Wrapper for a list of RNGs for LCIA characterization factors"""
    def __init__(self, arrays, rng, kwargs, biosphere_count):
        self.arrays = [rng(arry, sort=False, **kwargs) for arry in arrays]
        self.count = biosphere_count

    def iterate(self):
        return [sparse.coo_matrix((obj.iterate(), (obj['matrix_row'], 
            obj['matrix_row'])), (self.count, self.count)).tocsr() for obj \
            in self.arrays]


class MonteCarloLCA(LCA):
    """.. centered:: Monte Carlo class, including analysis functions

Inputs:

# This is a NumPy structured array: http://docs.scipy.org/doc/numpy/user/basics.rec.html

#. process: As for LCA
#. vary_technosphere: Vary technological parameters. Default is True.
#. mc_biosphere: Vary biosphere parameters. Default is True.
#. mc_weighting: Vary weighting parameters. Default is False.

Outputs:

# self.mc.weighted_results: Array of summed weighted biosphere flows for each MC instance. Dimensions are [biosphere flows, iterations].
# self.mc.biosphere_results: Array of summed biosphere flows for each MC instance. Dimensions are [biosphere flows, iterations].
# self.mc.supply_results: Array of summed technological inputs (supply vector) for each MC instance. Dimensions are [processes, iterations].

Internal Data:

#. self.mc.technosphere_positions: List of [uncertainty type, start index, end index], used to populate random data.
#. self.mc.biosphere_positions: List of [uncertainty type, start index, end index], used to populate random data.
#. self.mc.biosphere_matrix_instance: Last realization of the biosphere matrix
#. self.mc.technosphere_matrix_instance: Last realization of the technosphere matrix
    """
    def __init__(self, process, vary_technosphere=True, vary_biosphere=True, 
            vary_characterization=False, iterations=100, **kwargs):
        self.iterations = iterations
        self.mc = Container()
        self.mc.completed = False
        self.vary_biosphere = vary_biosphere
        self.vary_characterization = vary_characterization
        self.vary_technosphere = vary_technosphere
        self.rng = MCRandomNumberGenerator
        self.rng_kwargs = {}
        # Force parameter array to be created
        kwargs['use_inheritance'] = True
        super(MonteCarloLCA, self).__init__(process, **kwargs)

    @timelogger(logging.DEBUG)
    def build_new_technosphere_matrix_instance(self):
        self.mc.technosphere_random_data = \
            self.mc.technosphere_generator.iterate()
        self.mc.technosphere_matrix_instance = self.build_sparse_matrix( 
            self.mc.technosphere_random_data, 
            self.mc.technosphere_array['matrix_row'], 
            self.mc.technosphere_array['matrix_col'], self.technosphere_count, 
            self.technosphere_count, row_dict=self.dicts.technosphere, 
            return_proxy=False, is_technosphere=True)

    @timelogger(logging.DEBUG)
    def build_new_biosphere_matrix_instance(self):
        self.mc.biosphere_random_data = self.mc.biosphere_generator.iterate()
        self.mc.biosphere_matrix_instance = self.build_sparse_matrix( 
            self.mc.biosphere_random_data, self.mc.biosphere_array[ 
            'matrix_row'], self.mc.biosphere_array['matrix_col'], 
            self.biosphere_count, self.technosphere_count, return_proxy=False, 
            is_technosphere=False)

    @timelogger(logging.DEBUG)
    def build_mc_technosphere_array(self):
        ordering = argsort(self.technosphere_array["uncertainty_type"])
        self.mc.technosphere_array = self.technosphere_array[ordering]

    @timelogger(logging.DEBUG)
    def build_mc_biosphere_array(self):
        ordering = argsort(self.biosphere_array["uncertainty_type"])
        self.mc.biosphere_array = self.biosphere_array[ordering]

    @timelogger(logging.DEBUG)
    def build_mc_characterization_array(self):
        """Order all characterization factors arrays by uncertainty type"""
        self.mc.characterization_arrays = [obj[argsort(obj["uncertainty_type"] 
            )] for obj in self.ia_exchange_arrays]

    def lci(self):
        if self.logging:
            self.logger.error("Called Monte Carlo lci method")
        raise LCAOperationalError, "lci() can't be called separately in Monte Carlo LCA; use calculate()"

    def lcia(self):
        if self.logging:
            self.logger.error("Called Monte Carlo lcia method")
        raise LCAOperationalError, "lcia() can't be called separately in Monte Carlo LCA; use calculate()"

    @timelogger(logging.DEBUG)
    def iterate(self, index=0, only_lci=False):
        if self.vary_technosphere:
            self.build_new_technosphere_matrix_instance()
            supply_array = self.solve_linear_system(
                self.mc.technosphere_matrix_instance, self.demand_array.data)
        elif not hasattr(self.mc, "technosphere_matrix_instance"):
            # demand_array could have been changed, so still recalculate
            supply_array = self.solver(self.demand_array.data)

        if self.vary_biosphere:
            self.build_new_biosphere_matrix_instance()
        elif not hasattr(self, "biosphere_matrix_instance"):
            self.mc.biosphere_matrix_instance = self.biosphere_matrix.data

        calculated_biosphere = self.mc.biosphere_matrix_instance * \
            sparse.spdiags([supply_array], [0], self.technosphere_count, 
            self.technosphere_count)
        self.mc.supply_results[:,index] = supply_array

        if not only_lci:
            if self.vary_characterization:
                objs = self.mc.characterization_generator.iterate()
            else:
                objs = [o.data for o in self.characterization_matrices]

            for method_i, method in enumerate(objs):
                weighted_biosphere = method * calculated_biosphere
                self.mc.technosphere_results[:,method_i,index] = \
                    weighted_biosphere.sum(axis=0)
                # flatten transforms from (n,1) array to (n,) array
                self.mc.biosphere_results[:,method_i,index] = \
                    weighted_biosphere.sum(axis=1).flatten()

    @timelogger(logging.INFO)
    def calculate(self, only_lci=False):
        if not self.method and not only_lci:
            raise LCASetupError, "Must specify a method to perform LCIA"

        self.get_lci_matrices()
        self.build_demand_array()
        self.get_lcia_matrices()

        if self.vary_technosphere:
            self.build_mc_technosphere_array()
            self.mc.technosphere_generator = self.rng( 
                self.mc.technosphere_array, sort=False, **self.rng_kwargs)

        if self.vary_biosphere:
            self.build_mc_biosphere_array()
            self.mc.biosphere_generator = self.rng(self.mc.biosphere_array, 
                sort=False, **self.rng_kwargs)

        if not only_lci:
            self.build_mc_characterization_array()

            if self.vary_characterization:
                self.mc.characterization_generator = MethodArrayList( 
                    self.mc.characterization_arrays, self.rng, 
                    self.rng_kwargs, len(self.dicts.biosphere))

        self.inject_post_demand_array()

        # Arrays for saving MC runs
        # This saves the solution (input) array runs
        self.mc.supply_results = zeros((self.technosphere_count, 
            self.iterations))

        if not only_lci:
            # These save the weighted LCIA matrix column/row sums
            self.mc.technosphere_results = zeros((self.technosphere_count, 
                self.method_count, self.iterations))
            self.mc.biosphere_results = zeros((self.biosphere_count, 
                self.method_count, self.iterations))

        for index in range(self.iterations):
            self.iterate(index, only_lci)

        self.mc.completed = True

    def get_total_scores(self, index=0):
        return self.mc.biosphere_results[:,index,:].sum(axis=0)

    def get_supply_for_process(self, process):
        process_index = self.dicts.technosphere[process]
        return self.mc.supply_results[process_index,:]