# -*- coding: utf-8 -*-
from time import sleep
from scipy import sparse, linalg, matrix
from numpy import array, allclose, identity, zeros
from django.conf import settings
from brightway.core.lca.lca import LCA
from brightway.apps.inventory.models import Database, Exchange, Process
from brightway.tests.lca_test_case import LCATestCase
from brightway.core.cache_interface import CacheInterface

class CacheTestCase(LCATestCase):
    """Test LCA interace to CacheInterface. 
    CacheInterface has separate tests for pure cache functionality."""

    def test_reading(self):
        '''Test that cache loads correctly.'''
        # Write to cache
        lca_writer = LCA(self.processes.steel, read_from_cache=False,
            method=self.methods.human_health, **self.redirect_kwargs)
        lca_writer.lci()
        lca_writer.lcia()
        self.assertFalse(lca_writer.read_from_cache)
        self.assertFalse(lca_writer.from_cache.solution_matrices)
        self.assertFalse(lca_writer.from_cache.technosphere_matrix)
        self.assertFalse(lca_writer.from_cache.biosphere_matrix)

        # Read from cache
        lca_reader = LCA(self.processes.steel, write_to_cache=False, 
            method=self.methods.human_health, **self.redirect_kwargs)
        lca_reader.lci()
        lca_reader.lcia()

        self.assertTrue(lca_reader.from_cache.solution_matrices)
        self.assertFalse(hasattr(lca_reader.from_cache, "technosphere_matrix"))
        self.assertFalse(hasattr(lca_reader.from_cache, "biosphere_matrix"))

        self.assertEqual(lca_reader.dicts.technology, 
            lca_writer.dicts.technology)
        self.assertEqual(lca_reader.dicts.emission, 
            lca_writer.dicts.emission)
        self.assertTrue(allclose(lca_reader.calculated_emissions.data.todense(), 
            lca_writer.calculated_emissions.data.todense()))
        self.assertTrue(allclose(lca_reader.supply_array.data, 
            lca_writer.supply_array.data))
        self.assertTrue(allclose(lca_reader.weighting_matrix[0].data.todense(), 
            lca_writer.weighting_matrix[0].data.todense()))

    def test_dont_load_technosphere_matrix_if_solution_cache(self):
        '''Technology matrix should not be loaded if solution cache is found.'''
        lca_writer = LCA(self.processes.steel, read_from_cache=False,
            method=self.methods.human_health, **self.redirect_kwargs)
        lca_writer.lci()
        lca_writer.lcia()

        # Read from cache
        lca_reader = LCA(self.processes.steel, 
            write_to_cache=False, method=self.methods.human_health, 
            **self.redirect_kwargs)
        lca_reader.lci()
        self.assertFalse(hasattr(lca_reader, "technosphere_matrix"))

    def test_load_separate_processes_solution_caches(self):
        '''Test loading and scaling of separate solution caches.'''
        pass

    def test_load_separate_processes_fail_missing_process(self):
        '''Test cache loading failed if one process result is not cached.'''
        pass

    def test_cache_fails_if_process_modified(self):
        '''Cache loading should fail if a process is modified.'''
        cache_builder = LCA(self.processes.steel, read_from_cache=False, 
            **self.redirect_kwargs)
        cache_builder.lci()
        # Invalidate cache
        sleep(1)
        self.processes.steel.save()
        uncached_lca = LCA(self.processes.steel, **self.redirect_kwargs)
        self.assertFalse(uncached_lca._get_cached_matrices(
            filename=uncached_lca.names.lci_matrices,
            expiration=uncached_lca.database_modified))
    
    def test_cache_fails_if_database_modified(self):
        '''Cache loading should fail is a database is modified.'''
        cache_builder = LCA(self.processes.steel, read_from_cache=False, 
            **self.redirect_kwargs)
        cache_builder.lci()
        # Invalidate cache
        sleep(1)
        self.databases.master.save()
        uncached_lca = LCA(self.processes.steel, **self.redirect_kwargs)
        self.assertFalse(uncached_lca._get_cached_matrices(
            filename=uncached_lca.names.lci_matrices,
            expiration=uncached_lca.database_modified))
    
    def test_cache_directory_permissions(self):
        """Test cache directory permissions."""
        # TODO: This doesn't test for permissions for web server
        _cache_dir = settings.BRIGHTWAY_CACHE_PATH
        cache_writer = CacheInterface(_cache_dir, "test", [range(10),])
        del cache_writer
        cache_reader = CacheInterface(_cache_dir, "test")
        self.assertEquals(cache_reader.objs, [range(10),])

    def test_extra_string(self):
        """Test that the 'extra_string' parameter works."""
        lca_writer = LCA(self.processes.steel, extra_string="w00t",
            method=self.methods.human_health, **self.redirect_kwargs)
        lca_writer.lci()
        lca_writer.lcia()

        self.assertFalse(lca_writer.from_cache.solution_matrices)
        self.assertFalse(lca_writer.from_cache.technosphere_matrix)
        self.assertFalse(lca_writer.from_cache.biosphere_matrix)

        # Read from cache
        lca_reader = LCA(self.processes.steel, extra_string="w00t",
            method=self.methods.human_health, **self.redirect_kwargs)
        lca_reader.lci()
        lca_reader.lcia()

        self.assertTrue(lca_reader.from_cache.solution_matrices)
        self.assertFalse(hasattr(lca_reader.from_cache, "technosphere_matrix"))
        self.assertFalse(hasattr(lca_reader.from_cache, "biosphere_matrix"))
