# -*- coding: utf-8 -*
import os
import numpy
import datetime
import unittest
from django.conf import settings
from brightway.core.utils.file_utils import check_if_expired
from brightway.core.errors import CacheDirectoryError, CacheExpired, \
    CacheMissing
from functools import wraps

def catch_ioerror(func):
    @wraps(func)
    def this(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except IOError, err:
            return None
    return this


class SimpleCache(object):
    """
.. centered:: A fast interface for NumPy arrays.

**Inputs**

    #. cache_dir: The directory to store the cache.
    #. filename: The filename of the cache to save or load.
    #. obj: The NumPy array to cache. None if loading the cache.
    #. logging: A boolean or logging config dictionary.
    #. overwrite: Overwrite filename is it exists already.
    #. expiration: Don't load a cache older than expiration. Optional.

**Outputs**

When saving objects, SimpleCache does not return a value. When retrieving 
objects, SimpleCache returns False is loading was unsuccesful, or a list of 
objects if loading was successful.
    """
    def __init__(self, cache_dir, filename, obj=None, expiration=False):
        super(SimpleCache, self).__init__()
        self.filename = os.path.join(cache_dir, filename + ".array")

        if expiration and not isinstance(expiration, (datetime.datetime,
                datetime.timedelta)):
            raise TypeError, "expiration must be a python datetime." + \
                "datetime. You passed %s (type %s)" % (expiration, 
                type(expiration))
        self.expiration = expiration

        if obj is not None:
            self.save(obj)

    def save(self, obj):
        if not isinstance(obj, numpy.ndarray):
            raise TypeError, "SimpleCache.load_array obj must be a"+\
                " numpy array."
        with open(self.filename, "wb") as f:
            numpy.save(f, obj)
        return True

    @catch_ioerror
    def expired(self):
        if not self.expiration:
            return False
        if isinstance(self.expiration, datetime.timedelta):
            self.expiration = datetime.datetime.now() - self.expiration
        return check_if_expired(self.filename, self.expiration)

    @catch_ioerror
    def load(self):
        if self.expired():
            return None
        with open(self.filename, 'rb') as f:
            obj = numpy.load(f)
        return obj


class SimpleCacheTest(unittest.TestCase):
    def setUp(self):
        super(SimpleCacheTest, self).setUp()
        self.cache_dir = os.path.join(settings.BRIGHTWAY_ROOT_PATH, 
            'core/tests/cache/')

    def tearDown(self):
        self.clean_cache_directory()
        super(SimpleCacheTest, self).tearDown()

    def clean_cache_directory(self):
        def array_filter(x):
            if x.lower()[-4:] == ".array":
                return True
            return False

        files = os.listdir(self.cache_dir)
        files = filter(array_filter, files)
        for fyle in files:
            os.remove(os.path.join(self.cache_dir, fyle))

    def test_return_false_from_missing_cache(self):
        self.assertFalse(SimpleCache(self.cache_dir, "missing").load())

    def test_overwrite_existing_cache(self):
        a = numpy.ones((3,3))
        b = numpy.ones((3,3)) * 2
        Input = SimpleCache(self.cache_dir, "test_overwrite", obj=a)
        Input = SimpleCache(self.cache_dir, "test_overwrite", obj=b)
        Output = SimpleCache(self.cache_dir, "test_overwrite").load()
        self.assertTrue(numpy.allclose(Output, b))

    def test_numpy_array(self):
        a = numpy.ones((3,3))
        b = a * 10
        Input = SimpleCache(self.cache_dir, "test1", obj=a)
        Output = SimpleCache(self.cache_dir, "test1", None).load()
        self.assertTrue(numpy.allclose(a, Output))
        Input = SimpleCache(self.cache_dir, "test2", obj=b)
        Output = SimpleCache(self.cache_dir, "test2", None).load()
        self.assertTrue(numpy.allclose(b, Output))

    def test_struct_array(self):
        g = numpy.zeros((5,), dtype=[
            ('foo', 'u4'), ('baloney', 'b1'), ('bar', 'f4')])
        g[0] = (12, True, 3.14)
        Input = SimpleCache(self.cache_dir, "test_struct", obj=g)
        h = SimpleCache(self.cache_dir, "test_struct").load()
        self.assertEqual(g.shape, h.shape)
        self.assertTrue(numpy.allclose(g['foo'], h['foo']))
        self.assertTrue(numpy.allclose(g['baloney'], h['baloney']))
        self.assertTrue(numpy.allclose(g['bar'], h['bar']))

    def test_array_dtype(self):
        """Test array saving with non-default dtype"""
        a = numpy.ones((3,3), dtype='int16')
        Input = SimpleCache(self.cache_dir, "test1", obj=a)
        Output = SimpleCache(self.cache_dir, "test1", None).load()
        self.assertTrue(numpy.allclose(a, Output))

    def test_arbitrary_object_fails(self):
        class Foo(object):
            """A dummy class used for tests"""
            def __init__(self, arg):
                self.arg = arg

        self.assertRaises(TypeError, SimpleCache, self.cache_dir, 
            "test_object", Foo("this"))

    def test_invalid_path(self):
        self.assertRaises(CacheDirectoryError, SimpleCache, 
            "/foo/not/really", "false", ())

    def test_invalid_expiration(self):
        self.assertRaises(TypeError, SimpleCache.__init__, self.cache_dir,
            "nothing", {'obj': numpy.ones((2,)), 'expiration': '2008-10-27'})
        self.assertRaises(TypeError, SimpleCache.__init__, self.cache_dir,
            "nothing", {'obj': numpy.ones((2,)), 'expiration': (
            2010, 2, 16, 10, 6, 20, 107421)})

