import amara
import codecs
from xml.dom import XML_NAMESPACE as XML_NS
try:
    import babel
except ImportError:
    babel = False
from django.contrib.gis.db.models.query import GeoQuerySet
from django.db.models.query import QuerySet
from django.core.exceptions import ObjectDoesNotExist
from django.utils import simplejson
from brightway.apps.geo.models import GeographyDatabase, Geography
from brightway.io.file_sender import FileSender
from brightway.apps.geo.geometry_choices import GEOGRAPHY_COLUMN_DICT
from brightway.apps.geo.utils import check_coordinates_8_digits
from brightway.core.sql import sql

IS_VALID = """select st_isvalid(g.poly_shape) 
    from geo_geography g 
    where g.id = %s"""
IS_VALID_REASON = """select st_isvalidreason(g.poly_shape) 
    from geo_geography g 
    where g.id = %s"""

class GeographyDatabaseEcospold2Exporter(object):
    '''
Export GeographyDatabase(s) or a selection of Geographies into EcoSpold2 Format
    '''
    def __init__(self, arg, name=None, write=True, use_file_sender=True, 
            languages=('es', 'ja', 'fr', 'de', 'zh'), format="xml", 
            check_geos=False):
        self.name = name
        self.use_file_sender = use_file_sender
        self.format = format

        if babel and self.format == "xml":
            self.languages = languages
            self.language_dictionary = {}
            for language in self.languages:
                # TODO: Can split string based on en-US, etc.
                locale = babel.Locale(language)
                self.language_dictionary[language] = locale.territories
        else:
            self.languages = ()

        self.optional_attributes = {
            u'numeric': u'UNCode',
            u'longitude': u'longitude',
            u'latitude': u'latitude',
            u'alpha2': u'ISOTwoLetterCode',
            u'alpha3': u'ISOThreeLetterCode'
            }
        self.extra_attributes = {
            u'region': u'UNRegionCode',
            u'subregion': u'UNSubregionCode'
            }

        if not isinstance(arg, (QuerySet, GeoQuerySet, Geography, 
                GeographyDatabase)):
            raise AttributeError
        elif isinstance(arg, GeographyDatabase):
            self.geographies = Geography.objects.filter(database=arg)
        elif isinstance(arg, Geography):
            self.geographies = Geography.objects.filter(id=arg.id)
        elif arg.model == Geography:
            self.geographies = arg
        elif arg.model == GeographyDatabase:
            self.geographies = Geography.objects.filter(database__in=arg)
        else:
            raise AttributeError

        if check_geos:
            for geo in self.geographies:
                cursor = sql(IS_VALID, params=(geo.id,))
                if cursor.fetchone()[0] == True:
                    continue
                else:
                    print "Problem with geography %s" % geo
                    cursor = sql(IS_VALID_REASON, params=(geo.id,))
                    print "\t", cursor.fetchone()[0]            
            print "Checking conversion to 8 digits of precision:"
            check_coordinates_8_digits(self.geographies)

        self.create_document()
        if write:
            self.write_document()

    def create_document(self):
        if self.format == "xml":
            self.create_xml_document()
        elif self.format == "json":
            self.create_dict_document()
        else:
            raise AttributeError

    def create_dict_document(self):
        self.doc = []
        for obj in self.geographies:
            if not obj.geography:
                raise AttributeError
            obj_dict = {'uuid': obj.uuid,
                'geography': obj.geography,
                'description': obj.description,
                'pop_total': obj.pop_total,
                'area': obj.area,
                'pop_density_km': obj.pop_density_km,
                'database': obj.database.database,
                'code': obj.imported_database_code,
                }
            for attr in self.optional_attributes:
                value = getattr(obj, attr, None)
                if value:
                    obj_dict[attr] = value
            ewkt = getattr(obj.shape, "ewkt", None)
            if ewkt:
                obj_dict["ewkt"] = ewkt
            for attr in self.extra_attributes:
                try:
                    value = obj.attributes[attr]
                    if value:
                        obj_dict[attr] = value
                except:
                    pass
            self.doc.append(obj_dict)

    def create_xml_document(self):
        doc_root = amara.create_document()
        doc = doc_root.xml_create_element(u'validGeographies', 
            attributes={u"majorRelease":u"3", u"minorRelease":u"0",
            u'xmlns': u"http://www.EcoInvent.org/EcoSpold02"})
        for geography in self.geographies:
            print "Processing %s" % geography
            if not geography.geography:
                raise AttributeError
            element = doc.xml_create_element(u'geography',
                # content=geography.geography,
                attributes={ # These attributes are required
                    u'id': unicode(geography.uuid)})
            for attr in self.optional_attributes:
                if getattr(geography, attr, None):
                    element.xml_set_attribute(self.optional_attributes[attr], 
                        unicode(getattr(geography, attr)))
            for attr in self.extra_attributes:
                try:
                    value = unicode(geography.attributes[attr])
                    if value:
                        element.xml_set_attribute(self.extra_attributes[attr], 
                            value)
                except:
                    pass

            # Name
            name_element = doc.xml_create_element(u"name", 
                content=unicode(geography.geography), 
                attributes={(u'xml:lang', XML_NS): u'en'})
            element.xml_append(name_element)
            for language in self.languages:
                code = geography.alpha2
                try:
                    region = "%03d" % geography.attributes["region"]
                except ObjectDoesNotExist:
                    region = None
                try:
                    subregion = "%03d" % geography.attributes["subregion"]
                except ObjectDoesNotExist:
                    subregion = None
                
                for obj in (code, region, subregion):
                    if obj in self.language_dictionary[language]:
                        name_element = doc.xml_create_element(u"name", 
                            content=unicode(self.language_dictionary[ 
                                language][obj]), 
                            attributes={(u'xml:lang', XML_NS): unicode(
                                language)})
                        element.xml_append(name_element)
                        break

            # Shortname
            if geography.code:
                shortname = unicode(geography.code)
            else:
                shortname = unicode(geography.geography)
            short_name_element = doc.xml_create_element(u"shortname", 
                content=shortname, 
                attributes={(u'xml:lang', XML_NS): u'en'})
            element.xml_append(short_name_element)

            # Comment
            if geography.description:
                comment_element = doc.xml_create_element(u"comment")
                text_element = doc.xml_create_element(u"text",
                    content=unicode(geography.description),
                    attributes={u"index":u"0", (u'xml:lang', XML_NS): u'en'})
                comment_element.xml_append(text_element)
                element.xml_append(comment_element)
            if hasattr(geography, "shape") and hasattr(geography.shape, 
                    "kml"):
                kml_element = doc.xml_create_element(u"kml",) 
                # Remove for time being...
                #    attributes={u"xmlns": u"http://www.opengis.net/kml/2.2"})

                # TODO: Check if simplified shape should be exported
                kml_text = geography.kml
                if kml_text:
                    # xml_append_fragment doesn't work because it doesn't have 
                    # beginning XML declaration
                    kml_parsed = amara.parse(kml_text)
                    kml_placemark_element = doc.xml_create_element( 
                        u"Placemark")
                    kml_placemark_element.xml_append(kml_parsed.xml_children[0 
                        ])
                    kml_document_element = doc.xml_create_element(u"Document")
                    kml_name_element = doc.xml_create_element(u"name", 
                        content=geography.geography)
                    kml_document_element.xml_append(kml_name_element)
                    kml_document_element.xml_append(kml_placemark_element)
                    kml_element.xml_append(kml_document_element)
                element.xml_append(kml_element)
            doc.xml_append(element)
        doc_root.xml_append(doc)
        self.doc = doc_root

    def write_document(self):
        if self.name:
            filename = self.name
        else:
            filename = "Geographies"
        if self.use_file_sender:
            self.file_sender = FileSender(filename=filename, 
                content_type=self.format)
            if self.format == "xml":
                self.file_sender.file.write(self.doc.xml())
            elif self.format == "json":
                self.file_sender.file.write(simplejson.dumps(self.doc, 
                    indent=2))
            return True
        else:
            # Write to normal file
            # Don't need codecs because xml() writes encoded utf-8
            f = open(filename + "." + self.format, "w")
            if self.format == "xml":
                f.write(self.doc.xml())
            elif self.format == "json":
                f.write(simplejson.dumps(self.doc, indent=2))
            f.close()
            return filename
