import csv
import datetime
import random
# from cStringIO import StringIO 
from django.db.models.base import ModelBase
from brightway.core.array_proxies import ArrayProxy
from brightway.io.file_sender import FileSender

# Optional Excel writers
try:
    import xlwt
except ImportError:
    xlwt = False
try:
    import pyexcelerator
except ImportError:
    pyexcelerator = False

class MatrixProxyExporter(object):
    """
Class for exporting sparse or dense matrix proxies to various formats.
    """
    def __init__(self, proxy, row_model, column_model=None, format="csv", 
            filename=None, export=True, in_memory=False, zipped=False, 
            use_file_sender=True, **kwargs):
        super(MatrixProxyExporter, self).__init__()
        if not isinstance(proxy, ArrayProxy):
            raise AttributeError, "Must pass a matrix proxy to MatrixProxyExporter"
        self.data = proxy.data
        # if hasattr(self.data, "todense"):
        #     self.data = self.data.todense()
        self.row_dict = proxy.row_dict
        
        if hasattr(proxy, "col_dict"):
            self.col_dict = proxy.col_dict
        else:
            self.col_dict = None
        self.format = format
        self.in_memory = in_memory
        self.zipped = zipped
        self.use_file_sender = use_file_sender

        if filename:
            self.filename = filename
        else:
            # TODO
            date = datetime.datetime.today().isoformat()
            self.filename = "matrix_proxy_" + date.replace("-", "_") + "_" + str(int(random.random()*1000))
        self.kwargs = kwargs

        self.parse_models(row_model, column_model)
        self.build_reverse_row_dict()
        self.build_reverse_column_dict()

        if export:
            self.export()

    def parse_models(self, row_model, column_model):
        '''Passed values can be Models, string, or None'''
        if isinstance(row_model, ModelBase):
            self.row_model = row_model
            self.row_string = None
        elif row_model:
            self.row_model = None
            self.row_string = row_model
        else:
            self.row_model = None
            self.row_string = None
        if isinstance(column_model, ModelBase):
            self.column_model = column_model
            self.column_string = None
        elif column_model:
            self.column_model = None
            self.column_string = column_model
        else:
            self.column_model = None
            self.column_string = None

    def _label(self, key, model, string):
        if model:
            assert isinstance(key, int)
            return model.objects.get(id=key).__unicode__()
        elif string:
            return "%s %s" % (string, key)
        else:
            return unicode(key)

    def build_reverse_row_dict(self):
        '''Build reverse dictionary that maps rows to model strings'''
        labeled_rows = [self._label(x, self.row_model, self.row_string) 
            for x in self.row_dict.keys()]
        self.reverse_row_dict = dict(zip(self.row_dict.values(),
            labeled_rows))

    def build_reverse_column_dict(self):
        '''Build reverse dictionary that maps columns to model strings'''
        if not self.col_dict:
            self.reverse_column_dict = None
            return
        labeled_columns = [self._label(x, self.column_model, self.column_string) 
            for x in self.col_dict.keys()]
        self.reverse_column_dict = dict(zip(self.col_dict.values(),
            labeled_columns))

    def export(self):
        """Choose correct exporter based on format"""
        if self.format.lower() == "csv":
            self._export_plain(separator=",")
        elif self.format.lower() == "tsv":
            self._export_plain(separator="\t")
        elif self.format.lower() == "excel":
            # self._export_excel()
            pass
        elif self.format.lower() == "gdocs":
            pass
        elif self.format.lower() == "html":
            pass
        else:
            raise AttributeError, "Format not recognized"
        # return self.file

    def _export_plain(self, separator=","):
        """Export to plain text using standard library CSV"""
        if self.use_file_sender:
            self.file_sender = FileSender(filename=self.filename, 
                zip_=self.zipped, in_memory=self.in_memory, 
                content_type=self.format.lower())
            file_obj = self.file_sender.file
        else:
            # Write to normal file
            file_obj = open(self.filename, "w")

        if "dialect" in self.kwargs:
            w = csv.writer(file_obj, **self.kwargs)
        elif separator == ",":
            w = csv.writer(file_obj, dialect=csv.excel)
        elif separator == "\t":
            w = csv.writer(file_obj, dialect=csv.excel_tab)
        else:
            raise

        # Header
        if self.column_model:
            w.writerow(self.utf8_encoder([u"",] + [self.reverse_column_dict[x] \
                for x in range(self.data.shape[1])]))
        for x in range(self.data.shape[0]):
            try:
                row = self.data[x,:].todense()
            except AttributeError:
                row = self.data[x,:]
            w.writerow(self.utf8_encoder([self.reverse_row_dict[x],] + 
                row.tolist()[0]))

        if self.use_file_sender:
            self.file_sender.send()
        else:
            file_obj.close()
            return self.filename

    def utf8_encoder(self, data):
        def _encode(element):
            if hasattr(element, "encode"):
                return element.encode("utf-8")
            else:
                return element
        return [_encode(element) for element in data]