import time
import mimetypes
import tempfile
import zipfile
import codecs
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from django.http import HttpResponse 
from django.core.servers.basehttp import FileWrapper

CONTENT_TYPE_DICTIONARY = {
    'zip': 'application/zip',
    'excel': 'application/vnd.ms-excel',
    'csv': 'text/csv; charset=utf-8',
    'pdf': 'application/pdf',
    'tsv': 'text/csv; charset=utf-8',
    'word': 'application/msword',
    'text': 'text/plain',
    'xml': 'application/xml',
    'json': 'application/json',
    'html': 'text/html',
    'kml': 'application/vnd.google-earth.kml+xml',
    'kmz': 'application/vnd.google-earth.kmz',
}

FILE_SUFFIX_DICTIONARY = {
    'word': 'doc',
    'excel': 'xls',
}

class FileSender(object):
    """
Write a file-like object to a response, with correct encoding, etc. 
Instantiate the class, write data to FileSender.file, then return 
FileSender.send(). Use utility function FileSender.add_zip_file(filename, 
data) to add separate files to a zip archive.
    """
    def __init__(self, filename=None, file_=None, zip_=False, in_memory=False, 
            content_type="zip"):
        self.filename = filename
        self.content_type = content_type
        self.in_memory = in_memory
        self.file = file_
        self.zip = zip_
        self.create_file()

    def create_file(self):
        if self.file:
            # User-specified file object
            return
        if self.in_memory:
            # Use StringIO
            buffer = StringIO()
        else:
            buffer = tempfile.TemporaryFile()

        if self.zip:
            self._original_buffer = buffer
            self.file = zipfile.ZipFile(self._original_buffer, "w")
        else:
            self.file = buffer

    def add_zip_file(self, filename, data):
        if not self.zip:
            raise AttributeError, "Can only add files to zip archive."
        info = zipfile.ZipInfo(str(filename))
        info.date_time = time.localtime(time.time())[:6]
        info.compress_type = zipfile.ZIP_DEFLATED
        info.external_attr = 0666 << 16L
        self.file.writestr(info, data.encode('utf-8'))

    def parse_content_type(self):
        try:
            self.content_type = CONTENT_TYPE_DICTIONARY[self.content_type]
        except KeyError:
            pass
        if not self.content_type:
            # Need to write to a file, and get a name
            # This won't work as it is
            # self.content_type = self.guess_mimetype()
            content_type = 'plain'

    def get_data(self):
        # if 'utf-8' in self.content_type:
        #     self.response.write(codecs.BOM_UTF8)
        try:
            self.file.seek(0)
        except AttributeError: # Zipfile
            pass
        if self.in_memory:
            data = self.file.getvalue()
        else:
            data = self.file.read()
        return data
    
    def guess_mimetype(self):
        if self.in_memory:
            data = self.file.getvalue()
            mimetype, encoding = mimetypes.guess_type(data)
        else:
            mimetype, encoding = mimetypes.guess_type(self.file)
        return mimetype
    
    def parse_filename(self):
        try:
            self.filename += "." + FILE_SUFFIX_DICTIONARY[self.content_type]
        except KeyError:
            self.filename += "." + self.content_type
    
    def send(self):
        self.parse_filename()
        if self.zip:
            # Close zipfile and access original file object
            self.file.close()
            self.file = self._original_buffer
        self.file.seek(0)
        self.parse_content_type()
        # self.response = HttpResponse(FileWrapper(self.file), 
        #    content_type=self.content_type)
        self.response = HttpResponse(self.get_data(), 
            content_type=self.content_type)
        self.response['Content-Disposition'] = 'attachment; filename=%s' % \
            self.filename.encode('utf-8')
        # Set file length
        try:
            self.response['Content-Length'] = self.file.tell()
        except:
            pass
        return self.response
