import os
from lxml import objectify
import logging as log
from django.core.exceptions import ObjectDoesNotExist
from brightway.io.imprt.ia_base import ImpactAssessmentBase
from brightway.apps.geo.models import Geography
from brightway.apps.impact_assessment.models import Method
from brightway.apps.inventory.models import Process, Database, Category
from brightway.apps.unit.models import Unit, UnitType
from brightway.apps.geo import WORLD

class EcospoldImpactAssessmentImporter(ImpactAssessmentBase):
    """Import impact assessment methods and weightings from ecospold XML format"""
    def __init__(self, filename, emissions_database, description=None, *args, 
            **kwargs):
        super(EcospoldImpactAssessmentImporter, self).__init__(*args,**kwargs)
        self.filename = filename
        self.description = description
        assert isinstance(emissions_database, Database)
        self.emissions_database = emissions_database
        # Note that this is only used for the first root method found in 
        # the file
        self.root = objectify.parse(open(self.filename)).getroot()

        self.start_transaction()
        try:
            # Only add root method if providing special description
            if self.description:
                self.add_root_method(self.root.dataset, first=True)
            for dataset in self.root.iterchildren():
                self.add_method(dataset)
            self.commit_transaction()
        except:
            self.rollback_transaction()
            raise
        finally:
            self.leave_transaction()

    def add_root_method(self, ds, first=False):
        method_name = ds.metaInformation.\
            processInformation.referenceFunction.get("category")

        if first:
            description = self.description or ""
        else:
            description = ""

        log.info(u"Starting import of LCIA method %s".encode('utf-8'
            ) % method_name)
        unit = self.get_unit(ds.metaInformation.\
            processInformation.referenceFunction.get("unit"))
        log.info(u"Using unit %s".encode('utf-8') % unit)
        root_method = self.create_method(method_name, 
            description=description, 
            other_info=str(self.filename),
            unit=unit)
        log.info(u"Created root method %s".encode('utf-8') % root_method)
        if first:
            self.root_method = root_method

    def add_method(self, ds):
        unit = self.get_unit(ds.metaInformation.processInformation.\
            referenceFunction.get("unit"))

        # One file can have multiple root methods
        name = ds.metaInformation.processInformation.referenceFunction.get( 
            "category")
        try:
            root_method = Method.objects.get(method=name, parent__isnull=True)
        except ObjectDoesNotExist:
            root_method = Method.objects.create(method=name, parent=None,
                unit=unit)
            log.info(u"Creating root method %s".encode('utf-8' 
                ) % root_method)

        name = ds.metaInformation.processInformation.referenceFunction.get( 
            "subCategory")
        try:
            submethod = Method.objects.get(method=name, parent=root_method)
        except ObjectDoesNotExist:
            submethod = Method.objects.create(method=name, parent=root_method, 
                unit=unit)
            log.info(u"Creating intermediate method %s".encode('utf-8' 
                ) % submethod)

        name = ds.metaInformation.processInformation.referenceFunction.get(
            "name")
        log.info(u"Creating LCIA method %s".encode('utf-8') % name)
        description = ds.metaInformation.processInformation.\
            referenceFunction.get("generalComment")
        description = description or ""
        if Method.objects.filter(method=name, parent=submethod).exists():
            log.error(u"Found duplicate method: %s %s".encode('utf-8') % ( 
                name, submethod))
            return
        method = self.create_method(method=name, 
            parent=submethod, 
            description=description, 
            unit=unit)
        for weighting in ds.flowData.iterchildren():
            self.add_weighting(weighting, method)

    def add_missing_biosphere_flow(self, wt, method):
        WARN_TEXT = u"""Adding missing biosphere flow:
            \tname: %s
            \tcategory: %s-%s
            \tunit: %s"""
        log.warn(WARN_TEXT.encode(
            'utf-8') % (wt.get("name"), wt.get("category"), wt.get( 
            "subCategory"), wt.get("unit")))
        unit = self.get_unit(wt.get("unit"))
        if wt.get("subCategory") == "unspecified":
            category = Category.objects.get_or_create(
                category=wt.get("category"), 
                database=self.emissions_database)[0]
        else:
            category = Category.objects.get_or_create(
                category=wt.get("subCategory"), 
                parent__category=wt.get("category"), 
                database=self.emissions_database)[0]
        process = Process.objects.create(creator=self.user, 
            process=wt.get("name"), 
            category=category, 
            unit=unit, 
            process_type=2, 
            database=self.emissions_database, 
            cas_number=wt.get("CASNumber") or '', 
            chemical_formula=wt.get("formula") or '', 
            imported_database_code=wt.get("normal") or '', 
            geography=WORLD())
        self.add_attributes(process, wt)
        return self.create_weighting(method, process, 
            float(wt.get("meanValue")))

    def add_weighting(self, wt, method):
        try:
            process = Process.objects.filter(
                database=self.emissions_database,
                category__database=self.emissions_database,
                process=wt.get("name"), 
                process_type__in=(1,2))
            if wt.get("subCategory") == "unspecified":
                process = process.filter(category__category=wt.get( 
                    "category")).get()
            else:
                process = process.filter( 
                    category__category=wt.get("subCategory"),
                    category__parent__category=wt.get("category")).get()
            self.add_attributes(process, wt)
            self.create_weighting(method, process, float(wt.get("meanValue")))
        except ObjectDoesNotExist:
            return self.add_missing_biosphere_flow(wt, method)

    def add_attributes(self, process, weighting):
        """Add CAS number, chemical formula, and imported database code"""
        changed = False
        for attr1, attr2 in (("cas_number", "CASNumber"), 
                ("imported_database_code", "number"), ("chemical_formula", 
                "formula")):
            if not getattr(process, attr1, None) and weighting.get(attr2):
                setattr(process, attr1, weighting.get(attr2))
                changed = True
        if changed:
            process.save(skip_datetime_update=True)

    def get_unit(self, unit):
        try:
            unit = Unit.objects.get(unit=unit)
        except ObjectDoesNotExist:
            try:
                unit = Unit.objects.get(abbreviation=unit)
            except ObjectDoesNotExist:
                unit = Unit.objects.create(unit=unit, abbreviation=unit, 
                    unittype=UnitType.objects.get_or_create( 
                    unittype="Unknown")[0])
        return unit


def import_dir(dirpath):
    db = Database.objects.get(database='biosphere flows')
    for filename in filter(lambda x: x.lower()[-4:] == ".xml", 
            os.listdir(dirpath)):
        filepath = os.path.join(dirpath, filename)
        print "Working on %s" % filepath
        EcospoldImpactAssessmentImporter(filepath, db)

