# -*- coding: utf-8 -*
import amara
import datetime
import os
import vobject
import math
import logging as log
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.db.models.query import QuerySet
from django.utils import simplejson
from django.conf import settings
from brightway.io.imprt.import_base import ImportBase
from brightway.apps.geo.models import Geography, GeographyDatabase
from brightway.apps.geo import WORLD
from brightway.apps.inventory.models import Process, Exchange, Database, \
    Category
from brightway.apps.unit.models import Unit, UnitType
from brightway.apps.data_source.models import Person
from brightway.io.errors import *
from brightway.io.imprt.process_finder import ProcessFinder
from brightway.core.utils.safe_hash import generate_safe_hash
from brightway.io.data.ecoinvent_mapping import \
    ECOINVENT_13_TO_2_PROCESS_MAPPING, ECOINVENT_13_TO_2_CATEGORY_MAPPING
from brightway.core import Container
from stats_toolkit.distributions import *
try:
    import progressbar as pb
except ImportError:
    pb = None

"""Design decisions:

Import scenarios:
1. Import of fresh database without meta information.
    - I don't think we want to support this. If this is an import of the Ecoinvent, the EcoinventImporter will have access to generic metainformation.

2. Import of fresh database with meta information.
    - Must follow expected directory structure (e.g. meta and datasets)

3. Update of database information.
    - Currently only supported for exchanges.

4. Import of fresh database built on top of one or more parent databases.
    - This is tricky if:
        * Names don't match exactly
        * Category names overlap in parent databases
        * Knowing where to insert new emissions

Major changes from Ecospold to Brightway data storage:

1. Process type information moves from categories to processes

Functions::

Geography databases:

Geographies are taken from either:

- A 'Default world' database
- A parent geography database
- Created in a new geography database

# Need global geography


Here is how broken EcoSpold is: Everything is referenced by ID numbers, because process names are not unique. However, ID numbers are ignored by import and export software, so can't be trusted at all. Consider the following possibilities:

1. Every process in a database has ID number 1

Culprit: US LCI database

2. Exchange number at not as a reference to an input, but as a sequential counter

Culprit: The Excel export macro, which is widely used
Culprit: US LCI database
Culprit: Simapro

3. BONUS: Dataset refers to standard meta-data files, e.g. "Categories.xml", but actually uses own, undefined meta-data. In other words, the cake is a lie, and the correct meta-data can only be (incompletely) guessed.

Culprit: Simapro
Culprit: US LCI database

The end result is lots of code tripping over itself, layer upon layer of try and except, and lots of extra add_missing_foo functions.

Importing and Exporting

Although there are public formats for distributing life cycle infomation, none of them are well designed for the exchange of data. Currently, Brightway only supports importing and exporting data using the EcoSpold data format. LCI data in other formats can be converted to EcoSpold using the OpenLCA converter. Data exchange between Brightway instances uses a custom format which fits the Brightway data schema better, and contains extra attributes that make interchange trivial.

How to import

How to export

Details of EcoSpold importer

There are several classes for importing different types of EcoSpold data, but all derive from Ecospold1Importer.

Ecospold1Importer
-----------------

Most common use cases should reference the specialized classes described below. This base class can import a complete database, or a database built on top of another database. 

Parameters
``````````

path:   Path to data files.
database:   String or Database instance
parent_databases:   QuerySet or None
add_geographies:    Boolean
add_units:  Boolean
add_categories: Boolean
default_local_language: String. Default is "de".
user:   String or User instance
behaviour:  String. Default is "update_if_provided".

TODO:

Clean up docs
Add possibility to use enhanced meta-files
Add people and verificaiton
Tests
"""

class Ecospold1Importer(ImportBase):
    """
Generic Ecospold import class. Not designed to work with any specific implementation, or reference any specific metadata.

Can have 4 different behaviours:
1. "ignore": Processes and Exchanges which already exist are ignored.
2. "update": Processes and Exchanges which already exist have the attributes specified in the XML file updated.
3. "update_if_provided": As 2, but won't change attributes if given None or an empty string.
4. "overwrite": Processes and Exchanges which already exist are overwritten, and given only with the values provided.
Note that this can raise validation errors if required attributes are not provided.
    """
    def __init__(self, path, database, parent_databases=None, 
            add_geographies=False, add_units=False, add_categories=True, 
            add_processes=True, add_exchanges=True,
            default_local_language="de", geography_databases=None, 
            commit=False, is_1_to_2=False, process_mapping=None, 
            category_mapping=None, *args, **kwargs):
        super(Ecospold1Importer, self).__init__(*args, **kwargs)
        # Strip all white space / "text_wrapper" junk.
        self.amara_rules = [amara.binderytools.ws_strip_element_rule(u'*'),]
        self.default_local_language = default_local_language

        self.setup_cache()

        self.start_transaction()
        self.check_path(path)
        self.get_databases(parent_databases, database)
        self.global_geography = WORLD(reset=True)
        self.get_geography_databases(geography_databases)
        process_mapping = process_mapping or ( 
            ECOINVENT_13_TO_2_PROCESS_MAPPING if is_1_to_2 else {})
        category_mapping = category_mapping or ( 
            ECOINVENT_13_TO_2_CATEGORY_MAPPING if is_1_to_2 else {})
        self.process_finder = ProcessFinder(
            database=self.database, 
            parent_databases=self.parent_databases, 
            process_name_mapping=process_mapping, 
            category_mapping=category_mapping)

        if add_categories:
            self.add_categories()
        if add_geographies:
            self.add_geographies()
        if add_units:
            self.add_units()
        if add_processes:
            self.add_processes()
        if add_exchanges:
            self.add_exchanges()
        
        # Update modified datetime
        self.database.save()
        
        if commit:
            self.commit_transaction()
        else:
            self.rollback_transaction()

    ### Setup ###
    def setup_cache(self):
        """Initiate cache"""
        self.cache = Container()
        self.cache.unit = {}
        self.cache.geography = {}
        self.cache.category = {}
    
    ### Setup ###
    def check_path(self, path):
        """Make sure passed path has """
        if not isinstance(path, (unicode, str)):
            raise ImportSetupError, "path must be filepath (string): %s" % path
        self.path = path
        try:
            files = os.listdir(self.path)
        except OSError:
            raise ImportSetupError, "Provided path does not exist"
        found_test = False
        index = 0
        # Iteratively search over candidate files until find metadata
        while not found_test:
            index += 1
            filename = files[index]
            if not self.xmlfilter(filename):
                continue
            parsedfile = amara.parse(os.path.join(self.path, filename), 
                rules=self.amara_rules)
            if not hasattr(parsedfile, "ecoSpold"):
                continue
            try:
                dataset = parsedfile.ecoSpold.childNodes[0]
                units_filename = dataset.validUnits
                category_filename = dataset.validCategories
                locations_filename = dataset.validRegionalCodes
                # Look for metafiles in 'meta' subdirectory (default for Brightway exports)
                if "meta" in files:
                    metafiles = os.listdir(os.path.join(self.path, "meta"))
                    if units_filename in metafiles:
                        self.units_file = os.path.join(self.path, 'meta',
                            units_filename)
                    if category_filename in metafiles:
                        self.category_file = os.path.join(self.path, 'meta',
                            category_filename)
                    if locations_filename in metafiles:
                        self.locations_file = os.path.join(self.path, 'meta',
                            locations_filename)
                # Look for metafiles in data directory
                if units_filename in files:
                    self.units_file = os.path.join(self.path, units_filename)
                if category_filename in files:
                    self.category_file = os.path.join(self.path, 
                        category_filename)
                if locations_filename in files:
                    self.locations_file = os.path.join(self.path, 
                        locations_filename)
                found_test = True
            except:
                continue
        # Default metafiles
        if not (hasattr(self, "units_filename") and \
                os.path.exists(self.units_filename)):
            self.units_filename = os.path.join(settings.BRIGHTWAY_ROOT_PATH, 
                "core/io/data/inventory/ecospold1/Units.xml")
        if not (hasattr(self, "category_filename") and \
                os.path.exists(self.category_filename)):
            self.category_filename = os.path.join(settings.BRIGHTWAY_ROOT_PATH, 
                "core/io/data/inventory/ecospold1/Categories.xml")
        if not (hasattr(self, "locations_filename") and \
                os.path.exists(self.locations_filename)):
            self.locations_filename = os.path.join(settings.BRIGHTWAY_ROOT_PATH, 
                "core/io/data/inventory/ecospold1/RegionalCodes.xml")
        self.dataset_files = filter(self.xmlfilter, os.listdir(self.path))
    
    def get_databases(self, parent_databases, database):
        """Correctly format, and create if necessary, self.parent_databases and self.database."""
        # TODO: If database already exists, then the existing links should be used. The importer is not the place to create these relationships. parent_databases should only be used in the case where a database is being created.
        if isinstance(parent_databases, Database):
            self.parent_databases = Database.objects.filter( 
                id=parent_databases.id)
        elif isinstance(parent_databases, QuerySet):
            self.parent_databases = parent_databases
        elif not parent_databases:
            self.parent_databases = Database.objects.none()
        elif parent_databases:
            raise ImportSetupError, "Can't parse value for 'parent_" + \
                "database': %s" % database

        if isinstance(database, Database):
            self.database = database
        elif isinstance(database, (unicode, str)):
            try:
                self.database = Database.objects.get(database=database)
            except ObjectDoesNotExist:
                log.warning(u"creating database %s".encode('utf-8') % \
                    database)
                self.database = Database.objects.create(
                    database=database, 
                    creator=self.user)
                # TODO: Redo with DAG
                # if self.parent_databases:
                #     for parent_database in self.parent_databases:
                #         self.database.parents.add(parent_database)
        else:
            raise ImportSetupError, u"Can't parse value for 'database': %s" \
                % database
    
    def get_geography_databases(self, geography_databases):
        """Define set of allowed geography databases."""
        if not geography_databases:
            self.geography_databases = GeographyDatabase.objects.filter(
                database__in=("UN Regions and Subregions", "Continents", 
                    "Political entities", "Custom ecoinvent geographies", 
                    "NERC Regions",  "European electricity grid", "World"))
        elif isinstance(geography_databases, GeographyDatabase):
            self.geography_databases = GeographyDatabase.objects.filter( 
                id=geography_databases.id)
        elif isinstance(geography_databases, QuerySet):
            self.geography_databases = geography_databases
        else:
            raise ImportSetupError, u"Can't parse value for " + \
                u"geography_databases: %s" % geography_databases
    
    ### Generic Add/Get Functions ###
    def get_or_create_geography(self, code, name=None):
        """Lookup geography based on a number of possible parameter combinations. ``code`` is assumed to be unique within a set of import files."""
        if code in self.cache.geography:
            return self.cache.geography[code]

        possibilities = [
            {"code": True},
            {"name": True},
            {"code": True, "name": True},
            {"alpha2": True, "name": True},
            {"alpha2": True}]
        if not name:
            possibilities = [
                {"code": True},
                {"alpha2": True}]
        for combin in possibilities:
            filters = {"database__in": self.geography_databases}
            if combin.get("code", None):
                filters["imported_database_code__iexact"] = code
            if combin.get("name", None):
                filters["geography__iexact"] = name
            if combin.get("alpha2", None):
                filters["alpha2__iexact"] = code
            try:
                return Geography.objects.get(**filters)
            except ObjectDoesNotExist, MultipleObjectsReturned:
                pass
        
        log.info(u"create geography %s, name %s".encode('utf-8') % (code, 
            name))
        if not hasattr(self, "created_geography_database"):
            log.info(u"create geography %s, name %s".encode('utf-8') % (code, 
                name))
            self.created_geography_database = \
                GeographyDatabase.objects.create(
                database=u"Created geographies for %s" % self.database)
        geography = Geography.objects.create(
            database=self.created_geography_database,
            imported_database_code=code,
            geography=name if name else code)
        self.cache.geography[code] = geography
        return geography
    
    def get_or_create_category(self, category=None, 
            parent=None, database=None, 
            other_info=''):
        categoryhash = u"%s-%s-%s" % (category, parent, database)
        if category in self.cache.category:
            return self.cache.category[categoryhash]
        elif isinstance(category, Category):
            self.cache.category[categoryhash] = category
            return category

        if not database:
            database = self.database

        if not category:
            # TODO: Should create subcategories if parent is given.
            try:
                return self.unknown_category
            except AttributeError:
                self.unknown_category = self.get_or_create_category(
                    category="Unknown", database=database)
                return self.unknown_category

        if not parent:
            # Top level category
            try:
                category = Category.objects.get(
                    parent__isnull=True,
                    category=category,
                    database=database)
                self.cache.category[categoryhash] = category
                return category
            except ObjectDoesNotExist:
                try:
                    category = Category.objects.get(
                        database__in=self.parent_databases,
                        parent__isnull=True, 
                        category=category)
                    self.cache.category[categoryhash] = category
                    return category
                except ObjectDoesNotExist:
                    category = Category.objects.create(
                        category=category,
                        database=self.database,
                        other_info=other_info)
                    self.cache.category[categoryhash] = category
                    log.info(u"creating category '%s' root category".encode('utf-8') % category)
                    return category
        else:
            # Secondary or lower-level category
            if isinstance(parent, Category):
                parent = parent.category
            try:
                category = Category.objects.get(
                    parent__category=parent, 
                    category=category,
                    database=database)
                self.cache.category[categoryhash] = category
                return category
            except ObjectDoesNotExist:
                try:
                    category = Category.objects.get(
                        database__in=self.parent_databases,
                        parent__category=parent, 
                        category=category)
                    self.cache.category[categoryhash] = category
                    return category
                except ObjectDoesNotExist:
                    parent = Category.objects.get_or_create(
                        category=parent,
                        database=self.database)[0]
                    if category:
                        category = Category.objects.get_or_create(
                            category=category,
                            database=self.database,
                            parent=parent)[0]
                        log.warning(u"create_missing_categories p:'%s' c:'%s'".encode('utf-8') \
                            % (parent, category))
                        return category
                    else:
                        log.warning(u"create_missing_categories p:'%s' only parent".encode('utf-8') \
                            % parent)
                        return parent
    
    def get_or_create_unit(self, unit=None, unittype=None):
        if unit in self.cache.unit:
            return self.cache.unit[unit]
        # Try by unit
        try:
            unit_obj = Unit.objects.get(unit=unit)
            self.cache.unit[unit] = unit_obj
            return unit_obj
        except ObjectDoesNotExist:
            pass
        except MultipleObjectsReturned:
            try:
                if not isinstance(unittype, UnitType):
                    unittype = UnitType.objects.get(unittype=unittype)
                unit_ = Unit.objects.get(unit=unit, unittype=unittype)
                self.cache.unit[unit] = unit_
                return unit_
            except ObjectDoesNotExist:
                pass
        # Try by abbreviation
        try:
            unit_obj = Unit.objects.get(abbreviation=unit)
            self.cache.unit[unit] = unit_obj
            return unit_obj
        except ObjectDoesNotExist:
            pass
        except MultipleObjectsReturned:
            try:
                if not isinstance(unittype, UnitType):
                    unittype = UnitType.objects.get(unittype=unittype)
                unit_ = Unit.objects.get(unit=unit, unittype=unittype)
                self.cache.unit[unit] = unit_
                return unit_
            except ObjectDoesNotExist:
                pass
        # Unit creation
        if not unit:
            # The zen of get 'Unknown Unit'
            if hasattr(self, "unknown_unit"):
                return self.unknown_unit
            if not hasattr(self, "unknown_unittype"):
                log.warning(u"creating unittype Unknown".encode('utf-8'))
                self.unknown_unittype = UnitType.objects.get_or_create(
                    unittype="Unknown")[0]
            self.unknown_unit = self.get_or_create_unit(unit="Unknown", 
                unittype=self.unknown_unittype)
            return self.unknown_unit
        if not unittype:
            if not hasattr(self, "unknown_unittype"):
                log.warning(u"creating unittype Unknown".encode('utf-8'))
                self.unknown_unittype = UnitType.objects.get_or_create(
                    unittype="Unknown")[0]
            unittype = self.unknown_unittype
        if not isinstance(unittype, UnitType):
            log.warning(u"creating unittype %s".encode('utf-8') % unittype)
            unittype = UnitType.objects.get_or_create(unittype=unittype)[0]
        log.info(u"creating unit %s".encode('utf-8') % unit)
        unit_ = Unit.objects.create(unittype=unittype, unit=unit)
        self.cache.unit[unit] = unit_
        return unit_
    
    ### Add Categories ###
    def process_category_child_node(self, child_node, parent=None):
        """Recursive processing function."""
        other_info = {}
        if 'type' in child_node.xml_attributes:
            other_info['type'] = child_node.type
        if 'localName' in child_node.xml_attributes:
            other_info['translated_name'] = {
                'language': self.default_local_language, 
                'name': child_node.localName}
        category = Category.objects.get_or_create(category=child_node.name,
                        other_info=simplejson.dumps(other_info),
                        database=self.database,
                        parent=parent)[0]
        for child_node in child_node.childNodes:
            self.process_category_child_node(child_node, parent=category)
    
    def add_categories(self):
        try:
            parsedfile = amara.parse(self.category_filename, 
                rules=self.amara_rules)
        except:
            raise ImportSetupError, \
                "Can't find or import categories definition file"
        for child_node in parsedfile.categories.childNodes:
            self.process_category_child_node(child_node)
    
    ### Add Geographies ###
    def add_geographies(self):
        """
There are references to countries which we already include, and references to custom entitys.
        """
        # TODO: Need separate class for importing geographies by themselves.
        # TODO: Add parsed data from io/data
        raise NotImplementedError
        # if not hasattr(self, "world_database"):
        #     self.get_parent_geography_databases()
        # 
        # try:
        #     parsedfile = amara.parse(self.locations_filename,
        #         rules=self.amara_rules)
        # except OSError:
        #     raise ImportSetupError, "Can't find locations definition file"
        # 
        # for child in parsedfile.regionalCodes.childNodes:
        #     # Don't insert existing geographies
        #     # TODO: Should insert to self.geography_database
        #     if not hasattr(self, "world_database"):
        #         self.world_database = GeographyDatabase.objects.get_or_create(
        #             database="World")[0]
        #     if Geography.objects.filter(
        #             database=self.world_database,
        #             alpha2=child.xml_child_text).count() == 1:
        #         continue
        #     
        #     if hasattr(child, "description"):
        #         geography = child.description
        #     else:
        #         geography = child.xml_child_text
        #     
        #     if geography == "Global":
        #         # Already created
        #         continue
        #     
        #     Geography.objects.get_or_create(
        #         database=self.geography_database,
        #         geography=geography,
        #         imported_database_code=str(child.xml_child_text))
    
    ### Add Units ###
    def add_units(self):
        """By default, there are no unit types in Ecospold."""
        try:
            parsedfile = amara.parse(self.units_filename, 
                rules=self.amara_rules)
        except:
            raise ImportSetupError, "Can't find or import 'Units.xml' file"
        for child_node in parsedfile.units.childNodes:
            other_info = {}
            if 'localUnit' in child_node.xml_attributes:
                other_info['translated_name'] = {
                    'language': self.default_local_language, 
                    'name': child_node.localUnit}
            
            if not hasattr(self, "unknown_unittype"):
                log.warning(u"creating unittype Unknown".encode('utf-8'))
                self.unknown_unittype = UnitType.objects.get_or_create(
                    unittype="Unknown")[0]
            
            Unit.objects.get_or_create(
                abbreviation=child_node.xml_child_text,
                unit=child_node.xml_child_text,
                unittype=self.unknown_unittype,
                other_info=simplejson.dumps(other_info))
    
    ### Person ###
    def create_vcard(self, person):
        if not hasattr(person, "name"):
            return None
        v = vobject.vCard()
        v.add('n') # Ecospold doesn't split first/last names
        v.add('fn')
        v.fn.value = person.name
        if hasattr(person, "email"):
            v.add('email')
            v.email.value = person.email
            v.email.type_param = 'INTERNET'
        if hasattr(person, "telephone"):
            v.add('tel')
            v.tel.value = person.telephone
            v.tel.type_param = 'WORK'
        if hasattr(person, "address"):
            v.add('adr')
            v.adr.value = person.address
        if hasattr(person, "companyCode"):
            v.add('org')
            v.org.value = person.companyCode
        if hasattr(person, "countryCode"):
            v.add('geo')
            v.geo.value = person.countryCode
        return v.serialize()
    
    def create_person(self, person):
        if hasattr(person, "vCard"):
            vcard = person.vCard
        else:
            vcard = self.create_vcard(person)
        
        if hasattr(person, "uuid"):
            uuid = person.uuid
        else:
            uuid = None
        
        return Person.objects.get_or_create(
            uuid=uuid,
            vcard=vcard,
            person=person.name)
    
    ### Processes ###
    # TODO: This should be changed when we decide upon a better format.
    def build_process_description_node(self, text, title):
        return u"%s\n%s\n\n%s\n\n" % (title, "="* len(title), text)
    
    def get_process_description(self, process_info, dataset):
        ref_func = process_info.referenceFunction
        description_elements = (
            (getattr(ref_func, "generalComment", "None provided."), 
                "General Comment"),
            (getattr(process_info.technology, "text", "None provided."), 
                "Technology"),
            (getattr(process_info.geography, "text", "None provided."), 
                "Geography"),
            (getattr(ref_func, "includedProcesses", "None provided."), 
                "Included Processes"),
            (getattr(
                getattr(dataset.metaInformation.modellingAndValidation, 
                    "representativeness", Container()),
                "samplingProcedure", "None provided."), 
            "Sampling Procedure"),
            )
        
        return u"".join([self.build_process_description_node(obj[0], obj[1]
            ) for obj in description_elements])
    
    def create_or_update_process(self, dataset):
        """CRUD method for processes. Because internal codes are so unreliable, we ignore them."""
        process_info = dataset.metaInformation.processInformation
        ref_func = process_info.referenceFunction

        finder = {
            "process": ref_func.name, 
            "category": [ref_func.category, ref_func.subCategory],
            "unit": ref_func.unit}
        if hasattr(process_info.geography, "location"):
            finder["geography"] = process_info.geography.location

        process = self.process_finder.find(finder)

        unit = self.get_or_create_unit(ref_func.unit)
        category = self.get_or_create_category(parent=ref_func.category,
            category=ref_func.subCategory)

        if hasattr(process_info.geography, "location"):
            geography = self.get_or_create_geography(
                process_info.geography.location)
        else:
            geography = self.global_geography
            log.error(u"create_process no geography %s".encode('utf-8') % \
                process_info.geography.xml())

        other_info_dict = {}
        try:
            parameters = (('generator', dataset, "generator",),
                ('schema version', dataset, "internalSchemaVersion"),
                ('infrastructure included', ref_func, 
                    "infrastructureIncluded"),
                ('relates to product', ref_func, "datasetRelatesToProduct"),
                ('infrastructure process', ref_func, "infrastructureProcess"))

            for p in parameters:
                if hasattr(p[1], p[2]):
                    other_info_dict[p[0]] = getattr(p[1], p[2])

            if hasattr(ref_func, "localName"):
                other_info_dict['translated_name'] = {
                    'language': self.default_local_language, 
                    'name': ref_func.localName}
        except AttributeError:
            # Malformed EcoSpold is missing attributes
            pass

        attributes = {
            'process': ref_func.name,
            'imported_database_code': str(dataset.number),
            'database': self.database,
            'category': category,
            'geography': geography,
            'unit': unit,
            'description': self.get_process_description(process_info, 
                dataset),
            'creator': self.user,
            'process_type': 0,
            'other_info': simplejson.dumps(other_info_dict)}

        if not process:
            process = Process(**attributes)
            process.save(skip_datetime_update=True)
            return process
        elif isinstance(process, int):
            log.critical("Help! Too many identical processes")
            log.critical(unicode(attributes).encode('utf-8'))
            raise IdenticalProcessesError
        elif self.behaviour == "ignore":
            pass
        else:
            process = self.change_by_behaviour(process, attributes)
            process.save(skip_datetime_update=True)
            return process

    def create_process_from_exchange(self, exchange):
        """Exchange references a process that cannot be found"""
        description = exchange.generalComment if hasattr(exchange, 
            "generalComment") else ""
        
        if hasattr(exchange, "location"):
            geography = self.get_or_create_geography(exchange.location)
        else:
            geography = self.global_geography
        
        unit = self.get_or_create_unit(exchange.unit)
        
        if hasattr(exchange, "category"):
            category = self.get_or_create_category(parent=exchange.category,
                category=exchange.subCategory)
        else:
            category = self.get_or_create_category()
        
        if hasattr(exchange, "inputGroup"):
            if exchange.inputGroup.xml_child_text == "4":
                # Resource
                process_type = 1
            elif exchange.inputGroup.xml_child_text in ("0", "1", "2", "3", 
                    "5"):
                process_type = 0
            else:
                raise ValueError
        elif hasattr(exchange, "outputGroup"):
            if exchange.outputGroup.xml_child_text == "4":
                # Emissions
                process_type = 2
            elif exchange.outputGroup.xml_child_text in ("0", "1", "2", "3", 
                    "5"):
                process_type = 0
            else:
                raise ValueError
        else:
            process_type = 0

        log.info(u"create_missing_process\n\tprocess: %s\n\tdatabase: %s\n\tcategory: %s\n\tprocess_type: %s".encode('utf-8') % \
            (exchange.name, self.database, category, process_type))

        return Process.objects.create(
            imported_database_code=exchange.number if exchange.number not in \
                ("0", "1") else '',
            process=exchange.name,
            database=self.database,
            category=category,
            geography=geography,
            unit=unit,
            description=description,
            creator=self.user,
            process_type=process_type,
            )

    def add_processes(self):
        if self.progressbar:
            widgets = ['Processes: ', pb.Percentage(), ' ', 
                pb.Bar(marker=pb.RotatingMarker()), ' ', pb.ETA()]
            pbar = pb.ProgressBar(widgets=widgets, maxval=len( 
                self.dataset_files)).start()
        for index, filename in enumerate(self.dataset_files):
            try:
                log.info(u"add_processes filename %s".encode('utf-8') % \
                    filename)
                parsedfile = amara.parse(os.path.join(self.path, filename), 
                    rules=self.amara_rules)
                if not hasattr(parsedfile, "ecoSpold"):
                    # Unrecognized file type
                    log.critical(u"skipping %s - no ecoSpold element".encode( 
                        'utf-8') % filename)
                    continue
                for dataset in parsedfile.ecoSpold.childNodes:
                    process = self.create_or_update_process(dataset)
                    if self.progressbar:
                        pbar.update(index)

            except AttributeError, e:
                log.critical(u"file %s raises attribute error while parsing process. Error:\n%s".encode('utf-8') % (filename, e))
                continue
            except:
                self.rollback_transaction()
                raise

        if self.progressbar:
            pbar.finish()
    
    ### Exchanges ###
    def translate_exchange_values(self, exchange):
        """
Translate crazy ecospold parameters into numbers that make sense.

Translation table:

=============== ======= ======= =============== ================================= ================= ===========
Distribution    ES ID   BW ID   Parameter       Ecospold                          Brightway         Translation
=============== ======= ======= =============== ================================= ================= ===========
Lognormal       1       2       Geometric Mu    meanValue                         number            
..              ..      ..      Geometric Sigma standardDeviation95 *(squared)*   sigma             sd95**0.5
Normal          2       3       Mu              meanValue                         number
..              ..      ..      Sigma           standardDeviation95 *(times two)* sigma             sd95/2
Triangular      3       5       Minimum         minValue                          minimum
..              ..      ..      Mode            mostLikelyValue                   number
..              ..      ..      Maximum         maxValue                          maximum
Uniform         4       4       Minimum         minValue                          minimum
..              ..      ..      Mean            meanValue                         number
..              ..      ..      Maximum         maxValue                          maximum
=============== ======= ======= =============== ================================= ================= ===========

Unknown of undefined uncertainty distributions are ID 0, and Brightway uses ID 1 for exchanges with no uncertainty (usually unit conversions).
        """
        
        attrs = {}

        (UndefinedUncertainty, NoUncertainty, 
            LognormalUncertainty, NormalUncertainty, UniformUncertainty, 
            TriangularUncertainty, BernoulliUncertainty)

        if hasattr(exchange, 'generalComment'):
            attrs['description'] = exchange.generalComment
        if not hasattr(exchange, "uncertaintyType"):
            attrs['amount'] = float(exchange.meanValue)
            # For aggregated processes, a lognormal distribution is implied
            # but not specified. Test using 'minimum' and 'maximum'
            # TODO: Figure out these values, as they don't seem to make sense:
            # e.g. meanValue="1.5859E-9" standardDeviation95="1.5502E-9" 
            # minValue="9.2394E-11" maxValue="5.908E-9"
            if hasattr(exchange, "standardDeviation95") and hasattr(exchange, 
                    'minValue') and hasattr(exchange, 'maxValue') and \
                    exchange.standardDeviation95 and exchange.maxValue and \
                    False:
                pass
        elif exchange.uncertaintyType == "1":
            # Lognormal
            attrs['uncertainty_type'] = LognormalUncertainty.id
            attrs['amount'] = float(exchange.meanValue)
            attrs['sigma'] = math.log(math.sqrt(float( 
                exchange.standardDeviation95)))
            if attrs['sigma'] == 0:
                # Bad ecoinvent data
                attrs = {'uncertainty_type':  UndefinedUncertainty.id, 
                    'amount': float(exchange.meanValue)}
        elif exchange.uncertaintyType == "2":
            # Normal
            attrs['uncertainty_type'] = NormalUncertainty.id
            attrs['amount'] = float(exchange.meanValue)
            attrs['sigma'] = float(exchange.standardDeviation95)/2
        elif exchange.uncertaintyType == "3":
            # Triangular
            attrs['uncertainty_type'] = TriangularUncertainty.id
            attrs['minimum'] = float(exchange.minValue)
            attrs['maximum'] = float(exchange.maxValue)
            # Sometimes this isn't included (though it SHOULD BE)
            if hasattr(exchange, "mostLikelyValue"):
                attrs['amount'] = float(exchange.mostLikelyValue)
            else:
                attrs['amount'] = float(exchange.meanValue)
        elif exchange.uncertaintyType == "4":
            # Uniform
            attrs['uncertainty_type'] = UniformUncertainty.id
            attrs['amount'] = float(exchange.meanValue)
            attrs['minimum'] = float(exchange.minValue)
            attrs['maximum'] = float(exchange.maxValue)
        elif exchange.uncertaintyType == "0":
            # None
            attrs['uncertainty_type'] = UndefinedUncertainty.id
            attrs['amount'] = float(exchange.meanValue)
        return attrs

    def create_exchanges(self, dataset):
        process_info = dataset.metaInformation.processInformation
        ref_func = process_info.referenceFunction

        finder = {
            "process": ref_func.name, 
            "category": [ref_func.category, ref_func.subCategory],
            "unit": ref_func.unit}
        if hasattr(process_info.geography, "location"):
            finder["geography"] = process_info.geography.location

        output = self.process_finder.find(finder)

        if not isinstance(output, Process):
            log.critical(u"create_exchanges unknown output process. process_finder found %i values. Finder inputs:\n%s".encode('utf-8'
                ) % (output, finder))
            raise MissingProcessError
        for exchange in dataset.flowData.childNodes:
            # Skip allocation
            if not hasattr(exchange, "name"):
                # TODO: Log this
                continue
            # Skip definitional exchange - we assume this already
            if exchange.name == ref_func.name and float(exchange.meanValue
                    ) == 1.0:
                continue

            finder = {"process": exchange.name}
            if hasattr(exchange, "unit"):
                finder["unit"] = exchange.unit
            category = []
            try:
                category.append(exchange.category)
            except AttributeError:
                pass
            try:
                # Skip subcategories that shouldn't exist
                sc = exchange.subCategory
                if sc != "unspecified":
                    category.append(sc)
            except AttributeError:
                pass
            if category:
                finder["category"] = category
            # "4" is a biosphere flow, which must have Global geography
            if hasattr(exchange, "location") and (exchange.outputGroup.\
                    xml_child_text != "4" if hasattr(exchange, "outputGroup"
                    ) else True):
                finder["geography"] = exchange.location
            input = self.process_finder.find(finder)

            if isinstance(input, Process):
                attributes = self.translate_exchange_values(exchange)
                existing = Exchange.objects.filter(input=input, 
                    output=output)
                count = existing.count()
                if count != 1:
                    if count > 1:
                        # Multiple existing exchanges were found. There is 
                        # no way to tell which should be modified, so log 
                        # this as an error and create a new exchange. See 
                        # also discussion in 
                        # apps/inventory/models.Exchange.meta
                        log.warning(u"create_exchanges multiple exchanges for input/output combination. Found %i exchanges. Input: %s, output: %s".encode( 
                            'utf-8') % (count, input, output))

                    e = Exchange(input=input, 
                        output=output,
                        creator=self.user,
                        **attributes)
                    e.save(skip_datetime_update=True, atomic=False)
                elif self.behaviour == "ignore":
                    continue
                else:
                    self.change_by_behaviour(existing[0], attributes)

            elif input > 0:
                log.critical(u"create_exchanges multiple input processes. process_finder found %i values. Finder inputs:\n%s".encode('utf-8'
                    ) % (input, finder))
                raise IdenticalProcessesError
            else:
                input = self.create_process_from_exchange(exchange)

    def add_exchanges(self):
        if self.progressbar:
            widgets = ['Exchanges: ', pb.Percentage(), ' ', 
                pb.Bar(marker=pb.RotatingMarker()), ' ', pb.ETA()]
            pbar = pb.ProgressBar(widgets=widgets, maxval=len( 
                self.dataset_files)).start()

        for index, filename in enumerate(self.dataset_files):
            try:
                log.debug(u"add_exchanges %s".encode('utf-8') % filename)
                parsedfile = amara.parse(os.path.join(self.path, filename), 
                    rules=self.amara_rules)
                # Sometimes there are multiple datasets in one XML file
                for dataset in parsedfile.ecoSpold.childNodes:
                    self.create_exchanges(dataset)
            except AttributeError, e:
                log.critical(u"file %s raises attribute error while parsing exchanges. Error:\n%s".encode('utf-8') % (filename, e))
            except:
                self.rollback_transaction()
                raise

            if self.progressbar:
                pbar.update(index)

        if self.progressbar:
            pbar.finish()
