import os
import datetime
import logging as log
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.contrib.auth.models import User
from django.conf import settings
try:
    import progressbar as pb
except ImportError:
    pb = None
from brightway.io.importer_user import get_importer_user

def get_importer_user():
    try:
        user = User.objects.get(username="Importer")
    except ObjectDoesNotExist:
        user = User.objects.create_user("Importer", 
            "fake@example.com", "Import Script")
        user.is_superuser = True
        user.is_staff = True
        user.save()
    return user


class ImportBase(object):
    """Base class that provides common functionality for importing."""

    def __init__(self, user=None, behaviour="update_if_provided", 
            show_progress=False):
        super(ImportBase, self).__init__()

        if pb and show_progress:
            self.progressbar = True
        else:
            self.progressbar = False

        self.behaviour = behaviour
        self.setup_logging()
        self.get_user(user)

    ### Logging ###
    def setup_logging(self):
        """
See http://docs.python.org/lib/module-logging.html.
Logging levels:

* CRITICAL. 50. Use for big bad errors.
* ERROR. 40. Use for errors which need immediate attention.
* WARNING. 30. Use for errors when something unexpected happened, e.g. calculation failed or cache load failed.
* INFO. 20. Use for results and timing of method call, for example.
* DEBUG. 10. Use for low level description of everything that is happening.
        """
        logging_config = {
            'level': log.INFO,
            'format': '%(lineno)d %(message)s',
            'filename': os.path.join(settings.BRIGHTWAY_ROOT_PATH,
                'logs/import-%s-%s.log' % (self.__class__.__name__, 
                datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S"))),
            }
        log.basicConfig(**logging_config)

    ### Filters ###
    def xmlfilter(self, filename):
        """Use file extension to find xml files"""
        if str(filename)[-4:].lower() == ".xml":
            return True
        else:
            return False

    ### Transactions ###
    def start_transaction(self):
        transaction.enter_transaction_management()
        transaction.managed(True)
    
    def rollback_transaction(self):
        transaction.rollback()
    
    def commit_transaction(self):
        transaction.commit()

    def leave_transaction(self):
        transaction.leave_transaction_management()

    ### Update functions by behaviour ###
    def change_by_behaviour(self, model, attributes):
        """Choose version of updating to apply to existing instances."""
        if self.behaviour == "ignore":
            # Don't change existing model
            return model
        if self.behaviour == "update":
            return self._update(model, attributes)
        if self.behaviour == "update_if_provided":
            return self._update_if_provided(model, attributes)
        if self.behaviour == "overwrite":
            return self._overwrite(model, attributes)
    
    def _update_if_provided(self, model, attributes):
        """Only update attributes if value is provided"""
        for key in attributes:
            if attributes[key] and getattr(model, key, None) != \
                    attributes[key]:
                log.debug(u"_update_if_provided model %s field %s value %s".encode('utf-8') % \
                    (model, key, attributes[key]))
                setattr(model, key, attributes[key])
        return model
    
    def _update(self, model, attributes):
        """Updates for all provided attributes"""
        for key in attributes:
            if getattr(model, key, None) != attributes[key]:
                log.debug(u"_update model %s field %s value %s".encode('utf-8') % (model, key, 
                    attributes[key]))
                setattr(model, key, attributes[key])
        return model
    
    def _overwrite(self, model, attributes):
        """Updates all model fields; attributes not provided set to None"""
        fields = [m.name for m in model._meta.fields]
        for field in fields:
            if field in attributes and getattr(model, field, None) != \
                    attributes[field]:
                log.debug(u"_overwrite model %s field %s value %s".encode('utf-8') % (model, 
                    field, attributes[field]))
                setattr(model, field, attributes[field])
            else:
                log.debug(u"_overwrite model %s setting %s to None".encode('utf-8') % (model, 
                    field))
                setattr(model, field, None)
        return model

    ### Defaults ###
    def get_user(self, user=None):
        """Get or create user. Use default 'Importer' if no user provided."""
        if not user:
            self.user = get_importer_user()
        elif isinstance(user, User):
            self.user = user
        elif isinstance(user, (str, unicode)):
            self.user = User.objects.get(username=user)
        else:
            raise ImportSetupError, "Can't parse given value for 'user': %s"\
                % user
