import os
import xlrd

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.db import transaction
from django.contrib.auth.models import User

from brightway.apps.inventory.models import Database, Category, Process, \
    Exchange
from brightway.apps.impact_assessment.models import Method, Weighting
from brightway.apps.unit.models import Unit, UnitType
from brightway.apps.geo.models import Geography, GeographyDatabase
from brightway.io.data.ecoinvent_mapping import GENERIC_CATEGORY_MAPPING, ECOINVENT_13_TO_2_CATEGORY_MAPPING, ECOINVENT_13_TO_2_PROCESS_MAPPING

class RecipeImporter(object):
    def __init__(self, filename=None, worksheet=1):
        if filename:
            self.filename = filename
        else:
            self.filename = os.path.join(settings.BRIGHTWAY_ROOT_PATH, "io/data/recipe/recipe_101.xls")
        self.file = xlrd.open_workbook(self.filename)
        self.worksheet = worksheet
        # Hackish
        self.database = Database.objects.get(database__icontains="ecoinvent 2")
        self.unit_dict = {}
        self.get_kilogram()
        self.category_dict = {}
        self.process_dict = {}
        self.bad_list = []
        self.get_process_creation_constants()

    def go(self, worksheet=None):
        self.start_transaction()
        try:
            self._import(worksheet)
        except:
            self.rollback_transaction()
            raise

    def _import(self, worksheet=None):
        self.recipe = Method.objects.get_or_create(method="ReCiPe 1.01")[0]
        
        if worksheet == None:
            worksheet = self.worksheet
        sheet = self.file.sheet_by_index(worksheet)
        
        for row_index in range(sheet.nrows):
            if row_index % 200 == 0 and row_index:
                print "Row %i" % row_index
            
            if row_index >= 5930:
                break
            
            row = sheet.row_values(row_index)
            
            # Blank lines
            if not row[0]:
                continue
            
            # Header
            if (not row[1] and not row[2]) or row[0] == "Compartment":
                continue
            
            # New method
            if (not row[2] and not row[3]):
                print "Method %s %s (%i)" % (row[0], row[1], row_index)
                current_method = Method.objects.get_or_create(method=row[0], description="Unit: %s" % row[1], parent=self.recipe)[0]
                continue
            
            # New value
            category = self.get_category(row)
            process = self.get_process(row, category)
            
            try:
                weighting = Weighting.objects.get(method=current_method, process=process)
                weighting.amount = row[4]
                weighting.save(skip_datetime_update=True)
            except ObjectDoesNotExist:
                weighting = Weighting.objects.create(method=current_method, process=process, amount=row[4])
        
        # Reset modified timestamp
        process.save()
        self.recipe.save()

    def get_process_creation_constants(self):
        try:
            self.user = User.objects.get(username="Importer")
        except ObjectDoesNotExist:
            self.user = User.objects.create_user("Importer", 
                "fake@example.com", "Import Script")
            self.user.is_superuser = True
            self.user.is_staff = True
            self.user.save()
        
        geography_database = GeographyDatabase.objects.get(
            database__in=("Default world", "World"))
        
        try:
            self.global_geography = Geography.objects.get(
                database=geography_database,
                geography="World")
        except ObjectDoesNotExist:
            self.global_geography = Geography.objects.create(
                database=geography_database,
                geography="World")

    def create_process(self, process, category, unit, process_type=2):
        if category.parent.category.lower() == "resource":
            process_type = 1
        print "Creating", process
        return Process.objects.create(
            process=process,
            category=category,
            process_type=process_type,
            database=self.database,
            geography=self.global_geography,
            creator=self.user,
            unit=unit)

    def get_process(self, row, category):
        p_key = (row[0], row[1], row[2], row[5])
        if p_key in self.process_dict:
            return self.process_dict[p_key]
        
        p_label = row[2]
        if p_label in ECOINVENT_13_TO_2_PROCESS_MAPPING:
            p_label = ECOINVENT_13_TO_2_PROCESS_MAPPING[p_label]

        unit = self.get_unit(row[5])
        
        # Simple lookup
        try:
            process = Process.objects.get(category=category, 
                process__iexact=p_label, 
                database=self.database,
                process_type__in=(1,2))
        except ObjectDoesNotExist:
            process = self.create_process(p_label, category, unit)
            self.bad_list.append(p_label)
        except MultipleObjectsReturned:
            process = Process.objects.get(category=category, 
                process__iexact=p_label, 
                unit=unit,
                database=self.database)
        
        if not process.cas_number and row[3]:
            process.cas_number = row[3]
            process.save(skip_datetime_update=True)
        
        self.process_dict[p_key] = process
        return process

    def get_unit(self, label):
        if label in self.unit_dict:
            return self.unit_dict[label]
        
        if label == "kg":
            self.unit_dict["kg"] = self.kilogram
            return self.kilogram
        else:
            try:
                unit = Unit.objects.get(unit=label)
            except ObjectDoesNotExist, MultipleObjectsReturned:
                unit = Unit.objects.get(abbreviation=label)
        
        self.unit_dict[label] = unit
        return unit

    def get_kilogram(self):
        try:
            self.kilogram = Unit.objects.get(unit="kilogram")
            return
        except ObjectDoesNotExist:
            pass
        
        try:
            self.kilogram = Unit.objects.get(abbreviation="kg")
            return
        except ObjectDoesNotExist:
            pass
        
        self.kilogram = Unit.objects.get(unit="kg")

    def get_category(self, row):
        c_key = (row[0], row[1])
        if c_key in self.category_dict:
            return self.category_dict[c_key]
        
        parent_category = Category.objects.filter(database=self.database, parent__isnull=True, category__iexact=row[0])
        c_label = row[1]
        if c_label in GENERIC_CATEGORY_MAPPING:
            c_label = GENERIC_CATEGORY_MAPPING[c_label]
        elif c_label in ECOINVENT_13_TO_2_CATEGORY_MAPPING:
            c_label = ECOINVENT_13_TO_2_CATEGORY_MAPPING[c_label]
        category = Category.objects.get(category__iexact=c_label, parent=parent_category)
        self.category_dict[c_key] = category
        return category

    def start_transaction(self):
        transaction.enter_transaction_management()
        transaction.managed(True)

    def rollback_transaction(self):
        transaction.rollback()

    def commit_transaction(self):
        transaction.commit()

if __name__ == "__main__":
    R = RecipeImporter()
    R.go()