# -*- coding: utf-8 -*
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.db.models import Q
from django.db.models.query import QuerySet
from brightway.apps.inventory.models import Category, Process, Database
from brightway.apps.geo.models import Geography
from brightway.core.utils.safe_hash import generate_safe_hash
from brightway.core import Container

class ProcessFinder(object):
    """A class that uses a series of attribute combinations to find Processes. Inputs are provided in a standardized dictionary format, and cached. Used to import data from poor importers where process references are not clear. Assume only that process name is given.

The general approach is to iterate through a series of parameters combinations until exactly one process is found. Note that this approach is dumb, in that it does not adjust the number of parameters based on whether it is too liberal or too restrictive.

.. rubric:: Inputs

* database: Database that processes are located in.
* process_name_mapping: dictionary, optional. Map old process names to new ones.
* category_mapping: dictionary, optional. Map old category names to new ones.
    """
    # TODO: Include DAG to get parent databases
    # TODO: Re-include progressive guessing
    def __init__(self, database=None, parent_databases=None,    
            process_name_mapping={}, category_mapping={}):
        super(ProcessFinder, self).__init__()
        self.database = self.get_database_qs(database)
        self.parent_databases = self.get_database_qs(parent_databases)
        self.process_name_mapping = process_name_mapping
        self.category_mapping = category_mapping
        self.cache = {}

    def get_database_qs(self, db):
        if isinstance(db, Database):
            return Database.objects.filter(id=db.id)
        elif isinstance(db, QuerySet):
            return db
        elif isinstance(db, int):
            return Database.objects.filter(id=db)
        else:
            return Database.objects.none()

    def get_hash(self, obj):
        """Generate hash for caching/looking up processes."""
        return generate_safe_hash(u".".join([unicode(obj.get(key, "n")
            ) for key in ("process", "category", "unit", "geography")]))

    def map_values(self, obj):
        if obj["process"] in self.process_name_mapping:
            obj["process"] = self.process_name_mapping[obj["process"]]
        for index, value in enumerate(obj.get("category", [])):
            if value in self.category_mapping:
                obj["category"][index] = self.category_mapping[value]
        return obj

    def get_category_filter_string(self, obj):
        """Build up filter strings, like parent__parent__category"""
        filters = {}
        for index, value in enumerate(obj["category"][::-1]):
            filters["category__" + "parent__" * index + "category"] = value
        return filters

    def find(self, obj):
        assert "process" in obj
        obj = self.map_values(obj)
        obj_hash = self.get_hash(obj)
        try:
            return self.cache[obj_hash]
        except KeyError:
            process = self._find(obj)
            if process:
                self.cache[obj_hash] = process
            return process

    def _find(self, obj):
        qs = Process.objects.filter(process__iexact=obj["process"]).filter( 
            Q(database__in=self.database) | Q( 
            database__in=self.parent_databases))
        if obj.get("unit", None):
            qs = qs.filter(Q(unit__unit__iexact=obj["unit"]) | Q(
                unit__abbreviation__iexact=obj['unit']))
        if obj.get("geography", None):
            qs = qs.filter(Q(geography__geography=obj['geography']
                ) | Q(geography__imported_database_code__iexact=obj[ 
                "geography"]))
        if obj.get("category", None):
            category_filters = self.get_category_filter_string(obj)
            qs = qs.filter(**category_filters)
        if qs.count() == 1:
            return qs[0]
        else:
            return qs.count()

