# -*- coding: utf-8 -*-
import datetime
import math
import decimal
import os
from numpy import matrix, identity, zeros, array
from scipy import sparse
from django.conf import settings
from django.contrib.auth.models import User, Group
from django.core.exceptions import ObjectDoesNotExist
from brightway.tests.base_test_case import BaseTestCase
from brightway.apps.inventory.models import Database, Category, Process, \
    Exchange
from brightway.apps.impact_assessment.models import Method, Weighting
from brightway.apps.unit.models import Unit, UnitType
from brightway.apps.geo.models import GeographyDatabase, Geography
# from brightway.core.logs.loggers import lca_test_logger
from brightway.core import Container
from django_dag.models import Node
from brightway.apps.geo import WORLD


class LCATestCase(BaseTestCase):
    def setUp(self):
        super(LCATestCase, self).setUp()
        self.cache_dir = os.path.join(settings.BRIGHTWAY_ROOT_PATH, 
            'core/tests/cache/')
        self.redirect_kwargs = {
            'cache_dir': self.cache_dir,
            # 'logging': lca_test_logger,
            'use_inheritance': True
            }
        self.setup_db()

    def tearDown(self):
        self.clean_cache_directory()
        super(LCATestCase, self).tearDown()

    def clean_cache_directory(self):
        def tar_filter(x):
            if x.lower()[-4:] == ".tar":
                return True
            return False
        
        files = os.listdir(self.cache_dir)
        tar_files = filter(tar_filter, files)
        for tf in tar_files:
            os.remove(os.path.join(self.cache_dir, tf))

    def setup_db(self):
        # For our test system, we set up an industrial system, with the 
        # following processes:
        process_names = ("coal,steel,electricity coal,electricity wind,electricity mix").split(",")
        # and in a separate database
        extra_process_name = "widget"

        # See code below for their attributes (category, geography, etc.).
        # Here is our technological matrix, with inputs on rows and outputs on
        # the columns:
        self.technological_matrix = matrix(  
            #Coal     Steel   E. Coal E. Wind E. Mix
            [[0,      .1,     .25,    0,      0],#Coal
            [.01,     0,      .01,    .02,    0],#Steel
            [0,       0,      0,      0,      .7],#Elec. Coal
            [0,       0,      0,      0,      .3],#Elec. Wind
            [.1,      .1,     .05,    .01,    0],#Elec. Mix
            ])

        self.emissions_matrix = matrix(
            [ #Coal       Steel     E. Coal   E. Wind E. Mix
            [0.05,      0.01,       0.5,    0,      0], # NOx
            [0.2,       0.1,        1,      -0.01,      0], # PM
            [1,         0,          0,      0,      0], # Coal in ground
            ])

        """Now the paperwork."""
        self.groups = Container()
        self.groups.admins = Group.objects.create(name="admins")
        self.groups.users = Group.objects.create(name="users")
        self.groups.test = Group.objects.create(name="test group")
        self.groups.everyone = Group.objects.get(name="Everyone")

        self.users = Container()
        self.users.admin = User.objects.create_user("test admin", "a@a.com", 
            "admin")
        self.users.admin.is_superuser = True
        self.users.admin.is_staff = True
        self.users.admin.groups = (self.groups.admins, self.groups.test, 
            self.groups.everyone)
        self.users.admin.save()

        self.users.testuser = User.objects.create_user("test user", "a@a.com", 
            "user")
        self.users.testuser.groups = (self.groups.users, self.groups.test, 
            self.groups.everyone)
        self.users.testuser.save()

        self.users.public = User.objects.get(username="public")

        #Databases
        self.databases = Container()
        self.databases.master = self.create_database(
            database="Test Master Database", 
            creator=self.users.admin, 
            database_type=0,
            managers=[self.groups.admins, self.groups.users, 
                self.groups.test], 
            viewers=[self.groups.everyone,])
        self.databases.user = self.create_database(
            database="Test User Database", 
            creator=self.users.admin, 
            database_type=2,
            managers=[self.groups.admins, self.groups.users, 
                self.groups.test])
        self.databases.user.add_parent(self.databases.master)
        self.databases.public = self.create_database(
            database="Public Database", 
            creator=self.users.public, 
            database_type=2,
            managers=[self.groups.everyone,])
        self.databases.admins_only = self.create_database(
            database="Admins Only Database", 
            creator=self.users.admin, 
            database_type=0,
            managers=[self.groups.admins,])

        #Geographies
        self.geographies = Container()
        self.geographies.worlddb = GeographyDatabase.objects.get( 
            database="World")

        self.geographies.switzerland = Geography.objects.create( 
            geography="Switzerland", database=self.geographies.worlddb)
        self.geographies.germany = Geography.objects.create( 
            geography="Germany", database=self.geographies.worlddb)
        self.geographies.europe = Geography.objects.create(geography="Europe", 
            database=self.geographies.worlddb)
        self.geographies.world = WORLD()

        #Unittypes
        self.units = Container()
        self.units.mass = UnitType.objects.create(unittype="Mass")
        self.units.volume = UnitType.objects.create(unittype="Volume")
        self.units.human_health = UnitType.objects.create(
            unittype="Human Helath")

        #Units
        self.units.kg = Unit.objects.create(unit="Kilogram", 
            abbreviation="kg", unittype=self.units.mass)
        self.units.nm3 = Unit.objects.create(unit="Normal Cubic Meter", 
            abbreviation="nm3", unittype=self.units.volume)
        self.units.kwh = Unit.objects.create(unit="Kilowatt hour", 
            abbreviation="kWh", unittype=self.units.volume)
        self.units.dalys = Unit.objects.create(unit="DALYs", 
            abbreviation="dalys", unittype=self.units.human_health)

        self._default_user = self.users.admin
        self._default_db = self.databases.master
        self._world = self.geographies.world
        self._kg = self.units.kg

        #Categories
        self.categories = Container()
        self.categories.fossilfuels = self.create_category("Fossil Fuels")
        self.categories.coal = self.create_category("Coal", 
            parent=self.categories.fossilfuels)
        self.categories.metals = self.create_category("Metals")
        self.categories.energy = self.create_category("Energy")
        self.categories.electricity = self.create_category("Electricity", 
            parent=self.categories.energy)
        self.categories.air = self.create_category("Air")
        self.categories.acidification = self.create_category("Acidification", 
            parent=self.categories.air)
        self.categories.products = self.create_category("Products")
        self.categories.airemissions = self.create_category("Air Emissions")
        self.categories.earth = self.create_category("Earth Resources")
        self.categories.private = self.create_category("Private", 
            database=self.databases.admins_only)

        "coal,steel,aluminium,plastic,oil,electricity coal,electricity oil,electricity wind,electricity mix, widget"
        ### Technological processes ###
        self.processes = Container()
        self.processes.coal = self.create_process("Coal", 
            self.categories.coal, self.geographies.world, self.units.kg)
        self.processes.coal.attributes['dirty'] = True
        # self.processes.coal.attributes['year'] = datetime.date(2000, 1, 1)
        self.processes.steel = self.create_process("Steel", 
            self.categories.metals, self.geographies.world, self.units.kg)
        self.processes.electricity = Container()
        self.processes.electricity.coal = self.create_process(
            "Electricity Coal", self.categories.electricity, 
            self.geographies.world, self.units.kwh)
        self.processes.electricity.coal.attributes['dirty'] = True
        # self.processes.electricity.coal.attributes['year'
        #    ] = datetime.date(2000, 1, 1)
        self.processes.electricity.wind = self.create_process( 
            "Electricity Wind", self.categories.electricity, 
            self.geographies.world, self.units.kwh)
        self.processes.electricity.wind.attributes['dirty'] = False
        # self.processes.electricity.wind.attributes['year' 
        #    ] = datetime.date(2010, 1, 1)
        self.processes.electricity.mix = self.create_process( 
            "Electricity Mix", self.categories.electricity, 
            self.geographies.world, self.units.kwh)
        self.processes.widget = self.create_process("Widget", 
            self.categories.products, self.geographies.world, self.units.kg, 
            database=self.databases.user)

        self.processes.technological = [self.processes.coal, 
            self.processes.steel, self.processes.electricity.coal, 
            self.processes.electricity.wind, self.processes.electricity.mix]

        self.processes.nox = self.create_process("NOx", 
            self.categories.airemissions, self.geographies.world, 
            self.units.kg, process_type=2)
        self.processes.pm = self.create_process("PM", 
            self.categories.airemissions, self.geographies.world, 
            self.units.kg, process_type=2)
        self.processes.coalresource = self.create_process("Coal Resource", 
            self.categories.earth, self.geographies.world, self.units.kg, 
            process_type=1)

        self.processes.emissions = [self.processes.nox, self.processes.pm, 
            self.processes.coalresource]

        def enter_exchanges(matrix, processes, col_processes=None):
            coo = sparse.lil_matrix(matrix).tocoo()
            indices = zip(coo.row, coo.col)
            if not col_processes:
                for row, col in indices:
                    self.create_exchange(processes[row], processes[col], 
                        matrix[row, col])
            else:
                for row, col in indices:
                    self.create_exchange(processes[row], col_processes[col], 
                        matrix[row, col])


        enter_exchanges(self.technological_matrix, 
            self.processes.technological)
        enter_exchanges(self.emissions_matrix, self.processes.emissions, 
            self.processes.technological)

        # Include to test for Ecoinvent conditions, where these exchanges 
        # exist. Raises total number of exchanges to 20
        self.create_exchange(self.processes.coal, self.processes.coal, 1, 
            uncertainty_type=1)

        # Methods and Weightings
        self.methods = Container()
        self.methods.human_health = Method.objects.create(
            method="Human Health", unit=self.units.dalys)

        self.create_weighting(self.methods.human_health, self.processes.nox, 
            1.0)
        self.create_weighting(self.methods.human_health, self.processes.pm, 
            2.0)

